/*
    Copyright 2009  Rivo Laks <rivolaks@hot.ee>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "loadingview.h"

#include "gameworld.h"
#include "ship.h"
#include "planet.h"
#include "model.h"
#include "datastore.h"
#include "shadermanager.h"

#include <kgllib/textrenderer.h>
#include <kgllib/geometrybuffer.h>

#include <KDebug>
#include <QTimer>
#include <QFile>
#include <QDomElement>
#include <QDir>
#include <QApplication>


namespace
{
    template<typename T> T sq(T x)  { return x*x; }
}


LoadingView::LoadingView(KGLLib::GLWidget* parent, GameWorld* world, const QString& levelname) : GLView(parent)
{
    mWorld = world;
    mLevelName = levelname;
    mProgress = 0;
    mFailed = false;

    mStateInfos[Load_NotStarted]            = LoadingStateInfo(  0, "Starting");
    mStateInfos[Load_InitWorld]             = LoadingStateInfo(  3, "Creating world");
    mStateInfos[Load_AsteroidModelInfos]    = LoadingStateInfo( 25, "Loading asteroid models");
    mStateInfos[Load_LevelXML]              = LoadingStateInfo( 30, "Loading level");
    mStateInfos[Load_ObjectModelInfos]      = LoadingStateInfo( 42, "Loading object models");
    mStateInfos[Load_FinalizeModels]        = LoadingStateInfo( 47, "Finalizing models");
    mStateInfos[Load_PrecompileShaders]     = LoadingStateInfo( 50, "Compiling shaders");
    mStateInfos[Load_AssignObjectModels]    = LoadingStateInfo( 55, "Creating objects");
    mStateInfos[Load_AssignAsteroidModels]  = LoadingStateInfo( 95, "Creating asteroids");
    mStateInfos[Load_Done]                  = LoadingStateInfo(100, "Starting game");
}

void LoadingView::init()
{
    mLoadingTextFont.setPointSize(14);
    mStatusTextFont.setPointSize(10);
}

void LoadingView::resize(int w, int h)
{
    mProgressRect.setRect(w/3, h/3, w/3, 20);

    mLoadingTextRect = mProgressRect;
    mLoadingTextRect.setBottom(mProgressRect.top() - 10);
    mLoadingTextRect.setTop(mProgressRect.top() - 10 - 20);

    mStatusTextRect = mProgressRect;
    mStatusTextRect.setTop(mProgressRect.bottom());
    mStatusTextRect.setBottom(mProgressRect.bottom() + 20);
}

void LoadingView::render()
{
    // Draw the progressbar
    glPushMatrix();
    setupOrthoView();

    glColor3f(0.1, 0.1, 0.4);
    glRecti(mProgressRect.left(), mProgressRect.top(), mProgressRect.right(), mProgressRect.bottom());
    glColor3f(0.0, 0.0, 0.0);
    float left = mProgressRect.left() + 1 + (mProgressRect.width() - 2) / 100.0f * mProgress;
    glRectf(left, mProgressRect.top()+1, mProgressRect.right()-1, mProgressRect.bottom()-1);

    glPopMatrix();

    // Draw the texts
    textRenderer()->begin(glWidget());
    glColor3f(0.2, 0.2, 0.7);
    // FIXME: there seems to be a text positioning bug in TextRenderer which results in text
    //  rendered a few pixels too low.
    textRenderer()->draw(mProgressRect.adjusted(0, -6, 0, -6), QString("%1 %").arg((int)mProgress), Qt::AlignCenter, mStatusTextFont);
    textRenderer()->draw(mStatusTextRect, mStatus, Qt::AlignRight | Qt::AlignTop, mStatusTextFont);
    if (mFailed) {
        glColor3f(0.8, 0.2, 0.2);
        textRenderer()->draw(mLoadingTextRect, "Loading failed!", Qt::AlignLeft | Qt::AlignBottom, mLoadingTextFont);
    } else {
        textRenderer()->draw(mLoadingTextRect, "Loading...", Qt::AlignLeft | Qt::AlignBottom, mLoadingTextFont);
    }
    textRenderer()->end();
}

void LoadingView::updateProgress(LoadingState currentState, int substate, int substateCount)
{
    kDebug() << currentState << substate << substateCount;

    LoadingState previousState = (LoadingState)((int)currentState - 1);
    mProgress = mStateInfos[previousState].progress;
    if (substateCount > 0) {
        mProgress += (mStateInfos[currentState].progress - mStateInfos[previousState].progress) * ((float)substate / substateCount);
    }
    mStatus = mStateInfos[currentState].status;

    update();
    qApp->processEvents();
}

void LoadingView::loadingAsteroid(QString name, int i, int total)
{
    updateProgress(Load_AsteroidModelInfos, i, total);
}

bool LoadingView::load()
{
    bool success = tryLoad();
    if (!success) {
        mFailed = true;
    }

    update();
    return success;
}

bool LoadingView::tryLoad()
{
    kDebug();

    connect(mWorld->data(), SIGNAL(loadingAsteroid(QString,int,int)), this, SLOT(loadingAsteroid(QString,int,int)));

    kDebug() << "Init world";
    updateProgress(Load_InitWorld);
    if (!mWorld->init()) {
        return false;
    }

    kDebug() << "Loading asteroid model infos";
    updateProgress(Load_AsteroidModelInfos);
    if (!mWorld->data()->loadAsteroidModels()) {
        return false;
    }

    kDebug() << "Loading level XML";
    updateProgress(Load_LevelXML);
    if (!loadLevel(mLevelName)) {
        return false;
    }
    kDebug() << "Found" << mModel2Objects.count() << "different models";

    updateProgress(Load_ObjectModelInfos);
    int i = 0;
    foreach (QString modelName, mModel2Objects.keys()) {
        updateProgress(Load_ObjectModelInfos, i, mModel2Objects.count());
        if (!mWorld->data()->loadModel(modelName)) {
            return false;
        }
        i++;
    }

    updateProgress(Load_FinalizeModels);
    mWorld->data()->finalizeModels();

    updateProgress(Load_PrecompileShaders);
    QStringList lightTypes;
    QSet<QString> materials;
    lightTypes << "directional" << "spot" << "point";
    materials << "asteroid";
    foreach (const QString& modelName, mModel2Objects.keys()) {
        materials << mWorld->data()->model(modelName)->material();
    }

    foreach (const QString& material, materials) {
        // For "light" renderMode, use all light types
        foreach (const QString& light, lightTypes) {
            if (!mWorld->data()->shaderManager()->precompileProgram("light", light, material)) {
                return false;
            }
        }
        // For "ambient" mode, just use "directional"
        if (!mWorld->data()->shaderManager()->precompileProgram("ambient", "directional", material)) {
            return false;
        }
    }

    updateProgress(Load_AssignObjectModels);
    // Objects
    QHashIterator<QString, QList<GameObject*> > it(mModel2Objects);
    while (it.hasNext()) {
        it.next();
        kDebug() << "Assigning model" << it.key() << "to" << it.value().count() << "objects";
        foreach (GameObject* obj, it.value()) {
            obj->setModel(mWorld->data()->model(it.key()));
        }
    }
    updateProgress(Load_AssignAsteroidModels);
    // Asteroids
    i = 0;
    foreach (GameObject* obj, *mWorld->objects()) {
        if (obj->type() == GameObject::T_Asteroid) {
            if (!obj->model()) {
                obj->setModel(mWorld->data()->randomAsteroidModel());
            }
            if (i % 30 == 0) {
                updateProgress(Load_AssignAsteroidModels, i, mWorld->objects()->count());
            }
        }
        i++;
    }

    updateProgress(Load_Done);
    return true;
}

bool LoadingView::loadLevel(const QString& levelname)
{
    QString levelfilepath(mWorld->dataPath() + "levels/" + levelname + ".level");
    kDebug() << "level file is" << levelfilepath;

    QFile file(levelfilepath);
    if (!file.open(QIODevice::ReadOnly)) {
        kError() << "Couldn't open level file" << levelfilepath;
        return false;
    }

    QDomDocument doc("level");
    if (!doc.setContent(&file)) {
        kError() << "Couldn't parse level file" << levelfilepath;
        return false;
    }
    file.close();

    QDomElement docElem = doc.documentElement();
    QDomNodeList tmp;

    if (!loadObjects(docElem.firstChildElement("Objects"))) {
        return false;
    }

    if (!loadTrack(docElem.firstChildElement("Track"))) {
        return false;
    }

    if (!loadPlayer(docElem.firstChildElement("Player"))) {
        return false;
    }

    kDebug() << "Level successfully loaded!";
    return true;
}

bool LoadingView::loadObjects(const QDomElement& objectsTag)
{
    if (objectsTag.isNull()) {
        kError() << "NULL objects tag";
        return false;
    }
    kDebug();

    // Misc objects
    QDomNodeList objectList = objectsTag.elementsByTagName("Object");
    for (int i = 0; i < objectList.count(); i++) {
        if (!loadObject(objectList.item(i).toElement())) {
            return false;
        }
    }

    // Asteroid fields
    QDomNodeList asteroidFieldList = objectsTag.elementsByTagName("AsteroidField");
    for (int i = 0; i < asteroidFieldList.count(); i++) {
        if (!loadAsteroidField(asteroidFieldList.item(i).toElement())) {
            return false;
        }
    }

    // Planets
    QDomNodeList planetList = objectsTag.elementsByTagName("Planet");
    for (int i = 0; i < planetList.count(); i++) {
        if (!loadPlanet(planetList.item(i).toElement())) {
            return false;
        }
    }

    return true;
}

bool LoadingView::loadObject(const QDomElement& elem)
{
    if (elem.isNull()) {
        kError() << "NULL element";
        return false;
    }

    QString name = elem.attribute("name");
    kDebug() << name;

    Vector3f pos;
    if (!loadVector(elem.firstChildElement("Position"), &pos)) {
        return false;
    }

    GameObject* obj = mWorld->createObject(name);
    if (!obj) {
        return false;
    }
    obj->setPosition(pos);
    mModel2Objects[name].append(obj);

    return true;
}

bool LoadingView::loadAsteroidField(const QDomElement& elem)
{
    if (elem.isNull()) {
        kError() << "NULL element";
        return false;
    }
    kDebug();

    float density = float(elem.attribute("density", "1").toDouble());

    QDomElement radiusElement = elem.firstChildElement("Radius");
    if (radiusElement.isNull()) {
        kError() << "NULL Radius element";
        return false;
    }
    float avg = float(radiusElement.attribute("avg", "10").toDouble());
    float variance = float(radiusElement.attribute("variance", "10").toDouble());
    float min = float(radiusElement.attribute("min", "10").toDouble());
    float max = float(radiusElement.attribute("max", "10").toDouble());

    QDomElement bboxElement = elem.firstChildElement("BBox");
    if (bboxElement.isNull()) {
        kError() << "NULL BBox element";
        return false;
    }
    Vector3f bboxmin, bboxmax;
    bboxmin.x() = float(bboxElement.attribute("x1", "0").toDouble());
    bboxmin.y() = float(bboxElement.attribute("y1", "0").toDouble());
    bboxmin.z() = float(bboxElement.attribute("z1", "0").toDouble());
    bboxmax.x() = float(bboxElement.attribute("x2", "0").toDouble());
    bboxmax.y() = float(bboxElement.attribute("y2", "0").toDouble());
    bboxmax.z() = float(bboxElement.attribute("z2", "0").toDouble());

    Vector3f bboxsize = bboxmax - bboxmin;
    // Volume of the bounding box of the asteroid field
    float bboxvolume = bboxsize.x() * bboxsize.y() * bboxsize.z();
    // Average volume of a single asteroid
    float asteroidvolume = M_PI * (sq(avg) + sq(qMin(max, avg+variance)) + sq(qMax(min, avg-variance))) / 3;
//     int count = int(bboxvolume * density / 1000);
    int count = int(bboxvolume/asteroidvolume * density/100);
    kDebug() << "Creating" << count << "asteroids, density:" << density << ", bvol:" << bboxvolume <<
            ", avol:" << asteroidvolume;

    mWorld->createAsteroids(count, avg, variance, min, max, bboxmin, bboxmax);

    return true;
}

bool LoadingView::loadVector(const QDomElement& elem, Eigen::Vector3f* pos)
{
    if (elem.isNull()) {
        kError() << "NULL element";
        return false;
    }

    pos->x() = float(elem.attribute("x", "0").toDouble());
    pos->y() = float(elem.attribute("y", "0").toDouble());
    pos->z() = float(elem.attribute("z", "0").toDouble());

    return true;
}

bool LoadingView::loadTrack(const QDomElement& elem)
{
    if (elem.isNull()) {
        kError() << "NULL element";
        return false;
    }

    QDomNodeList cpList = elem.elementsByTagName("Checkpoint");
    kDebug() << "Loading track of" << cpList.count() << "checkpoints";
    for (int i = 0; i < cpList.count(); i++) {
        QDomElement cpElem = cpList.item(i).toElement();
        if (cpElem.isNull()) {
            kError() << "NULL element for checkpoint" << i;
            return false;
        }

        Vector3f pos;
        if (!loadVector(cpElem.firstChildElement("Position"), &pos)) {
            return false;
        }

        GameObject* obj = mWorld->addTrackPoint(pos);
        mModel2Objects["checkpoint"].append(obj);

        // Set checkpoint's direction if it's given in the level file
        QDomElement directionElem = cpElem.firstChildElement("Direction");
        if (!directionElem.isNull()) {
            Vector3f dir;
            if (!loadVector(directionElem, &dir)) {
                return false;
            }
            obj->setDirection(dir);
        }
    }

    return true;
}

bool LoadingView::loadPlayer(const QDomElement& elem)
{
    if (elem.isNull()) {
        kError() << "NULL element";
        return false;
    }
    kDebug();

    Vector3f pos;
    if (!loadVector(elem.firstChildElement("Position"), &pos)) {
        return false;
    }
    Vector3f dir;
    if (!loadVector(elem.firstChildElement("Direction"), &dir)) {
        return false;
    }

    Ship* ship = new Ship();
    ship->setPosition(pos);
    ship->setDirection(dir);
    mWorld->setShip(ship);
    mModel2Objects["ship"].append(ship);

    return true;
}

bool LoadingView::loadPlanet(const QDomElement& elem)
{
    if (elem.isNull()) {
        kError() << "NULL element";
        return false;
    }

    Vector3f pos;
    if (!loadVector(elem.firstChildElement("Position"), &pos)) {
        return false;
    }

    QString type = elem.attribute("type");
    int radius = elem.attribute("radius").toInt();

    Planet* planet = mWorld->addPlanet(type);
    if (!planet) {
        return false;
    }
    planet->setPosition(pos);
    planet->setRadius(radius);

    return true;
}
