/*
 * Copyright 2008-2009  Rivo Laks <rivolaks@hot.ee>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mainwindow.h"

#include "ship.h"
#include "asteroid.h"
#include "gameworld.h"
#include "bullet.h"
#include "gameview.h"
#include "displaywidget.h"
#include "gameguiview.h"
#include "loadingview.h"
#include "datastore.h"
#include "menuview.h"

#include <kgllib/widgetproxy.h>

#include <QTimer>
#include <QApplication>
#include <QWheelEvent>

#include <KGlobal>
#include <KStandardDirs>
#include <KDebug>
#include <KAction>
#include <KActionCollection>
#include <KStandardGameAction>
#include <KToggleAction>
#include <KMessageBox>


MainWindow::MainWindow() : KXmlGuiWindow()
{
    mGameRunning = false;
    mWorld = 0;
    mView = 0;

    // Find where our data is
    QString datadir = KGlobal::dirs()->findResourceDir("appdata", "shaders/render-light.frag");
    if (datadir.isEmpty()) {
        KMessageBox::error(0, "Data directory could not be found!\nMake sure that you have properly installed AstroDodge.",
                           "Data not found");
        qApp->quit();
    }
    DataStore::initInstance(datadir);

    mDisplay = new DisplayWidget(this);

    setCentralWidget(mDisplay->proxy());

    setMinimumSize(640, 480);

    setupActions();
    setupGUI();

    mMenuView = new MenuView(mDisplay);
    connect(mMenuView, SIGNAL(startNewGame()), this, SLOT(startNewGame()));
    connect(mMenuView, SIGNAL(quitGame()), this, SLOT(close()));

    mDisplay->appendView(mMenuView);
}

MainWindow::~MainWindow()
{
}

void MainWindow::setupActions()
{
    // Dodge actions
    KAction* dodgeLeftAction = new KAction("Dodge left", this);
    dodgeLeftAction->setShortcut(Qt::Key_A);
    actionCollection()->addAction("dodge_left", dodgeLeftAction);
    connect(dodgeLeftAction, SIGNAL(triggered(bool)), this, SLOT(dodgeLeft()));

    KAction* dodgeRightAction = new KAction("Dodge right", this);
    dodgeRightAction->setShortcut(Qt::Key_D);
    actionCollection()->addAction("dodge_right", dodgeRightAction);
    connect(dodgeRightAction, SIGNAL(triggered(bool)), this, SLOT(dodgeRight()));

    KAction* dodgeUpAction = new KAction("Dodge up", this);
    dodgeUpAction->setShortcut(Qt::Key_W);
    actionCollection()->addAction("dodge_up", dodgeUpAction);
    connect(dodgeUpAction, SIGNAL(triggered(bool)), this, SLOT(dodgeUp()));

    KAction* dodgeDownAction = new KAction("Dodge down", this);
    dodgeDownAction->setShortcut(Qt::Key_S);
    actionCollection()->addAction("dodge_down", dodgeDownAction);
    connect(dodgeDownAction, SIGNAL(triggered(bool)), this, SLOT(dodgeDown()));

    KAction* fireAction = new KAction("Fire cannon", this);
    fireAction->setShortcut(Qt::Key_Space);
    actionCollection()->addAction("fire_cannon", fireAction);
    connect(fireAction, SIGNAL(triggered(bool)), this, SLOT(fireCannon(bool)));

    // Game state actions
    KStandardGameAction::gameNew(this, SLOT(startNewGame()), actionCollection());
    KStandardGameAction::pause(this, SLOT(setGamePaused(bool)), actionCollection());
    KStandardGameAction::quit(this, SLOT(close()), actionCollection());

    // Graphics actions
    KToggleAction* enableHDRAction = new KToggleAction("Use HDR rendering", this);
    enableHDRAction->setChecked(false);
    enableHDRAction->setShortcut(Qt::CTRL + Qt::Key_H);
    actionCollection()->addAction("hdr_enable", enableHDRAction);
    connect(enableHDRAction, SIGNAL(triggered(bool)), mDisplay, SLOT(setHdrRenderingActive(bool)));

    KAction* showHDRConfigAction = new KAction("Show HDR config", this);
    actionCollection()->addAction("hdr_config", showHDRConfigAction);
    connect(showHDRConfigAction, SIGNAL(triggered(bool)), mDisplay, SLOT(showHdrConfig()));

    mShowRenderInfoAction = new KToggleAction("Show renderinfo", this);
    mShowRenderInfoAction->setChecked(false);
    mShowRenderInfoAction->setShortcut(Qt::CTRL + Qt::Key_I);
    actionCollection()->addAction("show_renderinfo", mShowRenderInfoAction);

    mDetailIncreaseAction = new KAction("Increase detail", this);
    mDetailIncreaseAction->setShortcut(Qt::CTRL + Qt::Key_Plus);
    actionCollection()->addAction("detail_increase", mDetailIncreaseAction);
    mDetailDecreaseAction = new KAction("Decrease detail", this);
    mDetailDecreaseAction->setShortcut(Qt::CTRL + Qt::Key_Minus);
    actionCollection()->addAction("detail_decrease", mDetailDecreaseAction);

    mShipLightAction = new KToggleAction("Ship's light", this);
    mShipLightAction->setChecked(false);
    mShipLightAction->setShortcut(Qt::Key_L);
    actionCollection()->addAction("ship_light", mShipLightAction);
}

void MainWindow::startNewGame()
{
    kDebug();
    mMenuView->hide();

    mWorld = new GameWorld(this);
    mShipLightAction->setChecked(false);
    connect(mShipLightAction, SIGNAL(triggered(bool)), mWorld, SLOT(setShipLight(bool)));

    mLoadingView = new LoadingView(mDisplay, mWorld, "test1");
    connect(mLoadingView, SIGNAL(loadingComplete()), this, SLOT(gameAboutToStart()));

    mDisplay->appendView(mLoadingView);

    if (mLoadingView->load()) {
        gameAboutToStart();
    }
}

void MainWindow::gameAboutToStart()
{
    kDebug();

    connect(mDisplay, SIGNAL(signalRotateShip(int,int)), mWorld, SLOT(slotRotateShip(int,int)));

    mView = new GameView(mDisplay, mWorld);
    connect(mDetailIncreaseAction, SIGNAL(triggered(bool)), mView, SLOT(increaseDetailLevel()));
    connect(mDetailDecreaseAction, SIGNAL(triggered(bool)), mView, SLOT(decreaseDetailLevel()));
    mDisplay->appendView(mView);

    GameGUIView* guiView = new GameGUIView(mDisplay, mWorld);
    connect(mShowRenderInfoAction, SIGNAL(triggered(bool)), guiView, SLOT(setShowRenderInfo(bool)));
    guiView->setShowRenderInfo(mShowRenderInfoAction->isChecked());
    mDisplay->appendView(guiView);

    mDisplay->removeView(mLoadingView);

//     delete mLoadingView;

    mGameRunning = true;
}

void MainWindow::wheelEvent(QWheelEvent* e)
{
    if (!mGameRunning) {
        e->ignore();
        return;
    }

    mWorld->ship()->thrust += (e->delta() / 120.0f * 2.0f);
    mWorld->ship()->thrust = qBound(0.0f, mWorld->ship()->thrust, 20.0f);
}

void MainWindow::dodge(int x, int y)
{
    if (mGameRunning) {
        mWorld->ship()->dodge(x, y);
    }
}

void MainWindow::fireCannon(bool fire)
{
    kDebug() << fire;
    if (mGameRunning) {
        mWorld->slotShootBullet();
    }
}

void MainWindow::setGamePaused(bool paused)
{
    if (!mWorld) {
        return;
    }
    kDebug() << paused;
    if (paused == mWorld->gamePaused()) {
        return;
    }

    mWorld->slotToggleGamePaused();
    mGameRunning = !paused;
}
