/***************************************************************************
 *   Copyright (C) 2009 by Alexander Smirnov                               *
 *   alexander.v.smirnov@gmail.com                                         *
 *                                                                         *
 *   Copyright (C) 2010, Dmitry Chernov                                    *
 *   diman4ik.chernov@gmail.com                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.         *
 ***************************************************************************/
#ifndef BACKGAMMONGAME_H
#define BACKGAMMONGAME_H

#include <kgame/kgame.h>

#include "board.h"
#include "move.h"
#include "checker.h"
#include "games/igameengine.h"
#include "iplayer.h"

class IPlayer;

class BackgammonScene;


/**
	@author Alexander Smirnov <alexander.v.smirnov@gmail.com>
*/
//TODO: extract base class for all kinds of games
class BackgammonGame : public IGameEngine
{
    Q_OBJECT
public:
    typedef QPair<int, int > Rolls;
    
    BackgammonGame( QObject* parent = 0 );
    ~BackgammonGame();

    int cubeValue() const   {   return m_doublingCube; }
    int matchLength() const {   return m_matchLength;  }

    int pipCount(const int playerIdx) const; 

    //--- derived from IGameEngine
    virtual bool canPlaceChecker(const Checker::Color color, qint8 cell) const;
    virtual bool canMoveChecker(IPlayer* player, qint8 cell1, qint8 cell2) const;
    virtual QStringList initialBoardState() const;

    void updatePip();
    void processUserAction(qint8 cell);
    
    virtual void roll();

signals:
    /**
     *  Show changes on board
     *
     */
    void showMove( qint8 src, qint8 dst );
    
    /**
     * Added to history widget
     *
     */
    void gameEvent( const QString& info );
    
    /**
     * Message to player from game engine
     *
     */    
    void engineMessage( const QString& info );

protected:
    virtual bool playerInput(QDataStream &msg, KPlayer *player);
    
    bool createMove( const IPlayer& player );
    const Move makeMove( const IPlayer& player, QVector<Move>& changes );
    
    IPlayer * m_players[2];
    
    void setDice( Rolls rolls );
    
    /** 
     * Switch to next game state
     */
    void switchToNextState();
    void switchToState( EGameState state );
    EGameState getState() const;
    
    /** 
     * Check if one of the players won
     */
    bool checkGameOver();
    
    /** 
     * Check if given player can move checker from src position to dst position
     */
    bool canMakeMove( const IPlayer& player, const Move& move );
    
    /** 
     * Sends changes in game field to board for vizualization
     */
    void showBoardChanges( const IPlayer& player, const QVector<Move>& changes );
    
    static const int bar_ind = 26;
    static const int thrown_ind = 24;
    
    /** 
     * Debug helping routine
     */
    void showField();
    
    bool canUseFirstDie( const IPlayer& player );
    bool canUseSecondDie( const IPlayer& player );
    
    void setCurrentMoveSrc( int cell );
    void setCurrentMoveDst( int cell );
    
    bool moveReady( const IPlayer& player );
    
    /** 
     * Check if given player has any moves
     */
    bool gotMoves( const IPlayer& player );    
    
    bool gotCurrentMoveSrc();
    bool gotCurrentMoveDst();
    
    int getBarIndex( const IPlayer& player );
    int getOutIndex( const IPlayer& player );
    
    bool cellMatchesPlayer( const IPlayer& player, int src );
    
    static const int MAX_IN_ROW = 12;
    static const int FIELD_SIZE = 28;
    
    void setBoard( const qint8 ( & field ) [FIELD_SIZE] );
    void resetBoard();
    
    int getFirstDie();
    int getSecondDie();
    
    void setFirstDieUsed();
    void setSecondDieUsed();
    
    bool firstDieUsed();
    bool secondDieUsed();
    
    bool canThrow( const IPlayer& player );
    
    void showProposedMoves( int src, const IPlayer& player );
    void unshowProposedMoves();
    
private:    

    /** 
     * Check if given player can throw checkers from home
     */
    bool canThrow( IPlayer::direction dir );
    
    /** 
     * Cheks if given player has checkers at bar
     */
    bool checkCheckersAtBar( const IPlayer& player );

    /** 
     * Helper functions for field navigation and modification
     */
    int getBarIndex( IPlayer::direction dir );
    int getThrownIndex( IPlayer::direction dir );
    int getHomeIndex( IPlayer::direction dir );
    int getIndexFromBar( IPlayer::direction dir, int die );
    int getIndexFromSrc( int src, IPlayer::direction dir, int die );
    
    /** 
     * Helper functions to modify field
     */
    void addToBar( IPlayer::direction dir );
    void throwChecker( int ind, IPlayer::direction dir );
    void addChecker( int ind, int direction );
    void replaceChecker( int ind, int direction );
    
    
    bool canUseDie( const IPlayer& player, int die );
    
    /** 
     * The length of the match
     */
    int m_matchLength;
    /**
     * Player1 match score so far
     */
    int m_player1Score;
    /**
     * Player2 match score so far
     */
    int m_player2Score;
    /**
     * Current value of doubling cube
     */
    int m_doublingCube;

    
    Move m_currentMove;
    
    bool cellValueMatchesDirection( int cell, int dir );
    
    int getDieToBeUsed( const IPlayer& player );

    EGameState m_currentAction;

    qint8 m_dice1, m_dice2;
    
    bool m_dice1Used;
    bool m_dice2Used;
    
    /**
     * Game field
     */
    static const int TOTAL_CHECKERS = 15;
    static const int HOME_SIZE = 6;

    qint8 m_field[FIELD_SIZE];
    
    static const qint8 m_initCheckerPos[];
    
    void resetCurrentMove();
    void resetDice();
    void resetField();
    
    bool diceAreEqual();
    
    bool moveMatchesDirection( const Move& move, IPlayer::direction dir );
    
    bool cellValuesMatch( int src, int dst );
    
    void setDieUsed( int die );
    
    bool canThrowFromSrc( int src, IPlayer::direction dir, int die );
    
    int getLengthToThrown( int src, IPlayer::direction dir );
    
    bool cellMatchesPlayer( IPlayer::direction dir, int src );
    
    QList<int> getProposedMoves( int cell, const IPlayer& player );
};

#endif
