/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "ConfigWidget.h"

// FruitWars headers
#include "Globals.h"
#include "Main.h"

// KDE headers
#include <KConfigGroup>

ConfigWidget::ConfigWidget()
    : QWidget()
    , presetGames("fruitwarsgametypes.rc", KConfig::NoGlobals)
{
    setupUi(this);

    connect(startGameButton, SIGNAL(clicked()), SLOT(startButtonClicked()));

    // setting up the teams stuff
    tabWidget->setTabEnabled(2, false);
    connect(noTeams, SIGNAL(toggled(bool)), SLOT(teamsToggled(bool)));
    team3Group->setVisible(false);
    team4Group->setVisible(false);
    connect(fourTeams, SIGNAL(toggled(bool)), SLOT(fourTeamsToggled(bool)));

    // load default settings
    loadConfiguration(KGlobal::config()->group("DefaultSettings"));

    // other enable/disable things
    connect(hitEffectCombo, SIGNAL(currentIndexChanged(int)), SLOT(hitEffectChanged(int)));

    // preset gametypes
    foreach (const QString& group, presetGames.groupList()) {
        gameTypesCombo->addItem(group);
    }
    connect(gameTypesCombo, SIGNAL(currentIndexChanged(int)), SLOT(gameTypeChanged(int)));
}

void ConfigWidget::gameTypeChanged(int idx)
{
    QString gameType = gameTypesCombo->itemText(idx);
    loadConfiguration(presetGames.group(gameType));
}

void ConfigWidget::startButtonClicked()
{
    applyConfiguration();
    settingsTab->setEnabled(false);
    startGameButton->setEnabled(false);
    statusMessage->setText(i18n("The game has been started!"));

    MainWidget* win = new MainWidget(this);
    win->grabMouse();
    win->show();
    QCursor::setPos(win->mapToGlobal(QPoint(win->width()/2, win->height()/2)));
}

void ConfigWidget::teamsToggled(bool enabled)
{
    tabWidget->setTabEnabled(2, !enabled);
}

void ConfigWidget::fourTeamsToggled(bool enabled)
{
    team3Group->setVisible(enabled);
    team4Group->setVisible(enabled);
}

void ConfigWidget::hitEffectChanged(int idx)
{
    stunTimeEdit->setEnabled(idx == HE_Stun);
}

void ConfigWidget::applyConfiguration()
{
    // Game Settings
    ::hitEffect = hitEffectCombo->currentIndex();
    ::playerStunTime = stunTimeEdit->value();
    ::friendlyFire = friendlyFireCombo->currentIndex() == 0;
    ::extraHitPoints = hitPointsEdit->value();
    ::diversityBonus = diversityBonusEdit->value();
    ::maxTime = maxTimeEdit->value() * 60000;
    ::maxPoints = maxPointsEdit->value();
    if (gameTypeCombo->currentIndex() == GT_Points) {
        ::pointgame = true;
        ::timegame = false;
    } else if (gameTypeCombo->currentIndex() == GT_Time) {
        ::pointgame = false;
        ::timegame = true;
    } else {
        ::pointgame = ::timegame = true;
    }
    ::spawnRandom = spawnTypeCombo->currentIndex() == ST_Random;
    ::throwInfluence = throwEffectEdit->value() * 0.01;
    ::holdMode = holdModeCombo->currentIndex();
    ::drunkMode = fuzzyEndingCheck->isChecked();

    // Fruit Settings
    ::fruitAppleScore = applePoints->value();
    ::fruitLemonScore = lemonPoints->value();
    ::fruitOrangeScore = orangePoints->value();
    ::fruitAnanasScore = pineapplePoints->value();
    ::fruitStrawScore = strawberryPoints->value();

    ::minFruit = minFruitEdit->value();
    ::maxFruit = maxFruitEdit->value();
    ::minWait = minWaitEdit->value();
    ::maxWait = maxWaitEdit->value();
    ::spawnRadius = radiusEdit->value();

    ::oneFruitType = oneFruitTypeCheck->isChecked();
    switch (fruitTypeList->currentIndex()) {
        case FT_Apple:      ::theFruitType = ftApple; break;
        case FT_Lemon:      ::theFruitType = ftLemon; break;
        case FT_Orange:     ::theFruitType = ftOrange; break;
        case FT_Pineapple:  ::theFruitType = ftAnanas; break;
        case FT_Strawberry: ::theFruitType = ftStraw; break;
    }

    // Teams
    ::teamMode = 2;
    if (twoTeams->isChecked()) {
        ::teamMode = 0;
    } else if (fourTeams->isChecked()) {
        ::teamMode = 1;
    }
}

void ConfigWidget::loadConfiguration(const KConfigGroup& config)
{
    // Game Settings
    int hitEffect = config.readEntry("hit_effect", (int) HE_Stun);
    hitEffectCombo->setCurrentIndex(hitEffect);

    int stunTime = config.readEntry("stun_time", 3);
    stunTimeEdit->setValue(stunTime);

    bool friendlyFire = config.readEntry("friendly_fire", true);
    friendlyFireCombo->setCurrentIndex(friendlyFire ? 0 : 1);

    int hitPoints = config.readEntry("hit_points", 0);
    hitPointsEdit->setValue(hitPoints);

    int divBonus = config.readEntry("diversity_bonus", 5);
    diversityBonusEdit->setValue(divBonus);

    int maxTime = config.readEntry("max_time", 15);
    maxTimeEdit->setValue(maxTime);

    int maxPoints = config.readEntry("max_points", 50);
    maxPointsEdit->setValue(maxPoints);

    int gameType = config.readEntry("game_type", (int) GT_Points);
    gameTypeCombo->setCurrentIndex(gameType);

    int spawnType = config.readEntry("spawn_type", (int) ST_Basket);
    spawnTypeCombo->setCurrentIndex(spawnType);

    int throwEffect = config.readEntry("throw_effect", 0);
    throwEffectEdit->setValue(throwEffect);

    int holdMode = config.readEntry("hold_mode", (int) HM_Hold);
    holdModeCombo->setCurrentIndex(holdMode);

    bool fuzzyEnding = config.readEntry("fuzzy_ending", false);
    fuzzyEndingCheck->setChecked(fuzzyEnding);

    // Fruit Settings
    applePoints->setValue(config.readEntry("apple_points", 4));
    lemonPoints->setValue(config.readEntry("lemon_points", 1));
    orangePoints->setValue(config.readEntry("orange_points", 2));
    pineapplePoints->setValue(config.readEntry("pineapple_points", 3));
    strawberryPoints->setValue(config.readEntry("strawberry_points", 5));

    minFruitEdit->setValue(config.readEntry("min_fruit", 4));
    maxFruitEdit->setValue(config.readEntry("max_fruit", 8));
    minWaitEdit->setValue(config.readEntry("min_wait", 5));
    maxWaitEdit->setValue(config.readEntry("max_wait", 30));
    radiusEdit->setValue(config.readEntry("spawn_radius", 60));

    oneFruitTypeCheck->setChecked(config.readEntry("one_fruit_type", false));
    fruitTypeList->setCurrentIndex(config.readEntry("fruit_type", (int) FT_Strawberry));

    // Team Settings
    int teamCount = config.readEntry("teams", 0);
    switch (teamCount) {
        case 2:  twoTeams->setChecked(true); break;
        case 4:  fourTeams->setChecked(true); break;
        default: noTeams->setChecked(true); break;
    }
}
