/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "Instance.h"

// FruitWars headers
#include "Exceptions.h"
#include "Globals.h"
#include "Math3D.h"

#include <Eigen/LU>

USING_PART_OF_NAMESPACE_EIGEN
using Eigen::AngleAxisf;
using Eigen::Quaternionf;
using Eigen::Scaling3f;
using Eigen::Transform3f;
using Eigen::Translation3f;

Frustum instFrust;

Instance* Instance::load(Model3D *model)
{
    if (model->getHeader().ogltype == "STATIC") {
        return new StaticInstance((StaticModel*) model);
    } else if (model->getHeader().ogltype == "FRAME") {
        return new FrameInstance((FrameModel*) model);
    } else if (model->getHeader().ogltype == "ANIM") {
        return new AnimInstance((AnimModel*) model);
    } else if (model->getHeader().ogltype == "BONE") {
        return new BoneInstance((BoneModel*) model);
    } else {
        return NULL;
    }
}

Instance::Instance()
    : pos(0.0, 0.0, 0.0), rot(0.0, 0.0, 0.0), scale(1.0, 1.0, 1.0), bs(pos, 1000000.0)
{
    attach.attached = false;
}

Instance::~Instance()
{
}

void Instance::render()
{
}

void Instance::setFrame(unsigned int, float)
{
}

void Instance::setAnimSetBlend(unsigned int, float, float, int)
{
}

void Instance::setAnimBlend(unsigned int, unsigned int, float)
{
}

void Instance::calcAnim()
{
}

void Instance::setTexture(unsigned int mat, int tex)
{
    if ((unsigned int)texture.size() > mat) {
        texture[mat] = tex;
    }
}

void Instance::position(const Vector3f &pos)
{
    this->pos = pos;
}

void Instance::scaleObj(const Vector3f &scale)
{
    this->scale = scale;
}

void Instance::rotateX(float xAng)
{
    rot.x() = xAng;
}

void Instance::rotateY(float yAng)
{
    rot.y() = yAng;
}

void Instance::rotateZ(float zAng)
{
    rot.z() = zAng;
}

void Instance::rotate(const Vector3f &rot)
{
    this->rot = rot;
}

void Instance::setBS(const BS &bs)
{
    this->bs = bs;
}

void Instance::attachToFrame(BoneInstance *inst, int frame)
{
    if (!attach.attached) {
        attach.attached = true;
        attach.inst = inst;
        attach.frame = frame;
        attach.offset = /*fastMul(
                    fastMul(*/
                    Transform3f(attach.inst->transform() *
                    attach.inst->frame[frame].transform.inverse()) *
                    transform();
        attach.inst->transform() * transform();
        //attach.inst->frame[frame].transform.inverse() * transform();
        attach.inst->transform() * attach.inst->frame[frame].transform.inverse();
    }
}

void Instance::specialAttach(BoneInstance *inst, int frame, const Transform3f &offset)
{
    if (!attach.attached) {
        attach.attached = true;
        attach.inst = inst;
        attach.frame = frame;
        attach.offset = offset;
    }
}

void Instance::detach()
{
    if (attach.attached) {
        attach.attached = false;
        Transform3f t(Transform3f(attach.inst->transform() * attach.inst->frame[attach.frame].transform) * attach.offset);
//            .extract(pos, scale, rot); fastMul
        pos = t.translation();
        extractScaleRotation(t, scale, rot);
    }
}

Transform3f Instance::transform()
{
    return
        Translation3f(pos) *
        Transform3f(
            AngleAxisf(-rot.y(), Vector3f::UnitY()) *
            AngleAxisf(-rot.x(), Vector3f::UnitX()) *
            AngleAxisf(-rot.z(), Vector3f::UnitZ())) *
        Scaling3f(scale);
    /*fastMul(
            fastMul(
                fastMul(
                    fastMul(
                        Transform3f::fromPos(pos),
                        Transform3f::fromYAng(rot.y())
                        ),
                    Transform3f::fromXAng(rot.x())
                    ),
                Transform3f::fromZAng(rot.z())
                ),
            Transform3f::fromScale(scale)
            );*/
}

Transform3f Instance::attachTransform()
{
    return /*fastMul(
            fastMul(*/
            Transform3f(Transform3f(attach.inst->transform() *
            attach.inst->frame[attach.frame].transform) *
            attach.offset);
}

int Instance::getTex(unsigned int n)
{
    if ((unsigned int)texture.size() > n) {
        return texture[n];
    } else {
        return 0;
    }
}

Vector3f Instance::getPos()
{
    return pos;
}

Vector3f Instance::getRot()
{
    return rot;
}

Vector3f Instance::getScale()
{
    return scale;
}

BoneAttach Instance::getAttach()
{
    return attach;
}

BS Instance::getBS()
{
    return bs;
}

StaticInstance::StaticInstance(StaticModel *model)
    : model(model)
{
    texture.resize(model->materials.size());
    for (int i = 0; i < model->materials.size(); i++) {
        if (!model->materials[i].textured) {
            texture[i] = -1;
        } else {
            texture[i] = model->materials[i].textureID;
        }
    }
}

void StaticInstance::render()
{
    glPushMatrix();

    // Calc Matrix
    Transform3f theMat;
    if (!attach.attached) {
        theMat = transform();
    } else {
        theMat = attach.inst->transform();
        //theMat = fastMul(theMat, attach.inst->frame[attach.frame].transform);
        theMat = theMat * attach.inst->frame[attach.frame].transform;
        //theMat = fastMul(theMat, attach.offset);
        theMat = theMat * attach.offset;
    }

    // Setup BS
    BS theBS(theMat * bs.pos, bs.radius * qMax(qMax(scale.x(), scale.y()), scale.z()));
    if (bsInFrustum(instFrust, theBS)) {
        glMultMatrixf(theMat.data());

        model->setTextures(texture);
        if (!nprMode) {
            model->render();
        } else {
            Vector3f lvec(lightDir[0], lightDir[1], lightDir[2]);
            Transform3f matInv(extractRotation(theMat).inverse());
            model->renderCartoony( (matInv * lvec).normalized() );
        }
    }

    glPopMatrix();
}

void StaticInstance::setFrame(unsigned int, float)
{
}

void StaticInstance::setAnimSetBlend(unsigned int, float, float, int)
{
}

void StaticInstance::setAnimBlend(unsigned int, unsigned int, float)
{
}

void StaticInstance::calcAnim()
{
}

StaticModel* StaticInstance::getModel()
{
    return model;
}

FrameInstance::FrameInstance(FrameModel *model)
    : model(model)
{
    texture.resize(model->materials.size());
    for (int i = 0; i < model->materials.size(); i++) {
        if (!model->materials[i].textured) {
            texture[i] = -1;
        } else {
            texture[i] = model->materials[i].textureID;
        }
    }
}

void FrameInstance::render()
{
    glPushMatrix();

    // Calc Matrix
    Transform3f theMat = transform();

    // Setup BS
    BS theBS(theMat * bs.pos, bs.radius * qMax(qMax(scale.x(), scale.y()), scale.z()));
    if (bsInFrustum(instFrust, theBS)) {
        glMultMatrixf(theMat.data());
        model->setTextures(texture);
        model->render();
    }

    glPopMatrix();
}

void FrameInstance::setFrame(unsigned int, float)
{
}

void FrameInstance::setAnimSetBlend(unsigned int, float, float, int)
{
}

void FrameInstance::setAnimBlend(unsigned int, unsigned int, float)
{
}

void FrameInstance::calcAnim()
{
}

FrameModel* FrameInstance::getModel()
{
    return model;
}


AnimInstance::AnimInstance(AnimModel *model)
    : model(model)
{
    texture.resize(model->materials.size());
    for (int i = 0; i < model->materials.size(); i++) {
        if (!model->materials[i].textured) {
            texture[i] = -1;
        } else {
            texture[i] = model->materials[i].textureID;
        }
    }

    frame.resize(model->frames.size());
    for (int i = 0; i < frame.size(); i++) {
        frame[i].blendMult = model->frames[i].blendMult;
        frame[i].blendSum  = model->frames[i].blendSum;
        frame[i].priority  = model->frames[i].priority;
        frame[i].animPos   = model->frames[i].animPos;
        frame[i].animScale = model->frames[i].animScale;
        frame[i].animRot   = model->frames[i].animRot;
        frame[i].transform = model->frames[i].transform;
    }
}

void AnimInstance::render()
{
    glPushMatrix();

    // Calc Matrix
    Transform3f theMat = transform();

    // Setup BS
    BS theBS(theMat * bs.pos, bs.radius * qMax(qMax(scale.x(), scale.y()), scale.z()));
    if (bsInFrustum(instFrust, theBS)) {
        glMultMatrixf(theMat.data());
        model->setTextures(texture);

        for (int i = 0; i < frame.size(); i++) {
            model->frames[i].blendMult = frame[i].blendMult;
            model->frames[i].blendSum  = frame[i].blendSum;
            model->frames[i].priority  = frame[i].priority;
            model->frames[i].animPos   = frame[i].animPos;
            model->frames[i].animScale = frame[i].animScale;
            model->frames[i].animRot   = frame[i].animRot;
            model->frames[i].transform = frame[i].transform;
        }

        model->render();
    }

    glPopMatrix();
}

void AnimInstance::setFrame(unsigned int set, float time)
{
    if (set >= (unsigned int)model->animSets.size()) {
        return;
    }

    for (int i = 0; i < model->animSets[set].anims.size(); i++) {
        Anim *a = &model->animSets[set].anims[i];
        int f = a->frame;
        frame[f].transform.setIdentity();
        frame[f].animPos = Vector3f(0.0, 0.0, 0.0);
        frame[f].animScale = Vector3f(1.0, 1.0, 1.0);
        frame[f].animRot = Quaternionf(1.0, 0.0, 0.0, 0.0);
        for (int j = 0; j < a->keySets.size(); j++) {
            KeySet *ks = &a->keySets[j];
            float dTime;
            if (ks->keys.size() > 1) {
                dTime = ks->keys[ks->keys.size() - 1].time - ks->keys[ks->keys.size() - 2].time;
            } else {
                dTime = 0;
            }
            if ((time >= ks->keys[0].time) && (time < ks->keys[ks->keys.size() - 1].time + dTime)) {
                int x = 0;
                int y = ks->keys.size();
                if (y == 0) {
                    break;
                }
                while (x+1 != y) {
                    int h = (x + y) / 2;
                    if (time >= ks->keys[h].time) {
                        x = h;
                    } else {
                        y = h;
                    }
                }

                float part;
                Key *yKey;
                if (y == ks->keys.size()) {
                    part = (time - ks->keys[x].time) / dTime;
                    yKey = &ks->keys[0];
                } else {
                    part = (time - ks->keys[x].time) / (ks->keys[y].time - ks->keys[x].time);
                    yKey = &ks->keys[y];
                }

                switch (ks->animType) {
                    case 0:
                        frame[f].animRot = ks->keys[x].rot.slerp(part, yKey->rot);
                        break;
                    case 1:
                        frame[f].animScale = lerp(ks->keys[x].scale, yKey->scale, part);
                        break;
                    case 2:
                        frame[f].animPos = lerp(ks->keys[x].pos, yKey->pos, part);
                        break;
                    default:
                        frame[f].animRot = ks->keys[x].rot.slerp(part, yKey->rot);
                        frame[f].animScale = lerp(ks->keys[x].scale, yKey->scale, part);
                        frame[f].animPos = lerp(ks->keys[x].pos, yKey->pos, part);
                        break;
                }
            }
        }
    }

    for (int f = 0; f < frame.size(); f++) {
        frame[f].transform.fromPositionOrientationScale(frame[f].animPos, frame[f].animRot, frame[f].animScale);
    }
}

void AnimInstance::setAnimSetBlend(unsigned int set, float time, float blend, int priority)
{
    if (set >= (unsigned int)model->animSets.size()) {
        return;
    }
    model->animSets[set].time = time;
    model->animSets[set].blend = blend;
    model->animSets[set].priority = priority;
}

void AnimInstance::setAnimBlend(unsigned int set, unsigned int anim, float blend)
{
    if (set >= (unsigned int)model->animSets.size()) {
        return;
    }
    if (anim >= (unsigned int)model->animSets[set].anims.size()) {
        return;
    }
    model->animSets[set].anims[anim].blend = blend;
}

void AnimInstance::calcAnim()
{
    // PreCalcs
    for (int f = 0; f < frame.size(); f++) {
        frame[f].blendSum = 0.00;
        frame[f].priority = 0;
    }

    for (int i = 0; i < model->animSets.size(); i++) {
        for (int j = 0; j < model->animSets[i].anims.size(); i++) {
            int fid = model->animSets[i].anims[j].frame;
            if (frame[fid].priority <= model->animSets[i].priority) {
                if (frame[fid].priority < model->animSets[i].priority) {
                    frame[fid].priority = model->animSets[i].priority;
                    frame[fid].blendSum = 0.00;
                }
                frame[fid].blendSum += model->animSets[i].blend * model->animSets[i].anims[j].blend;
            }
        }
    }

    for (int f = 0; f < frame.size(); f++) {
        if (frame[f].blendSum != 0.00) {
            frame[f].blendMult = 1.00 / frame[f].blendSum;
        } else {
            frame[f].blendMult = 0.00;
        }
        if (!model->addToOne && (frame[f].blendMult != 0)) {
            frame[f].blendMult = 1.0;
        }
        if (frame[f].blendMult != 0.00) {
            frame[f].transform.setIdentity();
            frame[f].animPos = Vector3f(0.0, 0.0, 0.0);
            frame[f].animScale = Vector3f(0.0, 0.0, 0.0);
            frame[f].animRot = Quaternionf(0.0, 0.0, 0.0, 0.0);
        }
    }

    // Calc animation
    for (int set = 0; set < model->animSets.size(); set++) {
        float time = model->animSets[set].time;
        for (int i = 0; i < model->animSets[set].anims.size(); i++) {
            Anim* a = &model->animSets[set].anims[i];
            int f = a->frame;
            if (frame[f].priority == model->animSets[set].priority) {
                for (int j = 0; j < a->keySets.size(); j++) {
                    KeySet* ks = &a->keySets[j];
                    float dTime;
                    if (ks->keys.size() > 1) {
                        dTime = ks->keys[ks->keys.size() - 1].time - ks->keys[ks->keys.size() - 2].time;
                    } else {
                        dTime = 0.0;
                    }
                    if (time >= ks->keys[0].time && time < ks->keys[ks->keys.size() - 1].time + dTime) {
                        int x = 0;
                        int y= ks->keys.size();
                        if (y == 0) {
                            break;
                        }
                        while (x+1 != y) {
                            int h = (x + y) / 2;
                            if (time >= ks->keys[h].time) {
                                x = h;
                            } else {
                                y = h;
                            }
                        }

                        float part;
                        Key* yKey;
                        if (y == ks->keys.size()) {
                            part = (time - ks->keys[x].time) / dTime;
                            yKey = &ks->keys[0];
                        } else {
                            part = (time - ks->keys[x].time) / (ks->keys[y].time - ks->keys[x].time);
                            yKey = &ks->keys[y];
                        }

                        switch (ks->animType) {
                            case 0:
                                frame[f].animRot.coeffs() += ks->keys[x].rot.slerp(part, yKey->rot).coeffs() *
                                    (frame[f].blendMult * model->animSets[set].blend * a->blend);
                                break;
                            case 1:
                                frame[f].animScale += lerp(ks->keys[x].scale, yKey->scale, part) *
                                    (frame[f].blendMult * model->animSets[set].blend * a->blend);
                                break;
                            case 2:
                                frame[f].animPos += lerp(ks->keys[x].pos, yKey->pos, part) *
                                    (frame[f].blendMult * model->animSets[set].blend * a->blend);
                                break;
                            default:
                                frame[f].animRot.coeffs() += ks->keys[x].rot.slerp(part, yKey->rot).coeffs() *
                                    (frame[f].blendMult * model->animSets[set].blend * a->blend);
                                frame[f].animScale += lerp(ks->keys[x].scale, yKey->scale, part) *
                                    (frame[f].blendMult * model->animSets[set].blend * a->blend);
                                frame[f].animPos += lerp(ks->keys[x].pos, yKey->pos, part) *
                                    (frame[f].blendMult * model->animSets[set].blend * a->blend);
                                break;
                        }
                    }
                }
            }
        }
    }

    // Calc Frame Transforms
    for (int f = 0; f < frame.size(); f++) {
        frame[f].animRot.normalize();
        frame[f].transform.fromPositionOrientationScale(frame[f].animPos, frame[f].animRot, frame[f].animScale);
    }

    // Reset Blend Values
    for (int i = 0; i < model->animSets.size(); i++) {
        model->animSets[i].blend = 0.00;
        model->animSets[i].priority = -1;
        for (int j = 0; j < model->animSets[i].anims.size(); j++) {
            model->animSets[i].anims[j].blend = 1.00;
        }
    }
}

AnimModel* AnimInstance::getModel()
{
    return model;
}

BoneInstance::BoneInstance(BoneModel *model)
    : model(model), nKeyVerts(0)
{
    texture.resize(model->materials.size());
    for (int i = 0; i < model->materials.size(); i++) {
        if (!model->materials[i].textured) {
            texture[i] = -1;
        } else {
            texture[i] = model->materials[i].textureID;
        }
    }

    keyVert = new KeyVert[MAXKEYVERTS];
    frame.resize(model->frames.size());
    for (int i = 0; i < frame.size(); i++) {
        frame[i].blendMult = model->frames[i].blendMult;
        frame[i].blendSum  = model->frames[i].blendSum;
        frame[i].priority  = model->frames[i].priority;
        frame[i].animPos   = model->frames[i].animPos;
        frame[i].animScale = model->frames[i].animScale;
        frame[i].animRot   = model->frames[i].animRot;
        frame[i].transform = model->frames[i].transform;
    }
}

BoneInstance::BoneInstance()
{
}

BoneInstance::~BoneInstance()
{
    delete[] keyVert;
}

void BoneInstance::render()
{
    Transform3f theView;
    glGetFloatv(GL_MODELVIEW_MATRIX, theView.data());
    theView = viewInverse(theView);

    glPushMatrix();

    // Calc Matrix
    Transform3f theMat = transform();

    // Setup BS
    BS theBS(theMat * bs.pos, bs.radius * qMax(qMax(scale.x(), scale.y()), scale.z()));

    if (true || bsInFrustum(instFrust, theBS)) {
        glMultMatrixf(theMat.data());
        model->setTextures(texture);

        for (int i = 0; i < frame.size(); i++) {
            model->frames[i].blendMult = frame[i].blendMult;
            model->frames[i].blendSum  = frame[i].blendSum;
            model->frames[i].priority  = frame[i].priority;
            model->frames[i].animPos   = frame[i].animPos;
            model->frames[i].animScale = frame[i].animScale;
            model->frames[i].animRot   = frame[i].animRot;
            model->frames[i].transform = frame[i].transform;
        }

        if (nprMode) {
            Vector3f lVec(lightDir[0], lightDir[1], lightDir[2]);
            Transform3f matInv(extractRotation(theView).inverse());
            model->renderCartoony((matInv * lVec).normalized());
        } else {
            model->render();
        }

        // Get Key Vertices
        for (int i = 0; i < nKeyVerts; i++) {
            // Convert from view to world space
            keyVert[i].pos = theView * model->meshes[keyVert[i].meshID].vertexBuf[keyVert[i].vertID].pos;
        }
    } else {
        glMultMatrixf(theMat.data());

        for (int i = 0; i < frame.size(); i++) {
            model->frames[i].blendMult = frame[i].blendMult;
            model->frames[i].blendSum  = frame[i].blendSum;
            model->frames[i].priority  = frame[i].priority;
            model->frames[i].animPos   = frame[i].animPos;
            model->frames[i].animScale = frame[i].animScale;
            model->frames[i].animRot   = frame[i].animRot;
            model->frames[i].transform = frame[i].transform;
        }

        int last = -1;

        Transform3f curViewMat;
        glGetFloatv(GL_MODELVIEW_MATRIX, curViewMat.data());
        model->curViewMat = curViewMat;

        // Get Key Vertices
        for (int i = 0; i < nKeyVerts; i++) {
            if (last != keyVert[i].meshID) {
                model->setupBones(keyVert[i].meshID);
            }
            last = keyVert[i].meshID;
            keyVert[i].pos = theView * model->getVertPos(keyVert[i].meshID, keyVert[i].vertID);
        }
    }

    glPopMatrix();
}

void BoneInstance::setFrame(unsigned int set, float time)
{
    if (set >= (unsigned int)model->animSets.size()) {
        throw IllegalArgumentException(__FILE__, __LINE__, "set > number of sets");
    }

    for (int i = 0; i < model->animSets[set].anims.size(); i++) {
        Anim &a = model->animSets[set].anims[i];
        int f = a.frame;
        frame[f].transform.setIdentity();
        frame[f].animPos = Vector3f(0.0, 0.0, 0.0);
        frame[f].animScale = Vector3f(1.0, 1.0, 1.0);
        frame[f].animRot = Quaternionf(1.0, 0.0, 0.0, 0.0);
        for (int j = 0; j < a.keySets.size(); j++) {
            KeySet &ks = a.keySets[j];
            int dTime;
            if (ks.keys.size() > 1) {
                dTime = ks.keys[ks.keys.size() - 1].time - ks.keys[ks.keys.size() - 2].time;
            } else {
                dTime = 0;
            }
            if ((time >= ks.keys[0].time) && (time < ks.keys[ks.keys.size() - 1].time + dTime)) {
                int x = 0;
                int y = ks.keys.size();
                if (y == 0) {
                    break;
                }
                while (x+1 != y) {
                    int h = (x + y) / 2;
                    if (time >= ks.keys[h].time) {
                        x = h;
                    } else {
                        y = h;
                    }
                }

                float part;
                Key *yKey;
                if (y == ks.keys.size()) {
                    part = (time - ks.keys[x].time) / (float)dTime;
                    yKey = &ks.keys[0];
                } else {
                    part = (time - ks.keys[x].time) / (float)(ks.keys[y].time - ks.keys[x].time);
                    yKey = &ks.keys[y];
                }

                switch (ks.animType) {
                    case 0:
                        frame[f].animRot = ks.keys[x].rot.slerp(part, yKey->rot);
                        break;
                    case 1:
                        frame[f].animScale = lerp(ks.keys[x].scale, yKey->scale, part);
                        break;
                    case 2:
                        frame[f].animPos = lerp(ks.keys[x].pos, yKey->pos, part);
                        break;
                    default:
                        frame[f].animRot = ks.keys[x].rot.slerp(part, yKey->rot);
                        frame[f].animScale = lerp(ks.keys[x].scale, yKey->scale, part);
                        frame[f].animPos = lerp(ks.keys[x].pos, yKey->pos, part);
                        break;
                }
            }
        }
    }

    for (int f = 0; f < frame.size(); f++) {
        frame[f].transform.fromPositionOrientationScale(frame[f].animPos, frame[f].animRot, frame[f].animScale);
    }

    // Calc Combined matrices
    glPushMatrix();
    glLoadIdentity();
    for (int f = 0; f < model->rootFrames.size(); f++) {
        calcCombinedMatrix(model->rootFrames[f]);
    }
    glPopMatrix();
}

void BoneInstance::setAnimSetBlend(unsigned int set, float time, float blend, int priority)
{
    if (set >= (unsigned int)model->animSets.size()) {
        throw IllegalArgumentException(__FILE__, __LINE__, "set > number of sets");
    }
    model->animSets[set].time = time;
    model->animSets[set].blend = blend;
    model->animSets[set].priority = priority;
}

void BoneInstance::setAnimBlend(unsigned int set, unsigned int anim, float blend)
{
    if (set >= (unsigned int)model->animSets.size()) {
        throw IllegalArgumentException(__FILE__, __LINE__, "set > number of sets");
    }
    if (anim >= (unsigned int)model->animSets[set].anims.size()) {
        throw IllegalArgumentException(__FILE__, __LINE__, "anim > number of animations");
    }

    model->animSets[set].anims[anim].blend = blend;
}

void BoneInstance::calcAnim()
{
    // PreCalcs
    for (int f = 0; f < frame.size(); f++) {
        frame[f].blendSum = 0.00;
        frame[f].priority = 0;
    }

    for (int i = 0; i < model->animSets.size(); i++) {
        for (int j = 0; j < model->animSets[i].anims.size(); j++) {
            int fid = model->animSets[i].anims[j].frame;
            if (frame[fid].priority <= model->animSets[i].priority) {
                if (frame[fid].priority < model->animSets[i].priority) {
                    frame[fid].priority = model->animSets[i].priority;
                    frame[fid].blendSum = 0.00;
                }
                frame[fid].blendSum += model->animSets[i].blend * model->animSets[i].anims[j].blend;
            }
        }
    }

    for (int f = 0; f < frame.size(); f++) {
        if (frame[f].blendSum != 0.00) {
            frame[f].blendMult = 1.00 / frame[f].blendSum;
        } else {
            frame[f].blendMult = 0.00;
        }
        if (!model->addToOne && (frame[f].blendMult != 0)) {
            frame[f].blendMult = 1.0;
        }
        if (frame[f].blendMult != 0.00) {
            frame[f].transform.setIdentity();
            frame[f].animPos = Vector3f(0.0, 0.0, 0.0);
            frame[f].animScale = Vector3f(0.0, 0.0, 0.0);
            frame[f].animRot = Quaternionf(0.0, 0.0, 0.0, 0.0);
        }
    }

    // Calc animation
    for (int set = 0; set < model->animSets.size(); set++) {
        float time = model->animSets[set].time;
        for (int i = 0; i < model->animSets[set].anims.size(); i++) {
            Anim &a = model->animSets[set].anims[i];
            int f = a.frame;
            if (frame[f].priority == model->animSets[set].priority) {
                for (int j = 0; j < a.keySets.size(); j++) {
                    KeySet &ks = a.keySets[j];
                    int dTime;
                    if (ks.keys.size() > 1) {
                        dTime = ks.keys[ks.keys.size() - 1].time - ks.keys[ks.keys.size() - 2].time;
                    } else {
                        dTime = 0;
                    }
                    if (time >= ks.keys[0].time && time < ks.keys[ks.keys.size() - 1].time + dTime) {
                        int x = 0;
                        int y = ks.keys.size();
                        if (y == 0) {
                            break;
                        }
                        while (x+1 != y) {
                            int h = (x + y) / 2;
                            if (time >= ks.keys[h].time) {
                                x = h;
                            } else {
                                y = h;
                            }
                        }

                        float part;
                        Key *yKey;
                        if (y == ks.keys.size()) {
                            part = (time - ks.keys[x].time) / (float)dTime;
                            yKey = &ks.keys[0];
                        } else {
                            part = (time - ks.keys[x].time) / (float)(ks.keys[y].time - ks.keys[x].time);
                            yKey = &ks.keys[y];
                        }

                        switch (ks.animType) {
                            case 0:
                                frame[f].animRot.coeffs() += ks.keys[x].rot.slerp(part, yKey->rot).coeffs() *
                                    (frame[f].blendMult * model->animSets[set].blend * a.blend);
                                break;
                            case 1:
                                frame[f].animScale += lerp(ks.keys[x].scale, yKey->scale, part) *
                                    (frame[f].blendMult * model->animSets[set].blend * a.blend);
                                break;
                            case 2:
                                frame[f].animPos += lerp(ks.keys[x].pos, yKey->pos, part) *
                                    (frame[f].blendMult * model->animSets[set].blend * a.blend);
                                break;
                            default:
                                frame[f].animRot.coeffs() += ks.keys[x].rot.slerp(part, yKey->rot).coeffs() *
                                    (frame[f].blendMult * model->animSets[set].blend * a.blend);
                                frame[f].animScale += lerp(ks.keys[x].scale, yKey->scale, part) *
                                    (frame[f].blendMult * model->animSets[set].blend * a.blend);
                                frame[f].animPos += lerp(ks.keys[x].pos, yKey->pos, part) *
                                    (frame[f].blendMult * model->animSets[set].blend * a.blend);
                                break;
                        }
                    }
                }
            }
        }
    }

    // Calc Frame Transforms
    for (int f = 0; f < frame.size(); f++) {
        frame[f].animRot.normalize();
        frame[f].transform.fromPositionOrientationScale(frame[f].animPos, frame[f].animRot, frame[f].animScale);
    }

    // Reset Blend Values
    for (int i = 0; i < model->animSets.size(); i++) {
        model->animSets[i].blend = 0.00;
        model->animSets[i].priority = -1;
        for (int j = 0; j < model->animSets[i].anims.size(); j++) {
            model->animSets[i].anims[j].blend = 1.00;
        }
    }

    // Calc Combined matrices
    glMatrixMode(GL_MODELVIEW_MATRIX);
    glPushMatrix();
    glLoadIdentity();
    for (int f = 0; f < model->rootFrames.size(); f++) {
        calcCombinedMatrix(model->rootFrames[f]);
    }
    glPopMatrix();
}

void BoneInstance::addKeyVert(int mid, int vid)
{
    keyVert[nKeyVerts].meshID = mid;
    keyVert[nKeyVerts].vertID = vid;
    if (++nKeyVerts > MAXKEYVERTS) {
        --nKeyVerts;
    }
}

void BoneInstance::calcCombinedMatrix(int n)
{
    glPushMatrix();
    glMultMatrixf(frame[n].transform.data());
    glGetFloatv(GL_MODELVIEW_MATRIX, frame[n].transform.data());
    for (int i = 0; i < model->frames[n].subFrames.size(); i++) {
        calcCombinedMatrix(model->frames[n].subFrames[i]);
    }
    glPopMatrix();
}

Vector3f BoneInstance::getKeyPos(int kid) const
{
    return keyVert[kid].pos;
}

HelpFrame BoneInstance::getFrame(int fid)
{
    return frame[fid];
}

BoneModel* BoneInstance::getModel()
{
    return model;
}
