/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
// glew.h must be before any other opengl headers when it is used
#include <GL/glew.h>

#include "Main.h"

// FruitWars headers
#include "Math3D.h"
#include "Model3D.h"
#include "Instance.h"
#include "Camera.h"
#include "Globals.h"
#include "Collision.h"
#include "ConfigWidget.h"

// KDE headers
#include <KApplication>
#include <KAboutData>
#include <KCmdLineArgs>

// Qt headers
#include <QTimer>
#include <QKeyEvent>
#include <QTime>
#include <QPainter>

USING_PART_OF_NAMESPACE_EIGEN

MainWidget::MainWidget(QWidget* parent)
    : QGLWidget(glcontext = new QGLContext(QGLFormat::defaultFormat()))
{
    QTimer* timer = new QTimer(this);
    QObject::connect(timer, SIGNAL(timeout()), SLOT(idle()));
    timer->start();
    globaltime = new QTime();
    globaltime->start();
    setMouseTracking(true);
}

void MainWidget::paintGL()
{
    if (gameQuit) return;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    drunkActive = false; //XXX

    if (drunkActive) glViewport(0, 0, drunkSize, drunkSize);

    glLoadIdentity();

    // Camera
    myCam.render();

    // Light
    glLightfv(GL_LIGHT0, GL_POSITION, lightDir);


    // Do some rendering
    // Terrain
    myTer.render();
    myTerFinish.render();

    // Sky
    mySky.render();

    // Objects
    instFrust = Frustum::calc();

    // Players
    for (unsigned int i = 0; i < nplayers; i++) {
        myPlayers[i].render();
    }

    glDepthMask(GL_FALSE);
    for (unsigned int i = 0; i < nplayers; i++) {
        myPlayers[i].warpRender();
    }
    glDepthMask(GL_TRUE);

    // Fruit
    fruitSpawner.render();
    glDepthMask(GL_FALSE);
    fruitSpawner.warpRender2();
    glDepthMask(GL_TRUE);

    // Fruitbasket
    for (int i = 0; i < nBaskets; i++) {
        myBaskets[i].render();
    }

    // Shadows
    glDepthMask(GL_FALSE);
    for (unsigned int i = 0; i < nplayers; i++) {
        myPlayers[i].renderShadow();
    }
    glDepthMask(GL_TRUE);

    // Sea
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
    mySea.render();
    mySeaFinish.render();
    glDisable(GL_BLEND);
    glRotatef(glblRotate, 0.0, 1.0, 0.0);

    // Particles
    glDepthMask(GL_FALSE);
    fruitSpawner.warpRender();
    for (unsigned int i = 0; i < nplayers; i++) {
        myPlayers[i].warpRender2();
    }
    for (int i = 0; i < nBaskets; i++) {
        myBaskets[i].getStarShow().render();
    }
    glDepthMask(GL_TRUE);

    // Indicator
    glDisable(GL_DEPTH_TEST);
    theIndicator.render();
    glEnable(GL_DEPTH_TEST);

    // Crosshair
    glDisable(GL_DEPTH_TEST);
    theCrossHair.render();
    glEnable(GL_DEPTH_TEST);

    // Draw FPS
    if (drunkActive) {
        renderDrunkStuff();
    } else {
        drawFPS();
    }

    glFlush();
}

void MainWidget::renderDrunkStuff()
{
}

void MainWidget::drawFPS()
{
    int theTime = time();
    int dt = theTime - oldTime;
    oldTime = theTime;

    if (pastTime + dt >= 1000) {
        float fps = frameCount + (1000.0 - pastTime) / dt;
        pastTime = 0;
        frameCount = 0;
        fpsString = QString("FPS: %1 (%2)").arg(fps).arg(time());
        // TODO send score update to other players if host
    } else {
        pastTime += dt;
        frameCount++;
    }

    glEnable(GL_COLOR_MATERIAL);
    initTextDrawing();
    glColor3f(1.0, 1.0, 1.0);
    drawText(5, 20, fpsString);
    if (teamMode == 2) {
        drawRoundText(0); // playerId
    } else {
        drawRoundText(0); // teamId
    }

    if (keyState['H']) {
        drawHUDRect();
    }
    
    // drawChat
    deInitTextDrawing();
    glDisable(GL_COLOR_MATERIAL);
}

void MainWidget::initTextDrawing()
{
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(0, width(), height(), 0);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
}

void MainWidget::deInitTextDrawing()
{
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glEnable(GL_LIGHTING);
    glEnable(GL_DEPTH_TEST);
    glPopMatrix();
}

void MainWidget::drawText(int x, int y, const QString& text)
{
    QImage img(640, 40, QImage::Format_Mono);
    img.fill(0);
    QPainter p(&img);
    QFont font("Helvetica");
    font.setPixelSize(18);
    p.setFont(font);
    QFontMetrics m = p.fontMetrics();
    int a = m.ascent();
    p.setPen(Qt::white);
    p.drawText(qMin(x, 10), a, text);
    p.end();
    glRasterPos2i(qMax(0, x-10), y+(img.height()-a));
    img = img.mirrored();
    glBitmap(img.width(), img.height(), 0, 0, 0, 0, img.bits());
//    printf("[%3d, %3d]: %s\n", x, y, qPrintable(text));
}

void MainWidget::drawRoundText(int teamId)
{
    glColor3f(1.0, 0.0, 0.0);

    int place = 1;
    int maxScore = 0, globalMaxScore = 0;
    int myScore = 0;
    int m2 = 0;
    QString winner;

    if (teamMode == 2) {
        myScore = myBaskets[myPlayers[teamId].getBaskId()].getScore();
        for (unsigned int i = 0; i < nplayers; i++) {
            if (myPlayers[teamId].getBaskId() != myPlayers[i].getBaskId()) {
                if (myBaskets[myPlayers[i].getBaskId()].getScore() > myScore) place++;
                maxScore = qMax(maxScore, myBaskets[myPlayers[i].getBaskId()].getScore());
            }
            globalMaxScore = qMax(globalMaxScore, myBaskets[myPlayers[i].getBaskId()].getScore());
        }
        // XXX stuff to get winners when game has ended
    } else {
        myScore = myTeams[teamId].getScore();
        for (int i = 0; i < nTeams; i++) {
            if (!myTeams[i].isEmpty()) {
                if (teamId != i) {
                    if (myTeams[i].getScore() > myTeams[teamId].getScore()) place++;
                    maxScore = qMax(maxScore, myTeams[i].getScore());
                }
                globalMaxScore = qMax(globalMaxScore, myTeams[i].getScore());
            }
        }
        // XXX stuff to get winners when game has ended
    }

    QString placeText = QString("%1th").arg(place);
    switch (place) {
        case 1: placeText = "1st"; break;
        case 2: placeText = "2nd"; break;
        case 3: placeText = "3rd"; break;
    }
    drawText(width() - 35, height() - 50, placeText);

    QString scoreText = QString((maxScore <= myScore) ? "+%1" : "%1").arg(myScore - maxScore);
    drawText(width() - scoreText.length() * 12 - 5, height() - 30, scoreText);
    scoreText = QString("Score: %1").arg(myScore);
    drawText(width() - scoreText.length() * 10 - 5, 20, scoreText);
    
    int y = 40;
    if (pointgame) {
        QString text = QString("Points left: %1").arg((maxPoints - myScore));
        drawText(width() - text.length()*8 - 5, y, text);
        y += 20;
    }
    if (timegame) {
        QString text = QString("Time left: %1s").arg((maxTime - theTime) / 1000);
        drawText(width() - text.length()*8 - 10, y, text);
    }

    glColor3f(1.0, 1.0, 1.0);
}

void MainWidget::drawHUDRect()
{
    typedef QPair<int, int> IIPair;
    glColor4f(1.0, 1.0, 1.0, 0.60);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glBindTexture(GL_TEXTURE_2D, HUDRectTex);
    glEnable(GL_TEXTURE_2D);

    float halfWidth = width() / 2;
    float halfHeight = height() / 2;
    glBegin(GL_QUADS);
        glTexCoord2f(0.00, 1.00);
        glVertex3f(halfWidth - 200.0, halfHeight - 200.0, 0.90);
        glTexCoord2f(0.00, 0.00);
        glVertex3f(halfWidth - 200.0, halfHeight + 200.0, 0.90);
        glTexCoord2f(1.00, 0.00);
        glVertex3f(halfWidth + 200.0, halfHeight + 200.0, 0.90);
        glTexCoord2f(1.00, 1.00);
        glVertex3f(halfWidth + 200.0, halfHeight - 200.0, 0.90);
    glEnd();

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glColor4f(1.0, 1.0, 1.0, 1.0);

    // draw scores
    QList<QPair<int, int> > scores;
    if (teamMode == 2) {
        for (int i = 0; i < nplayers; i++) {
            scores.append(qMakePair(-myBaskets[myPlayers[i].getBaskId()].getScore(), i));
        }
    } else {
        for (int i = 0; i < nTeams; i++) {
            scores.append(qMakePair(-myTeams[i].getScore(), i));
        }
    }
    qSort(scores);
    if (teamMode == 2) {
        int y = halfHeight - 90;
        foreach (const IIPair& p, scores) {
            Vector3f c = myPlayerColor[myPlayers[p.second].getBaskId()];
            glColor3f(c.x(), c.y(), c.z());
            drawText(halfWidth - 150, y, myPlayers[p.second].getNickName());
            drawText(halfWidth + 100, y, QString("%1").arg(-p.first));
            y += 20;
        }
    } else {
        int y = halfHeight - 110;
        foreach (const IIPair& p, scores) {
            Vector3f c = myPlayerColor[myTeams[p.second].getColorId()];
            glColor3f(c.x(), c.y(), c.z());
            drawText(halfWidth - 150, y, myTeams[p.second].getTeamName());
            drawText(halfWidth + 150, y, QString("%1").arg(-p.first));
            for (int i = 0; i < myTeams[p.second].getPlayerCount(); i++) {
                Vector3f c = myPlayerColor[myPlayers[myTeams[p.second].getPlayId(i)].getColorId()];
                glColor3f(c.x(), c.y(), c.z());
                drawText(halfWidth - 130, y + 20 + 20 * i, myPlayers[myTeams[p.second].getPlayId(i)].getNickName());
            }
            y += 25 * (1 + (4 >> teamMode));
        }
    }

    glColor3f(1.0, 1.0, 1.0);
}

void MainWidget::resizeGL(int w, int h)
{
    if (w == 0 || h == 0) return;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    float wf = w;
    float hf = h;
    gluPerspective(FOV, wf / hf, NEARF, FARF);
    glViewport(0, 0, w, h);
    glMatrixMode(GL_MODELVIEW);
}

void MainWidget::keyPressEvent(QKeyEvent* event)
{
    int key = event->key();
    if (key == Qt::Key_Escape) {
        exit(0);
    }
    if (key == 'Z') {
        myPlayers[0].setStunTime(playerStunTime);
    }
    if (key >= 0 && key <= 0xff) {
        keyState[key] = true;
        if (time() - dLastTime[key] < DCDIF) dKeyState[key] = true;
    }
}

void MainWidget::keyReleaseEvent(QKeyEvent* event)
{
    int key = event->key();
    if (key >= 0 && key <= 0xff) {
        keyState[key] = false;
        dKeyState[key] = false;
        dLastTime[key] = false;
    }
}

void MainWidget::mouseMoveEvent(QMouseEvent* event)
{
    int dx = event->x() - lastPos.x();
    int dy = event->y() - lastPos.y();
    lastPos = event->pos();
//    QCursor::setPos(mapToGlobal(QPoint(width()/2, height()/2)));
    myCam.setTargetAngle(myCam.targetAngle() + Vector3f(dy * mouseSpeed, dx * mouseSpeed, 0.0));
    if (myCam.targetAngle().x() > MAXXANGLE) myCam.setTargetAngle(Vector3f(MAXXANGLE, myCam.targetAngle().y(), myCam.targetAngle().z()));
    if (myCam.targetAngle().x() < -MAXXANGLE) myCam.setTargetAngle(Vector3f(-MAXXANGLE, myCam.targetAngle().y(), myCam.targetAngle().z()));
}

void MainWidget::idle()
{
    theTime = time();
    if (gameQuit) return;

    if ((time() - syncTime) >= (unsigned int)(1000 / syncRate)) {
        syncTime = syncTime + (1000 / syncRate);

        // Player updating
        myPlayers[0].rotateY(myCam.targetAngle().y() + M_PI);
        myPlayers[0].update();

        for (unsigned int i = 0; i < nplayers; i++) {
            if (i != 0) {
                myPlayers[i].nwUpdate();
            }
        }

        // Basket updating
        for (int i = 0; i < nBaskets; i++) {
            myBaskets[i].getStarShow().update();
        }

        // Fruit updating
        fruitSpawner.update();

        calcCollisions();

        myCam.setCenter(myPlayers[0].getPos() + Vector3f(0, PLAYERHEIGHT, 0));
        myCam.update();

//        display();
        updateGL();
    }
}

void MainWidget::initializeGL()
{
    gl_13_available = true;
    glewInit();
    basicSetup();
    lightSetup();
    varInit();
    envInit();
    camInit();
    modelLoading();
}

int main(int argc, char** argv)
{
    KAboutData aboutData(
        "fruitwars",
        0,
        ki18n("Fruit Wars"),
        "1.0",
        ki18n("A simple and fun 3D multiplayer game."),
        KAboutData::License_GPL,
        ki18n("(c) 2008 Marijn Kruisselbrink")
    );

    KCmdLineArgs::init(argc, argv, &aboutData);
    KApplication app;

    srand(QTime::currentTime().msec());

    for (int i = 0; i < 8; i++) {
        thaPerm[i] = i;
    }
    for (int i = 0; i < 1000; i++) {
        int j = random() % 8;
        int k = random() % 8;
        int h = thaPerm[j];
        thaPerm[j] = thaPerm[k];
        thaPerm[k] = h;
    }

    ConfigWidget cfgWin;
    cfgWin.show();

    return app.exec();
    //return 0;
}
