/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
// glew.h must be before any other opengl related headers
#include <GL/glew.h>

#include "Terrain.h"

// FruitWars headers
#include "GameMath.h"
#include "Globals.h"
#include "Textures.h"

// Qt headers
#include <QImage>

USING_PART_OF_NAMESPACE_EIGEN

Terrain::Terrain(int xSegs, int zSegs, float sizeX, float sizeZ, float minHeight, float maxHeight)
    : xSegs(xSegs),
    zSegs(zSegs),
    xSegSize(sizeX / xSegs),
    zSegSize(sizeZ / zSegs),
    minHeight(minHeight),
    maxHeight(maxHeight),
    patchSize(qMax(xSegs, zSegs)),
    pos(0.0, 0.0, 0.0),
    texMeth(SingleTex)
{
    height.resize( (xSegs+1) * (zSegs+1) );
    normal.resize(height.size());
    color.resize(height.size());
    vertBuf = new VertBufTer[height.size()];
    for (unsigned int i = 0; i < height.size(); i++) {
        height[i] = minHeight;
        normal[i] = Vector3f(0.0, 1.0, 0.0);
        color[i].r = 1.0; color[i].g = 1.0; color[i].b = 1.0; color[i].a = 1.0;
    }

    for (int i = 0; i < 3; i++) {
        textures[i] = 0;
        texScale[i] = Vector2f(1.0, 1.0);
    }

    updateVertBuf();

    nPatches = 1;
    patch.resize(1);
    patch[0].bx = 0;
    patch[0].bz = 0;
    patch[0].ex = xSegs;
    patch[0].ez = zSegs;
    patch[0].bb.leftBottom = Vector3f(0.0, minHeight, 0.0);
    patch[0].bb.rightTop = Vector3f(sizeX, maxHeight, sizeZ);
    patch[0].nStrips = zSegs;
    patch[0].indices.resize(patch[0].nStrips);

    for (int i = 0; i < patch[0].nStrips; i++) {
        int stripCount = (xSegs + 1) * 2;
        patch[0].indices[i].resize(stripCount);
        int curIndex = (i + 1) * (xSegs + 1);
        for (int j = 0; j < stripCount; j++) {
            patch[0].indices[i][j] = curIndex;
            if (j % 2 == 0) {
                curIndex -= xSegs + 1;
            } else {
                curIndex += xSegs + 2;
            }
        }
    }
}

Terrain::Terrain()
{
}

Terrain::~Terrain()
{
    // delete[] vertBuf
}

void Terrain::useVertexBuffer(bool u)
{
    useVertBuf = u;
}

void Terrain::loadHeightMap(const QString& fileName)
{
    QImage img(fileName);
    if (img.isNull()) {
        fprintf(stderr, "Error loading %s\n", qPrintable(fileName));
        return;
    }

    int w = img.width();
    int h = img.height();

    for (int i = 0; i < xSegs; i++) {
        for (int j = 0; j < zSegs; j++) {
            float au = (float)i / xSegs;
            float av = (float)j / zSegs;
            int ax = (int)floor(au * (w - 1));
            int ay = (int)floor(av * (h - 1));
            float xpart = au * (w - 1) - ax;
            float ypart = av * (h - 1) - ay;

            float val1, val2, val;
            if (xpart > 0) {
                val1 = (1.0 - xpart) * qGray(img.pixel(ax, ay)) + xpart * qGray(img.pixel(ax+1, ay));
            } else {
                val1 = qGray(img.pixel(ax, ay));
            }

            if (ypart > 0) {
                if (xpart > 0) {
                    val2 = (1.0 - xpart) * qGray(img.pixel(ax, ay+1)) + xpart * qGray(img.pixel(ax+1, ay+1));
                } else {
                    val2 = qGray(img.pixel(ax, ay+1));
                }
                val = (1.0 - ypart) * val1 + ypart * val2;
            } else {
                val = val1;
            }

            height[j * (xSegs + 1) + i] = (val / 255.0) * (maxHeight - minHeight) + minHeight;
        }
    }

    calcNormals(0, 0, xSegs, zSegs);
    setPatchSize(patchSize);
}

void Terrain::loadColorMap(const QString& fileName)
{
    QImage img(fileName);
    if (img.isNull()) {
        fprintf(stderr, "Error loading %s\n", qPrintable(fileName));
        return;
    }

    for (int i = 0; i < xSegs; i++) {
        for (int j = 0; j < zSegs; j++) {
            float au = (float)i / xSegs;
            float av = (float)j / zSegs;
            int ax = (int)floor(au * (img.width() - 1));
            int ay = (int)floor(av * (img.height() - 1));
            float xpart = au * (img.width() - 1) - ax;
            float ypart = av * (img.height() - 1) - ay;

            Vector3f val1, val2, val;
            if (xpart > 0) {
                val1.x() = (1.0 - xpart) * qBlue(img.pixel(ax, ay)) + xpart * qBlue(img.pixel(ax+1, ay));
                val1.y() = (1.0 - xpart) * qGreen(img.pixel(ax, ay)) + xpart * qGreen(img.pixel(ax+1, ay));
                val1.z() = (1.0 - xpart) * qRed(img.pixel(ax, ay)) + xpart * qRed(img.pixel(ax+1, ay));
            } else {
                val1.x()  = qBlue(img.pixel(ax, ay));
                val1.y()  = qGreen(img.pixel(ax, ay));
                val1.z()  = qRed(img.pixel(ax, ay));
            }

            if (ypart > 0) {
                if (xpart > 0) {
                    val2.x() = (1.0 - xpart) * qBlue(img.pixel(ax, ay+1)) + xpart * qBlue(img.pixel(ax+1, ay+1));
                    val2.y() = (1.0 - xpart) * qGreen(img.pixel(ax, ay+1)) + xpart * qGreen(img.pixel(ax+1, ay+1));
                    val2.z() = (1.0 - xpart) * qRed(img.pixel(ax, ay+1)) + xpart * qRed(img.pixel(ax+1, ay+1));
                } else {
                    val2.x() = qBlue(img.pixel(ax, ay+1));;
                    val2.y() = qGreen(img.pixel(ax, ay+1));
                    val2.z() = qRed(img.pixel(ax, ay+1));
                }
                val = val1 * (1 - ypart) + val2 * ypart;
            } else {
                val = val1;
            }

            color[j * (xSegs + 1) + i].r = val.z() / 255.0;
            color[j * (xSegs + 1) + i].g = val.y() / 255.0;
            color[j * (xSegs + 1) + i].b = val.x() / 255.0;
            color[j * (xSegs + 1) + i].a = 1.0;
        }
    }
}

void Terrain::setHeight(int ax, int az, float y)
{
    if (ax >= 0 && az >= 0 && ax <= xSegs && az <= zSegs) {
        height[az * (xSegs + 1) + ax] = y;
    }
}

void Terrain::setNormal(int ax, int az, const Vector3f& norm)
{
    if (ax >= 0 && az >= 0 && ax <= xSegs && az <= zSegs) {
        normal[az * (xSegs + 1) + ax] = norm;
    }
}

void Terrain::setColor(int ax, int az, const Color& col)
{
    if (ax >= 0 && az >= 0 && ax <= xSegs && az <= zSegs) {
        color[az * (xSegs + 1) + ax] = col;
    }
}

void Terrain::position(const Vector3f& pos)
{
    this->pos = pos;
}

void Terrain::calcNormals(int bx, int bz, int ex, int ez)
{
    if (bx < 0 || bx > xSegs || bz < 0 || bz > zSegs ||
        ex < 0 || ex > xSegs || ez < 0 || ez > zSegs) return;

    int bbx = qMax(bx - 1, 0);
    int bbz = qMax(bz - 1, 0);
    int eex = qMin(ex + 1, xSegs);
    int eez = qMin(ez + 1, zSegs);
    int nFaces = (eex - bbx) * (eez - bbz) * 2;

    // Calc face normals
    Array<Vector3f> norms(nFaces);
    for (int i = bbx; i < eex; i++) {
        for (int j = bbz; j < eez; j++) {
            Vector3f v1(xSegSize, getHeight(i + 1, j) - getHeight(i, j), 0.0);
            Vector3f v2(xSegSize, getHeight(i + 1, j + 1) - getHeight(i, j), -zSegSize);
            norms[((j - bbz) * (eex - bbx) + (i - bbx)) * 2] = v1.cross(v2).normalized();
            v1 = Vector3f(xSegSize, getHeight(i + 1, j + 1) - getHeight(i, j), -zSegSize);
            v2 = Vector3f(0.0, getHeight(i, j + 1) - getHeight(i, j), -zSegSize);
            norms[((j - bbz) * (eex - bbx) + (i - bbx)) * 2 + 1] = v1.cross(v2).normalized();
        }
    }

    // Calc vertex normals
    for (int i = bx; i <= ex; i++) {
        for (int j = bz; j <= ez; j++) {
            int vert = j * (xSegs + 1) + i;
            int face = ((j - bbz) * (eex - bbx) + (i - bbx)) * 2;
            normal[vert] = Vector3f(0, 0, 0);
            if (i < xSegs && j < zSegs) {
                normal[vert] += norms[face];
                normal[vert] += norms[face+1];
            }
            if (i > 0 && j > 0) {
                normal[vert] += norms[face + (bbx - eex - 1) * 2];
                normal[vert] += norms[face + (bbx - eex - 1) * 2 + 1];
            }
            if (i > 0 && j < zSegs) {
                normal[vert] += norms[face - 2];
            }
            if (i < xSegs && j > 0) {
                normal[vert] += norms[face + (bbx - eex) * 2 + 1];
            }
            normal[vert].normalize();
        }
    }
}

void Terrain::setTexMethod(TexMethod mTex)
{
    texMeth = mTex;
    updateVertBuf();
}

void Terrain::setPatchSize(int size)
{
    patchSize = size;

    // Update patches
    int xcount = (int)ceil((float)xSegs / size);
    int zcount = (int)ceil((float)zSegs / size);
    bool extrax = xcount == (int)floor((float)xSegs / size);
    bool extraz = zcount == (int)floor((float)zSegs / size);
    nPatches = xcount * zcount;

    patch.resize(nPatches);
    for (int i = 0; i < xcount; i++) {
        for (int j = 0; j < zcount; j++) {
            int p = j * xcount + i;
            patch[p].bx = i * size;
            patch[p].bz = j * size;
            if (extraz || j < zcount-1) {
                patch[p].ez = (j + 1) * size;
                patch[p].nStrips = size;
                patch[p].indices.resize(patch[p].nStrips);
            } else {
                patch[p].ez = zSegs;
                patch[p].nStrips = zSegs - (int)floor((float)zSegs / size) * size;
                patch[p].indices.resize(patch[p].nStrips);
            }
            if (extrax || i < xcount-1) {
                patch[p].ex = (i + 1) * size;
                for (int k = 0; k < patch[p].nStrips; k++) {
                    patch[p].indices[k].resize((size + 1) * 2);
                    int curIndex = (patch[p].bz + k + 1) * (xSegs + 1) + patch[p].bx;
                    for (unsigned int l = 0; l < patch[p].indices[k].size(); l++) {
                        patch[p].indices[k][l] = curIndex;
                        if (l % 2 == 0) curIndex -= xSegs + 1;
                        else curIndex += xSegs + 2;
                    }
                }
            } else {
                patch[p].ex = xSegs;
                for (int k = 0; k < patch[p].nStrips; k++) {
                    patch[p].indices[k].resize( (patch[p].ex - patch[p].bx + 1) * 2 );
                    int curIndex = (patch[p].bz + k + 1) * (xSegs + 1) + patch[p].bx;
                    for (unsigned int l = 0; l < patch[p].indices[k].size(); l++) {
                        patch[p].indices[k][l] = curIndex;
                        if (l % 2 == 0) curIndex = curIndex - xSegs - 1;
                        else curIndex = curIndex + xSegs + 2;
                    }
                }
            }

            float minh = maxHeight;
            float maxh = minHeight;
            for (int k = patch[p].bx; k <= patch[p].ex; k++) {
                for (int l = patch[p].bz; l <= patch[p].ez; l++) {
                    float h = getHeight(k, l);
                    if (h < minh) minh = h;
                    if (h > maxh) maxh = h;
                }
            }
            patch[p].bb.leftBottom = Vector3f(patch[p].bx * xSegSize + pos.x(),
                            minh + pos.y(),
                            pos.z() - patch[p].ez * zSegSize);
            patch[p].bb.rightTop = Vector3f(patch[p].ex * xSegSize + pos.x(),
                            maxh + pos.y(),
                            pos.z() - patch[p].bz * zSegSize);
        }
    }
}

void Terrain::setMinHeight(float h)
{
    minHeight = h;
}

void Terrain::setMaxHeight(float h)
{
    maxHeight = h;
}

void Terrain::setTexture(int tex, GLuint texId)
{
    if (tex >= 0 && tex <= 2) {
        textures[tex] = texId;
    }
}

void Terrain::setTexScale(int tex, const Vector2f& scale)
{
    if (tex >= 0 && tex <= 2) {
        texScale[tex] = scale;
    }
}

void Terrain::loadTex(int tex, const QString& texFile)
{
    if (tex >= 0 && tex <= 2) {
        if (!loadTexture(texFile, textures[tex], glcontext)) {
            textures[tex] = 0;
            return;
        }

        glBindTexture(GL_TEXTURE_2D, textures[tex]);

        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

        if (texMeth == DetailTex && tex == 1) {
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        }
    }
}

void Terrain::calcSeaTex(int tex, Terrain* terrain, int size, float maxDif)
{
    if ((size == 128 || size == 64 || size == 256) && tex >= 0 && tex <= 2) {
        float data[256*256*4]; /*size*size*4*/

        for (int i = 0; i < size; i++) {
            for (int j = 0; j < size; j++) {
                float posx = pos.x() + (xSegSize * xSegs) * (i / (size - 1.0));
                float posz = pos.z() - (zSegSize * zSegs) * (j / (size - 1.0));
                float hdif;
                if (i > 0 && i < size-1 && j > 0 && j < size-1) {
                    hdif = getGroundHeight(Vector2f(posx, posz)) - terrain->getGroundHeight(Vector2f(posx, posz));
                } else {
                    hdif = getGroundHeight(Vector2f(posx, posz)) - terrain->minHeight;
                }
                float val = qMax(hdif / maxDif, 0.0f);
                val = powf(val, 0.33);
                data[(j * size + i) * 4 + 0] = 1.0 - sqrt(val);
                data[(j * size + i) * 4 + 1] = 1.0 - val;
                data[(j * size + i) * 4 + 2] = 1.0 - val * val * 0.75;
                data[(j * size + i) * 4 + 3] = 0.3 + sqrt(val) * 0.70;
            }
        }

        glGenTextures(1, &textures[tex]);
        glBindTexture(GL_TEXTURE_2D, textures[tex]);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

        gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, size, size, GL_RGBA, GL_FLOAT, data);

        glBindTexture(GL_TEXTURE_2D, textures[tex]);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    }
}

void Terrain::updateVertBuf()
{
    float invXSegs = 1.0 / xSegs;
    float invZSegs = 1.0 / zSegs;
    Vector3f lightVec(lightDir[0], lightDir[1], lightDir[2]);
    switch (texMeth) {
        case SingleTex: {
            for (int i = 0; i <= xSegs; i++) {
                for (int j = 0; j <= zSegs; j++) {
                    int vert = j * (zSegs + 1) + i;
                    vertBuf[vert].pos = Vector3f(pos.x() + i * xSegSize, pos.y() + height[vert], pos.z() - j * zSegSize);
                    vertBuf[vert].norm = normal[vert];
                    vertBuf[vert].texCoord1[0] = (i * invXSegs * texScale[0][0]);
                    vertBuf[vert].texCoord1[1] = (j * invZSegs * texScale[0][1]);
                }
            }
            break;
        }
        case ColorTex: {
            for (int i = 0; i <= xSegs; i++) {
                for (int j = 0; j <= zSegs; j++) {
                    int vert = j * (zSegs + 1) + i;
                    vertBuf[vert].pos = Vector3f(pos.x() + i * xSegSize, pos.y() + height[vert], pos.z() - j * zSegSize);
                    vertBuf[vert].norm = normal[vert];
                    vertBuf[vert].color = color[vert];
                    vertBuf[vert].texCoord1[0] = (i * invXSegs * texScale[0][0]);
                    vertBuf[vert].texCoord1[1] = (j * invZSegs * texScale[0][1]);
                }
            }
            break;
        }
        case DetailTex: {
            for (int i = 0; i <= xSegs; i++) {
                for (int j = 0; j <= zSegs; j++) {
                    int vert = j * (zSegs + 1) + i;
                    vertBuf[vert].pos = Vector3f(pos.x() + i * xSegSize, pos.y() + height[vert], pos.z() - j * zSegSize);
                    if (nprMode) vertBuf[vert].norm = lightVec;
                    else vertBuf[vert].norm = normal[vert];
                    vertBuf[vert].texCoord1[0] = (i * invXSegs * texScale[0][0]);
                    vertBuf[vert].texCoord1[1] = (j * invZSegs * texScale[0][1]);
                    vertBuf[vert].texCoord2[0] = (i * invXSegs * texScale[1][0]);
                    vertBuf[vert].texCoord2[1] = (j * invZSegs * texScale[1][1]);
                    if (nprMode) vertBuf[vert].texCoord3[0] = lightVec.dot(normal[vert]);
                }
            }
            break;
        }
        case ColorMix: {
            for (int i = 0; i <= xSegs; i++) {
                for (int j = 0; j <= zSegs; j++) {
                    int vert = j * (zSegs + 1) + i;
                    vertBuf[vert].pos = Vector3f(pos.x() + i * xSegSize, pos.y() + height[vert], pos.z() - j * zSegSize);
                    vertBuf[vert].norm = normal[vert];
                    vertBuf[vert].color = color[vert];
                    vertBuf[vert].texCoord1[0] = (i * invXSegs * texScale[0][0]);
                    vertBuf[vert].texCoord1[1] = (j * invZSegs * texScale[0][1]);
                    vertBuf[vert].texCoord2[0] = (i * invXSegs * texScale[1][0]);
                    vertBuf[vert].texCoord2[1] = (j * invZSegs * texScale[1][1]);
                }
            }
            break;
        }
        case MixTex: {
            for (int i = 0; i <= xSegs; i++) {
                for (int j = 0; j <= zSegs; j++) {
                    int vert = j * (zSegs + 1) + i;
                    vertBuf[vert].pos = Vector3f(pos.x() + i * xSegSize, pos.y() + height[vert], pos.z() - j * zSegSize);
                    vertBuf[vert].norm = normal[vert];
                    vertBuf[vert].texCoord1[0] = (i * invXSegs * texScale[0][0]);
                    vertBuf[vert].texCoord1[1] = (j * invZSegs * texScale[0][1]);
                    vertBuf[vert].texCoord2[0] = (i * invXSegs * texScale[1][0]);
                    vertBuf[vert].texCoord2[1] = (j * invZSegs * texScale[1][1]);
                    vertBuf[vert].texCoord3[0] = (i * invXSegs * texScale[2][0]);
                    vertBuf[vert].texCoord3[1] = (j * invZSegs * texScale[2][1]);
                }
            }
            break;
        }
    }
}

void Terrain::render()
{
    Frustum frust = Frustum::calc();

    switch(texMeth) {
        case SingleTex: case ColorTex: {
            glEnableClientState(GL_VERTEX_ARRAY);
            glEnableClientState(GL_NORMAL_ARRAY);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            if (texMeth == ColorTex) glEnableClientState(GL_COLOR_ARRAY);
            glVertexPointer(3, GL_FLOAT, sizeof(VertBufTer), vertBuf);
            glNormalPointer(GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].norm);
            glTexCoordPointer(2, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].texCoord1);
            if (texMeth == ColorTex) glColorPointer(4, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].color);

            if (texMeth == ColorTex) glEnable(GL_COLOR_MATERIAL);
            glBindTexture(GL_TEXTURE_2D, textures[0]);
            glEnable(GL_TEXTURE_2D);

            for (int i = 0; i < nPatches; i++) {
                if (aabbInFrustum(frust, patch[i].bb)) renderPatch(i);
            }

            glDisable(GL_TEXTURE_2D);
            glDisableClientState(GL_VERTEX_ARRAY);
            glDisableClientState(GL_NORMAL_ARRAY);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
            if (texMeth == ColorTex) glDisableClientState(GL_COLOR_ARRAY);

            if (texMeth == ColorTex) glDisable(GL_COLOR_MATERIAL);
            break;
        }
        case DetailTex: {
            glEnableClientState(GL_VERTEX_ARRAY);
            glEnableClientState(GL_NORMAL_ARRAY);
            glVertexPointer(3, GL_FLOAT, sizeof(VertBufTer), vertBuf);
            glNormalPointer(GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].norm);
            glClientActiveTexture(GL_TEXTURE0);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            glTexCoordPointer(2, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].texCoord1);
            glClientActiveTexture(GL_TEXTURE1);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            glTexCoordPointer(2, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].texCoord2);
            if (nprMode) {
                glClientActiveTexture(GL_TEXTURE2);
                glEnableClientState(GL_TEXTURE_COORD_ARRAY);
                glTexCoordPointer(1, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].texCoord3);
            }

            glActiveTexture(GL_TEXTURE0);
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D, textures[0]);
            glActiveTexture(GL_TEXTURE1);
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D, textures[1]);

            if (nprMode) {
                glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
                glActiveTexture(GL_TEXTURE2);
                glEnable(GL_TEXTURE_1D);
                glBindTexture(GL_TEXTURE_1D, cartoonTex);
                glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
            }

            glActiveTexture(GL_TEXTURE0);

            for (int i = 0; i < nPatches; i++) {
                if (aabbInFrustum(frust, patch[i].bb)) renderPatch(i);
            }

            if (nprMode) {
                glActiveTexture(GL_TEXTURE2);
                glDisable(GL_TEXTURE_1D);
            }
            glActiveTexture(GL_TEXTURE1);
            glDisable(GL_TEXTURE_2D);
            glActiveTexture(GL_TEXTURE0);
            glDisable(GL_TEXTURE_2D);
            if (!nprMode) {
                glDisableClientState(GL_VERTEX_ARRAY);
                glDisableClientState(GL_NORMAL_ARRAY);
            } else {
                glClientActiveTexture(GL_TEXTURE2);
                glDisableClientState(GL_TEXTURE_COORD_ARRAY);
            }
            glClientActiveTexture(GL_TEXTURE1);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
            glClientActiveTexture(GL_TEXTURE0);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);

            if (nprMode) {
                glEnable(GL_COLOR_MATERIAL);
                glColor3f(0.0, 0.0, 0.0);

                glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

                glLineWidth(2.0);
                glPointSize(2.0);

                glCullFace(GL_FRONT);

                for (int i = 0; i < nPatches; i++) {
                    if (aabbInFrustum(frust, patch[i].bb)) renderPatch(i);
                }

                glCullFace(GL_BACK);

                glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

                glColor3f(1.0, 1.0, 1.0);
                glDisable(GL_COLOR_MATERIAL);

                glDisableClientState(GL_VERTEX_ARRAY);
                glDisableClientState(GL_NORMAL_ARRAY);
            }
            break;
        }
        case ColorMix: {
            glEnableClientState(GL_VERTEX_ARRAY);
            glEnableClientState(GL_NORMAL_ARRAY);
            glEnableClientState(GL_COLOR_ARRAY);
            glVertexPointer(3, GL_FLOAT, sizeof(VertBufTer), vertBuf);
            glNormalPointer(GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].norm);
            glClientActiveTexture(GL_TEXTURE0);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            glTexCoordPointer(2, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].texCoord1);
            glClientActiveTexture(GL_TEXTURE1);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            glTexCoordPointer(2, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].texCoord2);
            glColorPointer(4, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].color);

            glEnable(GL_COLOR_MATERIAL);

            glActiveTexture(GL_TEXTURE0);
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D, textures[0]);
            glActiveTexture(GL_TEXTURE1);
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D, textures[1]);

            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
            glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
            glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
            glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB, GL_PRIMARY_COLOR);
            glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
            glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
            glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB, GL_SRC_COLOR);
            glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_INTERPOLATE);

            glActiveTexture(GL_TEXTURE0);

            for (int i = 0; i < nPatches; i++) {
                if (aabbInFrustum(frust, patch[i].bb)) renderPatch(i);
            }

            glActiveTexture(GL_TEXTURE1);
            glDisable(GL_TEXTURE_2D);
            glActiveTexture(GL_TEXTURE0);
            glDisable(GL_TEXTURE_2D);
            glDisableClientState(GL_VERTEX_ARRAY);
            glDisableClientState(GL_NORMAL_ARRAY);
            glDisableClientState(GL_COLOR_ARRAY);
            glClientActiveTexture(GL_TEXTURE1);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
            glClientActiveTexture(GL_TEXTURE0);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);

            glDisable(GL_COLOR_MATERIAL);
            break;
        }
        case MixTex: {
            glEnableClientState(GL_VERTEX_ARRAY);
            glEnableClientState(GL_NORMAL_ARRAY);
            glVertexPointer(3, GL_FLOAT, sizeof(VertBufTer), vertBuf);
            glNormalPointer(GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].norm);
            glClientActiveTexture(GL_TEXTURE0);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            glTexCoordPointer(2, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].texCoord1);
            glClientActiveTexture(GL_TEXTURE1);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            glTexCoordPointer(2, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].texCoord2);
            glClientActiveTexture(GL_TEXTURE2);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            glTexCoordPointer(2, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].texCoord3);

            glActiveTexture(GL_TEXTURE0);
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D, textures[0]);
            glActiveTexture(GL_TEXTURE1);
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D, textures[1]);

            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);

            glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_TEXTURE);
            glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
            glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_REPLACE);
            glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_ALPHA, GL_PREVIOUS);
            glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA, GL_SRC_ALPHA);
            glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA, GL_REPLACE);

            glActiveTexture(GL_TEXTURE2);
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D, textures[2]);

            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);

            glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
            glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
            glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE2_RGB, GL_PREVIOUS);
            glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
            glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
            glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND2_RGB, GL_SRC_ALPHA);
            glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_INTERPOLATE);


            glActiveTexture(GL_TEXTURE0);

            for (int i = 0; i < nPatches; i++) {
                if (aabbInFrustum(frust, patch[i].bb)) renderPatch(i);
            }

            //XXX niet disable(GL_TEXTURE_2D voor tex2?
            glActiveTexture(GL_TEXTURE1);
            glDisable(GL_TEXTURE_2D);
            glActiveTexture(GL_TEXTURE0);
            glDisable(GL_TEXTURE_2D);
            glDisableClientState(GL_VERTEX_ARRAY);
            glDisableClientState(GL_NORMAL_ARRAY);
            glClientActiveTexture(GL_TEXTURE2);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
            glClientActiveTexture(GL_TEXTURE1);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
            glClientActiveTexture(GL_TEXTURE0);
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
            break;
        }
    }
}

float Terrain::getGroundHeight(const Vector2f& pos) const
{
    int xTile = (int)floor((pos[0] - this->pos.x()) / xSegSize);
    int zTile = (int)floor(-(pos[1] - this->pos.z()) / zSegSize);
    float relx = pos[0] - this->pos.x() - xTile * xSegSize;
    float relz = -(pos[1] - this->pos.z()) - zTile * zSegSize;
    float u = relx / xSegSize;
    float v = relz / zSegSize;
    if (u >= v) {
        return (1.0 - u) * getHeight(xTile, zTile) + (u - v) * getHeight(xTile + 1, zTile) + v * getHeight(xTile + 1, zTile + 1);
    } else {
        return (1.0 - v) * getHeight(xTile, zTile) + (v - u) * getHeight(xTile, zTile + 1) + u * getHeight(xTile + 1, zTile + 1);
    }
}

Plane Terrain::getGroundPlane(const Vector2f& pos) const
{
    int xTile = (int)floor((pos[0] - this->pos.x()) / xSegSize);
    int zTile = (int)floor(-(pos[1] - this->pos.z()) / zSegSize);
    float relx = pos[0] - this->pos.x() - xTile * xSegSize;
    float relz = -(pos[1] - this->pos.z()) - zTile * zSegSize;
    float u = relx / xSegSize;
    float v = relz / zSegSize;
    if (pos[0] < this->pos.x() || pos[1] > this->pos.z() || pos[0] > this->pos.x() + xSegs * xSegSize || pos[1] < this->pos.z() - zSegs * zSegSize) {
        return Plane(0.0, 1.0, 0.0, -minHeight);
    }
    if (u >= v) {
        Vector3f norm = Vector3f(xSegSize, getHeight(xTile + 1, zTile) - getHeight(xTile, zTile), 0.0).cross(Vector3f(xSegSize, getHeight(xTile + 1, zTile + 1) - getHeight(xTile, zTile), -zSegSize)).normalized();
        return Plane(norm.x(), norm.y(), norm.z(), -norm.dot(Vector3f(pos[0], getGroundHeight(pos), pos[1])));
    } else {
        Vector3f norm = Vector3f(xSegSize, getHeight(xTile + 1, zTile + 1) - getHeight(xTile, zTile), -zSegSize).cross(Vector3f(0.0, getHeight(xTile, zTile + 1) - getHeight(xTile, zTile), -zSegSize)).normalized();
        return Plane(norm.x(), norm.y(), norm.z(), -norm.dot(Vector3f(pos[0], getGroundHeight(pos), pos[1])));
    }
}

float Terrain::getHeight(int x, int z) const
{
    if (x < 0 || z < 0 || x > xSegs || z > zSegs) return minHeight;
    return height[z * (xSegs + 1) + x];
}

Vector3f Terrain::getNormal(int x, int z) const
{
    if (x < 0 || z < 0 || x > xSegs || z > zSegs) return Vector3f(0.0, 1.0, 0.0);
    return normal[z * (xSegs + 1) + x];
}

GLuint Terrain::getTex(int t) const
{
    if (t < 0 || t > 2) return 100;
    return textures[t];
}

Vector2f Terrain::getTexScale(int t) const
{
    if (t < 0 || t > 2) return Vector2f(1.0, 1.0);
    return texScale[t];
}

Color Terrain::getColor(int x, int z) const
{
    if (x < 0 || z < 0 || x > xSegs || z > zSegs) return Color( 1.0, 1.0, 1.0, 1.0 );
    return color[z * (xSegs + 1) + x];
}

void Terrain::renderPatch(int n)
{
    for (int i = 0; i < patch[n].nStrips; i++) {
        glDrawElements(GL_TRIANGLE_STRIP, patch[n].indices[i].size(), GL_UNSIGNED_INT, &patch[n].indices[i][0]);
    }
}

SkyCube::SkyCube(float size)
    : pos(0.0, 0.0, 0.0),
    size(size)
{
    for (int i = 0; i < 5; i++) {
        textures[i] = 0;
    }
}

SkyCube::SkyCube()
{
}

void SkyCube::setTexture(int tex, GLuint texId)
{
    if (tex < 0 || tex > 4) return;
    textures[tex] = texId;
}

void SkyCube::setSize(float size)
{
    this->size = size;
}

void SkyCube::loadTex(int tex, const QString& texName)
{
    if (tex >= 0 && tex <= 4) {
        if (!loadTexture(texName, textures[tex], glcontext)) {
            textures[tex] = 0;
            return;
        }

        glBindTexture(GL_TEXTURE_2D, textures[tex]);

        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    }
}

void SkyCube::position(const Vector3f& pos) {
    this->pos = pos;
}

void SkyCube::render()
{
    glNormal3f(lightDir[0], lightDir[1], lightDir[2]);
    glEnable(GL_TEXTURE_2D);

    // The top
    glBindTexture(GL_TEXTURE_2D, textures[SKY_TOP]);
    glBegin(GL_QUADS);
        glTexCoord2f(0.0, 1.0);
        glVertex3f(pos.x() - size, pos.y() + size, pos.z() + size);
        glTexCoord2f(0.0, 0.0);
        glVertex3f(pos.x() - size, pos.y() + size, pos.z() - size);
        glTexCoord2f(1.0, 0.0);
        glVertex3f(pos.x() + size, pos.y() + size, pos.z() - size);
        glTexCoord2f(1.0, 1.0);
        glVertex3f(pos.x() + size, pos.y() + size, pos.z() + size);
    glEnd();

    // The front
    glBindTexture(GL_TEXTURE_2D, textures[SKY_FRONT]);
    glBegin(GL_QUADS);
        glTexCoord2f(0.0, 1.0);
        glVertex3f(pos.x() - size, pos.y() + size, pos.z() - size);
        glTexCoord2f(0.0, 0.0);
        glVertex3f(pos.x() - size, pos.y() - size, pos.z() - size);
        glTexCoord2f(1.0, 0.0);
        glVertex3f(pos.x() + size, pos.y() - size, pos.z() - size);
        glTexCoord2f(1.0, 1.0);
        glVertex3f(pos.x() + size, pos.y() + size, pos.z() - size);
    glEnd();

    // The back
    glBindTexture(GL_TEXTURE_2D, textures[SKY_BACK]);
        glBegin(GL_QUADS);
        glTexCoord2f(0.0, 1.0);
        glVertex3f(pos.x() + size, pos.y() + size, pos.z() + size);
        glTexCoord2f(0.0, 0.0);
        glVertex3f(pos.x() + size, pos.y() - size, pos.z() + size);
        glTexCoord2f(1.0, 0.0);
        glVertex3f(pos.x() - size, pos.y() - size, pos.z() + size);
        glTexCoord2f(1.0, 1.0);
        glVertex3f(pos.x() - size, pos.y() + size, pos.z() + size);
    glEnd();

    // The left
    glBindTexture(GL_TEXTURE_2D, textures[SKY_LEFT]);
    glBegin(GL_QUADS);
        glTexCoord2f(0.0, 1.0);
        glVertex3f(pos.x() - size, pos.y() + size, pos.z() + size);
        glTexCoord2f(0.0, 0.0);
        glVertex3f(pos.x() - size, pos.y() - size, pos.z() + size);
        glTexCoord2f(1.0, 0.0);
        glVertex3f(pos.x() - size, pos.y() - size, pos.z() - size);
        glTexCoord2f(1.0, 1.0);
        glVertex3f(pos.x() - size, pos.y() + size, pos.z() - size);
    glEnd();

    // The right
    glBindTexture(GL_TEXTURE_2D, textures[SKY_RIGHT]);
    glBegin(GL_QUADS);
        glTexCoord2f(0.0, 1.0);
        glVertex3f(pos.x() + size, pos.y() + size, pos.z() - size);
        glTexCoord2f(0.0, 0.0);
        glVertex3f(pos.x() + size, pos.y() - size, pos.z() - size);
        glTexCoord2f(1.0, 0.0);
        glVertex3f(pos.x() + size, pos.y() - size, pos.z() + size);
        glTexCoord2f(1.0, 1.0);
        glVertex3f(pos.x() + size, pos.y() + size, pos.z() + size);
    glEnd();

    glEnable(GL_DEPTH_TEST);
    glDepthMask(GL_TRUE);
}

GLuint SkyCube::getTexture(int tex)
{
    if (tex < 0 || tex > 4) return 0;
    return textures[tex];
}

TerrainFinish::TerrainFinish(Terrain* terrain, float height, int splitCount, int tileCount, int xDistCount, int zDistCount, Color color1, Color color2)
    : height(height)
    , xDistCount(xDistCount)
    , zDistCount(zDistCount)
    , terrain(terrain)
    , splitCount(splitCount)
    , tileCount(tileCount)
{
    xSegSize = ((terrain->xSegs + 2 * xDistCount) * terrain->xSegSize) / splitCount;
    zSegSize = ((terrain->zSegs + 2 * zDistCount) * terrain->zSegSize) / splitCount;

    nPatches = 12;
    nVertices = 4 * (terrain->xSegs + terrain->zSegs + 2);
    nVertices += (2 * tileCount + splitCount + 1) * (tileCount + 1) * 2;
    nVertices += (splitCount - 1) * 2 * (tileCount + 1);

    vertBuf = new VertBufTer[nVertices];
    patch.resize(nPatches);

    int vCount = 0;

    // Terrain vertices
    for (int i = 0; i <= terrain->zSegs; i++) {
        vertBuf[vCount].pos = Vector3f(terrain->pos.x(), terrain->getHeight(0, i), terrain->pos.z() - i * terrain->zSegSize);
        vertBuf[vCount].norm = terrain->getNormal(0, i);
        vertBuf[vCount].color = terrain->getColor(0, i);
        vCount++;
        vertBuf[vCount].pos = Vector3f(terrain->pos.x() + terrain->xSegs * terrain->xSegSize, terrain->getHeight(terrain->xSegs, i), terrain->pos.z() - i * terrain->zSegSize);
        vertBuf[vCount].norm = terrain->getNormal(terrain->xSegs, i);
        vertBuf[vCount].color = terrain->getColor(terrain->xSegs, i);
        vCount++;
    }
    for (int i = 1; i < terrain->xSegs; i++) {
        vertBuf[vCount].pos = Vector3f(terrain->pos.x() + i * terrain->xSegSize, terrain->getHeight(i, 0), terrain->pos.z());
        vertBuf[vCount].norm = terrain->getNormal(i, 0);
        vertBuf[vCount].color = terrain->getColor(i, 0);
        vCount++;
        vertBuf[vCount].pos = Vector3f(terrain->pos.x() + i * terrain->xSegSize, terrain->getHeight(i, terrain->zSegs), terrain->pos.z() - terrain->zSegs * terrain->zSegSize);
        vertBuf[vCount].norm = terrain->getNormal(i, terrain->zSegs);
        vertBuf[vCount].color = terrain->getColor(i, terrain->zSegs);
        vCount++;
    }

    int finishBegin = vCount;

    // Finishers
    for (int i = -1; i <= terrain->zSegs + 1; i++) {
        int extra;
        if (i == -1) extra = -zDistCount + 1;
        else if (i == terrain->zSegs + 1) extra = zDistCount - 1;
        else extra = 0;
        vertBuf[vCount].pos = Vector3f(terrain->pos.x() - terrain->xSegSize * xDistCount, height, terrain->pos.z() - (i + extra) * terrain->zSegSize);
        vertBuf[vCount].norm = Vector3f(0.0, 1.0, 0.0);
        vertBuf[vCount].color = color1;
        vCount++;
        vertBuf[vCount].pos = Vector3f(terrain->pos.x() + (terrain->xSegs + xDistCount) * terrain->xSegSize, height, terrain->pos.z() - (i + extra) * terrain->zSegSize);
        vertBuf[vCount].norm = Vector3f(0.0, 1.0, 0.0);
        vertBuf[vCount].color = color1;
        vCount++;
    }
    for (int i = 0; i <= terrain->xSegs; i++) {
        vertBuf[vCount].pos = Vector3f(terrain->pos.x() + i * terrain->xSegSize, height, terrain->pos.z() + terrain->zSegSize * zDistCount);
        vertBuf[vCount].norm = Vector3f(0.0, 1.0, 0.0);
        vertBuf[vCount].color = color1;
        vCount++;
        vertBuf[vCount].pos = Vector3f(terrain->pos.x() + i * terrain->xSegSize, height, terrain->pos.z() - (terrain->zSegs + zDistCount) * terrain->zSegSize);
        vertBuf[vCount].norm = Vector3f(0.0, 1.0, 0.0);
        vertBuf[vCount].color = color1;
        vCount++;
    }

    int tileBegin = vCount;
    int tileBegin2 = tileBegin;

    // The large tiles
    float posx = terrain->pos.x() - terrain->xSegSize * xDistCount - xSegSize * tileCount;
    float posz = terrain->pos.z() + terrain->zSegSize * zDistCount + zSegSize * tileCount;

    for (int i = 0; i <= 2 * tileCount + splitCount; i++) {
        for (int j = 0; j <= 2 * tileCount + splitCount; j++) {
            if (i <= tileCount || i >= tileCount + splitCount || j <= tileCount || j >= tileCount + splitCount) {
                vertBuf[vCount].pos = Vector3f(posx + i * xSegSize, height, posz - j * zSegSize);
                vertBuf[vCount].norm = Vector3f(0.0, 1.0, 0.0);
                if (i < tileCount || j < tileCount || i > tileCount + splitCount || j > tileCount + splitCount) {
                    float r = frandom();
                    vertBuf[vCount].color.r = color1.r * r + (1 - r) * color2.r;
                    vertBuf[vCount].color.g = color1.g * r + (1 - r) * color2.g;
                    vertBuf[vCount].color.b = color1.b * r + (1 - r) * color2.b;
                    vertBuf[vCount].color.a = color1.a * r + (1 - r) * color2.a;
                } else {
                    vertBuf[vCount].color = color1;
                }
                vCount++;
            }
        }
    }
    printf("vCount = %d, nVertices = %d\n", vCount, nVertices);

    // And now fill the patches...
    int rows;
    // Patch 0
    patch[0].bb = AABB(Vector3f(posx, height, posz), Vector3f(terrain->pos.x() - terrain->xSegSize, height, terrain->pos.z() + terrain->zSegSize));
    patch[0].nStrips = tileCount;
    patch[0].indices.resize(patch[0].nStrips);
    rows = 2 * tileCount + splitCount + 1;
    for (int i = 0; i < patch[0].nStrips; i++) {
        patch[0].indices[i].resize((tileCount + 1) * 2);
        int index = 0;
        for (int j = 0; j <= tileCount; j++) {
            patch[0].indices[i][index++] = tileBegin + j * rows + i + 1;
            patch[0].indices[i][index++] = tileBegin + j * rows + i;
        }
    }

    // Patch 1
    tileBegin = tileBegin + tileCount + splitCount;
    posx = terrain->pos.x() - terrain->xSegSize * xDistCount - xSegSize * tileCount;
    posz = terrain->pos.z() - terrain->zSegSize * terrain->zSegs - terrain->zSegSize * zDistCount;

    patch[1].bb = AABB(Vector3f(posx, height, posz), Vector3f(posx + xSegSize * tileCount, height, posz - zSegSize * tileCount));
    patch[1].nStrips = tileCount;
    patch[1].indices.resize(patch[1].nStrips);
    rows = 2 * tileCount + splitCount + 1;
    for (int i = 0; i < patch[1].nStrips; i++) {
        patch[1].indices[i].resize((tileCount + 1) * 2);
        int index = 0;
        for (int j = 0; j <= tileCount; j++) {
            patch[1].indices[i][index++] = tileBegin + j * rows + i + 1;
            patch[1].indices[i][index++] = tileBegin + j * rows + i;
        }
    }

    // Patch 2
    tileBegin = (tileBegin - tileCount - splitCount) + rows * (tileCount + 1) + (splitCount - 1) * 2 * (tileCount + 1);
    posx = terrain->pos.x() + terrain->xSegSize * terrain->xSegs + terrain->xSegSize * xDistCount;
    posz = terrain->pos.z() + terrain->zSegSize * zDistCount + zSegSize * tileCount; //XXX was xDistCount

    patch[2].bb = AABB(Vector3f(posx, height, posz), Vector3f(posx + xSegSize * tileCount, height, posz - zSegSize * tileCount));
    patch[2].nStrips = tileCount;
    patch[2].indices.resize(patch[2].nStrips);
    rows = 2 * tileCount + splitCount + 1;
    for (int i = 0; i < patch[2].nStrips; i++) {
        patch[2].indices[i].resize((tileCount + 1) * 2);
        int index = 0;
        for (int j = 0; j <= tileCount; j++) {
            patch[2].indices[i][index++] = tileBegin + j * rows + i + 1;
            patch[2].indices[i][index++] = tileBegin + j * rows + i;
        }
    }

    // Patch 3
    tileBegin = tileBegin + tileCount + splitCount;
    posx = terrain->pos.x() + terrain->xSegSize * terrain->xSegs + terrain->xSegSize * xDistCount;
    posz = terrain->pos.z() - terrain->zSegSize * terrain->zSegs - terrain->zSegSize * zDistCount;

    patch[3].bb = AABB(Vector3f(posx, height, posz), Vector3f(posx + xSegSize * tileCount, height, posz - zSegSize * tileCount));
    patch[3].nStrips = tileCount;
    patch[3].indices.resize(patch[3].nStrips);
    rows = 2 * tileCount + splitCount + 1;
    for (int i = 0; i < patch[3].nStrips; i++) {
        patch[3].indices[i].resize((tileCount + 1) * 2);
        int index = 0;
        for (int j = 0; j <= tileCount; j++) {
            patch[3].indices[i][index++] = tileBegin + j * rows + i + 1;
            patch[3].indices[i][index++] = tileBegin + j * rows + i;
        }
    }

    // Patch 4
    tileBegin = tileBegin2 + tileCount;
    posx = terrain->pos.x() - terrain->xSegSize * xDistCount - xSegSize * tileCount;
    posz = terrain->pos.z() + terrain->zSegSize * zDistCount;

    patch[4].bb = AABB(Vector3f(posx, height, posz), Vector3f(posx + xSegSize * tileCount, height, posz - zSegSize * tileCount));
    patch[4].nStrips = splitCount;
    patch[4].indices.resize(patch[4].nStrips);
    rows = 2 * tileCount + splitCount + 1;
    for (int i = 0; i < patch[4].nStrips; i++) {
        patch[4].indices[i].resize((tileCount + 1) * 2);
        int index = 0;
        for (int j = 0; j <= tileCount; j++) {
            patch[4].indices[i][index++] = tileBegin + j * rows + i + 1;
            patch[4].indices[i][index++] = tileBegin + j * rows + i;
        }
    }

    // Patch 5
    tileBegin = tileBegin2 + tileCount + rows * (tileCount + 1) + (splitCount - 1) * 2 * (tileCount + 1);
    posx = terrain->pos.x() + terrain->xSegSize * terrain->xSegs + terrain->xSegSize * xDistCount;
    posz = terrain->pos.z() + terrain->zSegSize * zDistCount;

    patch[5].bb = AABB(Vector3f(posx, height, posz), Vector3f(posx + xSegSize * tileCount, height, posz - zSegSize * tileCount));
    patch[5].nStrips = splitCount;
    patch[5].indices.resize(patch[5].nStrips);
    rows = 2 * tileCount + splitCount + 1;
    for (int i = 0; i < patch[5].nStrips; i++) {
        patch[5].indices[i].resize((tileCount + 1) * 2);
        int index = 0;
        for (int j = 0; j <= tileCount; j++) {
            patch[5].indices[i][index++] = tileBegin + j * rows + i + 1;
            patch[5].indices[i][index++] = tileBegin + j * rows + i;
        }
    }

    // Patch 6
    tileBegin = tileBegin2 + rows * tileCount;
    posx = terrain->pos.x() - terrain->xSegSize * xDistCount;
    posz = terrain->pos.z() + terrain->zSegSize * zDistCount + zSegSize * tileCount;

    patch[6].bb = AABB(Vector3f(posx, height, posz), Vector3f(posx + xSegSize * tileCount, height, posz - zSegSize * tileCount));
    patch[6].nStrips = tileCount;
    patch[6].indices.resize(patch[6].nStrips);
    rows = 2 * tileCount + splitCount + 1;
    for (int i = 0; i < patch[6].nStrips; i++) {
        patch[6].indices[i].resize((splitCount + 1) * 2);
        int index = 0;
        for (int j = 0; j <= splitCount; j++) {
            int extra = 0;
            if (j > 0) extra = (j - 1) * (splitCount - 1);
            patch[6].indices[i][index++] = tileBegin + j * rows - extra + i + 1;
            patch[6].indices[i][index++] = tileBegin + j * rows - extra + i;
        }
    }

    // Patch 7
    tileBegin = tileBegin + tileCount + splitCount;
    posx = terrain->pos.x() - terrain->xSegSize * xDistCount;
    posz = terrain->pos.z() - terrain->zSegSize * terrain->zSegs - terrain->zSegSize * zDistCount;

    patch[7].bb = AABB(Vector3f(posx, height, posz), Vector3f(posx + xSegSize * tileCount, height, posz - zSegSize * tileCount));
    patch[7].nStrips = tileCount;
    patch[7].indices.resize(patch[7].nStrips);
    rows = 2 * tileCount + splitCount + 1;
    for (int i = 0; i < patch[7].nStrips; i++) {
        patch[7].indices[i].resize((splitCount + 1) * 2);
        int index = 0;
        for (int j = 0; j <= splitCount; j++) {
            int extra;
            if (j < splitCount) extra = j * (splitCount - 1);
            else extra = (j - 1) * (splitCount - 1);
            patch[7].indices[i][index++] = tileBegin + j * rows - extra + i + 1;
            patch[7].indices[i][index++] = tileBegin + j * rows - extra + i;
        }
    }

    // Patch 8
    float minh = height;
    float maxh = height;
    for (int i = 0; i <= terrain->zSegs; i++) {
        float h = terrain->getHeight(0, i);
        if (h < minh) minh = h;
        if (h > maxh) maxh = h;
    }

    patch[8].bb = AABB(Vector3f(terrain->pos.x() - terrain->xSegSize * xDistCount, minh, terrain->pos.z() + terrain->zSegSize * zDistCount),
            Vector3f(terrain->pos.x(), maxh, terrain->pos.z() - terrain->zSegSize * (zDistCount + terrain->zSegs)));
    patch[8].nStrips = 1;
    patch[8].indices.resize(patch[8].nStrips);
    patch[8].indices[0].resize(2 * (terrain->zSegs + 3));
    patch[8].indices[0][0] = finishBegin;
    patch[8].indices[0][1] = finishBegin + 2 * (terrain->zSegs + 3);
    int index = 2;
    for (int i = 0; i <= terrain->zSegs; i++) {
        patch[8].indices[0][index++] = finishBegin + 2 * (i + 1);
        patch[8].indices[0][index++] = 2 * i;
    }
    patch[8].indices[0][index++] = finishBegin + 2 * (terrain->zSegs + 2);
    patch[8].indices[0][index++] = finishBegin + 2 * (terrain->zSegs + 3) + 1;

    // Patch 9
    minh = height;
    maxh = height;
    for (int i = 0; i <= terrain->zSegs; i++) {
        float h = terrain->getHeight(terrain->xSegs, i);
        if (h < minh) minh = h;
        if (h > maxh) maxh = h;
    }

    patch[9].bb = AABB(Vector3f(terrain->pos.x() + terrain->xSegSize * terrain->xSegs, minh, terrain->pos.z() + terrain->zSegSize * zDistCount),
            Vector3f(terrain->pos.x() + terrain->xSegSize * (terrain->xSegs + xDistCount), maxh, terrain->pos.z() - terrain->zSegSize * (zDistCount + terrain->zSegs)));
    patch[9].nStrips = 1;
    patch[9].indices.resize(patch[9].nStrips);
    patch[9].indices[0].resize(2 * (terrain->zSegs + 3));
    patch[9].indices[0][0] = finishBegin + 2 * (terrain->zSegs + 3) + 2 * terrain->xSegs;
    patch[9].indices[0][1] = finishBegin + 1;
    index = 2;
    for (int i = 0; i <= terrain->zSegs; i++) {
        patch[9].indices[0][index++] = 2 * i + 1;
        patch[9].indices[0][index++] = finishBegin + 2 * (i + 1) + 1;
    }
    patch[9].indices[0][index++] = finishBegin + 2 * (terrain->zSegs + 3) + 2 * terrain->xSegs + 1;
    patch[9].indices[0][index++] = finishBegin + 2 * (terrain->zSegs + 3) - 1;

    // Patch 10
    minh = height;
    maxh = height;
    for (int i = 0; i <= terrain->xSegs; i++) {
        float h = terrain->getHeight(i, 0);
        if (h < minh) minh = h;
        if (h > maxh) maxh = h;
    }

    patch[10].bb = AABB(Vector3f(terrain->pos.x() - terrain->xSegSize * xDistCount, minh, terrain->pos.z() + terrain->zSegSize * zDistCount),
                Vector3f(terrain->pos.x() + terrain->xSegSize * (terrain->xSegs + xDistCount), maxh, terrain->pos.z()));
    patch[10].nStrips = 1;
    patch[10].indices.resize(patch[10].nStrips);
    patch[10].indices[0].resize(2 * (terrain->xSegs + 1));
    patch[10].indices[0][0] = 0;
    patch[10].indices[0][1] = finishBegin + 2 * (terrain->zSegs + 3);
    index = 2;
    for (int i = 1; i < terrain->xSegs; i++) {
        patch[10].indices[0][index++] = 2 * (terrain->zSegs + 1) + 2 * (i - 1);
        patch[10].indices[0][index++] = finishBegin + 2 * (terrain->zSegs + 3 + i);
    }
    patch[10].indices[0][index++] = 1;
    patch[10].indices[0][index++] = finishBegin + 2 * (terrain->zSegs + terrain->xSegs + 3);

    // Patch 11
    minh = height;
    maxh = height;
    for (int i = 0; i <= terrain->xSegs; i++) {
        float h = terrain->getHeight(i, terrain->zSegs);
        if (h < minh) minh = h;
        if (h > maxh) maxh = h;
    }

    patch[11].bb = AABB(Vector3f(terrain->pos.x() - terrain->xSegSize * xDistCount, minh, terrain->pos.z() - terrain->zSegSize * terrain->zSegs),
                Vector3f(terrain->pos.x() + terrain->xSegSize * (terrain->xSegs + xDistCount), maxh, terrain->pos.z() - terrain->zSegSize * (zDistCount + terrain->zSegs)));
    patch[11].nStrips = 1;
    patch[11].indices.resize(patch[11].nStrips);
    patch[11].indices[0].resize(2 * (terrain->xSegs + 1));
    patch[11].indices[0][0] = finishBegin + 2 * (terrain->zSegs + 3) + 1;
    patch[11].indices[0][1] = 2 * terrain->zSegs;
    index = 2;
    for (int i = 1; i < terrain->xSegs; i++) {
        patch[11].indices[0][index++] = finishBegin + 2 * (terrain->zSegs + 3 + i) + 1;
        patch[11].indices[0][index++] = 2 * (terrain->zSegs + 1) + 2 * (i - 1) + 1;
    }
    patch[11].indices[0][index++] = finishBegin + 2 * (terrain->zSegs + terrain->xSegs + 3) + 1;
    patch[11].indices[0][index++] = 2 * terrain->zSegs + 1;
}

TerrainFinish::TerrainFinish()
{
}

TerrainFinish::~TerrainFinish()
{
}

void TerrainFinish::render()
{
    Frustum frust = Frustum::calc();

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glVertexPointer(3, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].pos);
    glNormalPointer(GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].norm);
    glColorPointer(4, GL_FLOAT, sizeof(VertBufTer), &vertBuf[0].color);

    glEnable(GL_COLOR_MATERIAL);

    for (int i = 0; i < nPatches; i++) {
        if (aabbInFrustum(frust, patch[i].bb)) renderPatch(i);
    }

    glDisable(GL_TEXTURE_2D); //XXX geen texture geenabled?
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);

    glDisable(GL_COLOR_MATERIAL);
}

void TerrainFinish::setEdgeColor(const QString& texFile)
{
    QImage img(texFile);
    if (img.isNull()) {
        fprintf(stderr, "Error loading image %s\n", qPrintable(texFile));
        return;
    }

    for (int j = 0; j <= terrain->zSegs; j++) {
        float au = 0.0;
        float av = (float)j / terrain->zSegs;
        int ax = 0;
        int ay = (int)floor(av * (img.height() - 1));
        float xpart = 0.0;
        float ypart = av * (img.height() - 1) - ay;
        Color val, val1, val2;
        val1 = Color(
            qRed(img.pixel(ax, ay)) / 255.0,
            qGreen(img.pixel(ax, ay)) / 255.0,
            qBlue(img.pixel(ax, ay)) / 255.0,
            1.0 );
        if (ypart > 0) {
            val2 = Color(
                qRed(img.pixel(ax, ay+1)) / 255.0,
                qGreen(img.pixel(ax, ay+1)) / 255.0,
                qBlue(img.pixel(ax, ay+1)) / 255.0,
                1.0 );
            val.r = (1.0 - ypart) * val1.r + ypart * val2.r;
            val.g = (1.0 - ypart) * val1.g + ypart * val2.g;
            val.b = (1.0 - ypart) * val1.b + ypart * val2.b;
            val.a = 1.0;
        } else {
            val = val1;
        }
        vertBuf[2 * j].color = val;

        au = 1.0;
        av = (float)j / terrain->zSegs;
        ax = img.width() - 1;
        ay = (int)floor(av * (img.height() -1));
        xpart = 0.0;
        ypart = av * (img.height() - 1) - ay;
        val1 = Color(
            qRed(img.pixel(ax, ay)) / 255.0,
            qGreen(img.pixel(ax, ay)) / 255.0,
            qBlue(img.pixel(ax, ay)) / 255.0,
            1.0 );
        if (ypart > 0) {
            val2 = Color(
                qRed(img.pixel(ax, ay+1)) / 255.0,
                qGreen(img.pixel(ax, ay+1)) / 255.0,
                qBlue(img.pixel(ax, ay+1)) / 255.0,
                1.0 );
            val.r = (1.0 - ypart) * val1.r + ypart * val2.r;
            val.g = (1.0 - ypart) * val1.g + ypart * val2.g;
            val.b = (1.0 - ypart) * val1.b + ypart * val2.b;
            val.a = 1.0;
        } else {
            val = val1;
        }
        vertBuf[2 * j + 1].color = val;
    }

    for (int i = 1; i < terrain->xSegs; i++) {
        float au = (float)i / terrain->xSegs;
        float av = 0.0;
        int ax = (int)floor(au * (img.width() - 1));
        int ay = 0;
        float xpart = au * (img.width() - 1) - ax;
        float ypart = 0.0;
        Color val;
        if (xpart > 0) {
            val = Color(
                ((1.0 - xpart) * qRed(img.pixel(ax, ay)) + xpart * qRed(img.pixel(ax+1, ay))) / 255.0,
                ((1.0 - xpart) * qGreen(img.pixel(ax, ay)) + xpart * qGreen(img.pixel(ax+1, ay))) / 255.0,
                ((1.0 - xpart) * qBlue(img.pixel(ax, ay)) + xpart * qBlue(img.pixel(ax+1, ay))) / 255.0,
                1.0 );
        } else {
            val = Color(
                qRed(img.pixel(ax, ay)) / 255.0,
                qGreen(img.pixel(ax, ay)) / 255.0,
                qBlue(img.pixel(ax, ay)) / 255.0,
                1.0 );
        }
        vertBuf[2 * (terrain->zSegs + i)].color = val;

        au = (float)i / terrain->xSegs;
        av = 1.0;
        ax = (int)floor(au * (img.width() - 1));
        ay = img.height() - 1;
        xpart = au * (img.width() - 1) - ax;
        ypart = 0.0;
        if (xpart > 0) {
            val = Color(
                ((1.0 - xpart) * qRed(img.pixel(ax, ay)) + xpart * qRed(img.pixel(ax+1, ay))) / 255.0,
                ((1.0 - xpart) * qGreen(img.pixel(ax, ay)) + xpart * qGreen(img.pixel(ax+1, ay))) / 255.0,
                ((1.0 - xpart) * qBlue(img.pixel(ax, ay)) + xpart * qBlue(img.pixel(ax+1, ay))) / 255.0,
                1.0 );
        } else {
            val = Color(
                qRed(img.pixel(ax, ay)) / 255.0,
                qGreen(img.pixel(ax, ay)) / 255.0,
                qBlue(img.pixel(ax, ay)) / 255.0,
                1.0 );
        }
        vertBuf[2 * (terrain->zSegs + i) + 1].color = val;
    }
}

void TerrainFinish::calcEdgeSea(Terrain* ter, float maxDif)
{
    if (!ter) return;

    float posx, posz, hdif, val;
    for (int i = 0; i < terrain->zSegs; i++) {
        posx = vertBuf[2 * i].pos.x();
        posz = vertBuf[2 * i].pos.z();
        hdif = vertBuf[2 * i].pos.y() - ter->getGroundHeight(Vector2f(posx, posz));
        val = qMax(hdif / maxDif, 0.0f);
        val = powf(val, 0.33);
        vertBuf[2 * i].color.r = 1.0 - sqrt(val);
        vertBuf[2 * i].color.g = 1.0 - val;
        vertBuf[2 * i].color.b = 1.0 - val * val * 0.75;
        vertBuf[2 * i].color.a = 0.3 + sqrt(val) * 0.7;

        posx = vertBuf[2 * i + 1].pos.x();
        posz = vertBuf[2 * i + 1].pos.z();
        hdif = vertBuf[2 * i + 1].pos.y() - ter->getGroundHeight(Vector2f(posx, posz));
        val = qMax(hdif / maxDif, 0.0f);
        val = powf(val, 0.33);
        vertBuf[2 * i + 1].color.r = 1.0 - sqrt(val);
        vertBuf[2 * i + 1].color.g = 1.0 - val;
        vertBuf[2 * i + 1].color.b = 1.0 - val * val * 0.75;
        vertBuf[2 * i + 1].color.a = 0.3 + sqrt(val) * 0.7;
    }

    for (int j = 0; j < terrain->xSegs; j++) {
        posx = vertBuf[2 * (terrain->zSegs + j)].pos.x();
        posz = vertBuf[2 * (terrain->zSegs + j)].pos.z();
        hdif = vertBuf[2 * (terrain->zSegs + j)].pos.y() - ter->getGroundHeight(Vector2f(posx, posz));
        val = qMax(hdif / maxDif, 0.0f);
        val = powf(val, 0.33);
        vertBuf[2 * (terrain->zSegs + j)].color.r = 1.0 - sqrt(val);
        vertBuf[2 * (terrain->zSegs + j)].color.g = 1.0 - val;
        vertBuf[2 * (terrain->zSegs + j)].color.b = 1.0 - val * val * 0.75;
        vertBuf[2 * (terrain->zSegs + j)].color.a = 0.3 + sqrt(val) * 0.7;

        posx = vertBuf[2 * (terrain->zSegs + j) + 1].pos.x();
        posz = vertBuf[2 * (terrain->zSegs + j) + 1].pos.z();
        hdif = vertBuf[2 * (terrain->zSegs + j) + 1].pos.y() - ter->getGroundHeight(Vector2f(posx, posz));
        val = qMax(hdif / maxDif, 0.0f);
        val = powf(val, 0.33);
        vertBuf[2 * (terrain->zSegs + j) + 1].color.r = 1.0 - sqrt(val);
        vertBuf[2 * (terrain->zSegs + j) + 1].color.g = 1.0 - val;
        vertBuf[2 * (terrain->zSegs + j) + 1].color.b = 1.0 - val * val * 0.75;
        vertBuf[2 * (terrain->zSegs + j) + 1].color.a = 0.3 + sqrt(val) * 0.7;
    }

    for (int i = 0; i < nVertices; i++) {
        vertBuf[i].color = vertBuf[0].color; // XXX huh?
    }
}

void TerrainFinish::setOuterColor(int level, Color color)
{
    int tileBegin = 4 * (terrain->xSegs + terrain->zSegs + 2);
    if (level < tileCount) {
        for (int i = level; i <= 2 * tileCount + splitCount - level; i++) {
            vertBuf[tileBegin + getVertId(level, i)].color = color;
            vertBuf[tileBegin + getVertId(2 * tileCount + splitCount - level, i)].color = color;
            vertBuf[tileBegin + getVertId(i, level)].color = color;
            vertBuf[tileBegin + getVertId(i, 2 * tileCount + splitCount - level)].color = color;
        }
    }
}

void TerrainFinish::setCornerColor(int level, Color color)
{
    int tileBegin = 4 * (terrain->xSegs + terrain->zSegs + 2);
    if (level < tileCount) {
        vertBuf[tileBegin + getVertId(level, level)].color = color;
        vertBuf[tileBegin + getVertId(2 * tileCount + splitCount - level, level)].color = color;
        vertBuf[tileBegin + getVertId(2 * tileCount + splitCount - level, 2 * tileCount + splitCount - level)].color = color;
        vertBuf[tileBegin + getVertId(level, 2 * tileCount + splitCount - level)].color = color;
    }
}

void TerrainFinish::renderPatch(int n)
{
//	if (n <= 7) return;
    for (int i = 0; i < patch[n].nStrips; i++) {
        glDrawElements(GL_TRIANGLE_STRIP, patch[n].indices[i].size(), GL_UNSIGNED_INT, &patch[n].indices[i][0]);
    }
}

int TerrainFinish::getVertId(int x, int z)
{
    int dif = 2 * tileCount + splitCount + 1;
    if (x <= tileCount) {
        return x * dif + z;
    } else if (x >= tileCount + splitCount) {
        return x * dif + z - (splitCount - 1) * (splitCount - 1);
    } else {
        int missed;
        if (z > tileCount) missed = (x - tileCount) * (splitCount - 1);
        else missed = (x - tileCount - 1) * (splitCount - 1);
        return x * dif + z - missed;
    }
}
