/*
 * Copyright 2008 Marijn Kruisselbrink <m.kruisselbrink@student.tue.nl>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "Weapons.h"

// FruitWars headers
#include "Globals.h"
#include "Instance.h"

USING_PART_OF_NAMESPACE_EIGEN

Banana::Banana(Model3D* model, int owner)
    : inst(Instance::load(model))
    , colBS(Vector3f(), 0)
    , carriedBy(owner-1)
    , state(0)
    , throwSpeed()
    , throwTime(0.0)
    , upAcc(WEAPONBANANAUPACC)
    , breakFac(WEAPONBANANAUPBRAKE)
    , returnTime(0.0)
    , oldPos()
    , trail(8)
{
    Matrix4f m;
    m << BANANAMATRIX[0], BANANAMATRIX[1], BANANAMATRIX[2], BANANAMATRIX[3],
            BANANAMATRIX[4], BANANAMATRIX[5], BANANAMATRIX[6], BANANAMATRIX[7],
            BANANAMATRIX[8], BANANAMATRIX[9], BANANAMATRIX[10], BANANAMATRIX[11],
            BANANAMATRIX[12], BANANAMATRIX[13], BANANAMATRIX[14], BANANAMATRIX[15];
    attachMat = m;

    inst->setBS(BS(Vector3f(WEAPONBANANABSOFFSET[0], WEAPONBANANABSOFFSET[1], WEAPONBANANABSOFFSET[2]), WEAPONBANANABSRAD));
    inst->scaleObj(Vector3f(WEAPONBANANASCALE, WEAPONBANANASCALE, WEAPONBANANASCALE));
    colBS = inst->getBS();
    colBS.radius = WEAPONBANANACOLRADIUS;
    oldPos = getColBS().pos;
}

Banana::~Banana()
{
}

void Banana::update(float dt)
{
    if (state == -1) {
        state = 0;
        getIt();
        // XXX sendMessage ATTACHBANANA
    } else if (state == -2) {
        state = -1;
    } else if (state == 1) {
        spin(dt);
        throwTime -= dt;
        if (throwTime <= 0.0) {
            throwTime = 0.0;
            state = 2;
        }
    } else if (state == 2) {
        spin(dt);
        throwSpeed += Vector3f(-breakFac * throwSpeed.x(), upAcc * dt, -breakFac * throwSpeed.z());
        if (getPos().y() > myTer.getMaxHeight()) {
            state = 3;
            returnTime = qMin(WEAPONBANANARETURNMAX, (myPlayers[carriedBy].getInst()->getKeyPos(1) - getColBS().pos).norm() * WEAPONBANANARETURNFAC);
        }
    } else if (state == 3) {
        spin(dt);
        Vector3f dif = myPlayers[carriedBy].getInst()->getKeyPos(1) - getColBS().pos;
        Vector3f theAcc = (dif - (throwSpeed * returnTime)) * (2.0 / (returnTime * returnTime));
        throwSpeed += theAcc * dt;
        returnTime -= dt;
        if (returnTime <= 0.0) {
            state = 0;
            inst->specialAttach(myPlayers[carriedBy].getInst(), PLAYERRIGHTARMFRAME, attachMat);
            // Send ATTACHBANANA
        }
    }

    position(getPos() + throwSpeed * dt);

    Vector3f posA, posB;
    if (!inst->getAttach().attached) {
        posA = inst->transform() * Vector3f(WEAPONBANANATRAILOFFSET[0], WEAPONBANANATRAILOFFSET[1], WEAPONBANANATRAILOFFSET[2]);
        posB = inst->transform() * Vector3f(WEAPONBANANATRAILOFFSET[3], WEAPONBANANATRAILOFFSET[4], WEAPONBANANATRAILOFFSET[5]);
    } else {
        posA = inst->attachTransform() * Vector3f(WEAPONBANANATRAILOFFSET[0], WEAPONBANANATRAILOFFSET[1], WEAPONBANANATRAILOFFSET[2]);
        posB = inst->attachTransform() * Vector3f(WEAPONBANANATRAILOFFSET[3], WEAPONBANANATRAILOFFSET[4], WEAPONBANANATRAILOFFSET[5]);
    }

    trail.update(posA, posB);
}

void Banana::render()
{
    inst->render();
}

void Banana::trailRender()
{
    if (state > 0) trail.render();
}

void Banana::throwIt(const Vector3f& dir)
{
    if (state == 0) {
        state = 1;
        throwSpeed = dir * WEAPONBANANATHROWSPEED;
        throwTime = WEAPONBANANARANGE / WEAPONBANANATHROWSPEED;
        inst->detach();
        oldPos = getColBS().pos;
    }
}

void Banana::spin(float dt)
{
    rotateX(getRot().x() + WEAPONBANANASPINSPEED * dt);
}

void Banana::getIt()
{
    state = 0;
    inst->specialAttach(myPlayers[carriedBy].getInst(), PLAYERRIGHTARMFRAME, attachMat);
    positionBS(myPlayers[carriedBy].getInst()->getKeyPos(1));
}

void Banana::scale(const Vector3f& scale)
{
    inst->scaleObj(scale);
}

void Banana::rotate(const Vector3f& rot)
{
    inst->rotate(rot);
}

void Banana::rotateX(float ang)
{
    inst->rotateX(ang);
}

void Banana::rotateY(float ang)
{
    inst->rotateY(ang);
}

void Banana::rotateZ(float ang)
{
    inst->rotateZ(ang);
}

void Banana::position(const Vector3f& pos)
{
    inst->position(pos);
}

void Banana::positionBS(const Vector3f& pos)
{
    inst->position(pos - inst->transform().linear() * colBS.pos);
}

Vector3f Banana::getScale()
{
    return inst->getScale();
}

Vector3f Banana::getRot()
{
    return inst->getRot();
}

Vector3f Banana::getPos()
{
    return inst->getPos();
}

BS Banana::getColBS()
{
    return BS(inst->transform() * colBS.pos, colBS.radius * getScale().x());
}
