/*
    listcollectionjob.cpp - Update collection in list view.

    Copyright (c) 2006      by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#include "listcollectionjob.h"

// Qt includes
#include <qguardedptr.h>
#include <qevent.h>

// KDE includes
#include <klocale.h>
#include <kapplication.h>

// Gamefu includes
#include <collectionaccess.h>
#include <rommetainformation.h>

#include "collectionviewer.h"

static const int ListedItemEventId = 10200;

class ListedItemEvent : public QCustomEvent
{
public:
	ListedItemEvent(const Gamefu::RomMetaInformation &romInfo)
	 : QCustomEvent(ListedItemEventId), m_romInfo(romInfo)
	{}

	Gamefu::RomMetaInformation romInformation() const
	{
		return m_romInfo;
	}
private:
	Gamefu::RomMetaInformation m_romInfo;
};

class ListCollectionJob::Private
{
public:
	Private()
	 : listingFinished(false)
	{}

	QGuardedPtr<CollectionViewer> collectionViewer;
	QString sqlQuery;
	volatile bool listingFinished;
};

ListCollectionJob::ListCollectionJob(QObject *parent, CollectionViewer *collectionViewer, const QString &sqlQuery)
 : ThreadWeaver::DependentJob(parent, "ListCollectionJob"), d(new Private)
{
	d->collectionViewer = collectionViewer;
	d->sqlQuery = sqlQuery;

	setDescription( i18n("Updating collection list view...") );
}

ListCollectionJob::~ListCollectionJob()
{
	delete d;
}

bool ListCollectionJob::doJob()
{
	d->collectionViewer->clear();

	QValueList<Gamefu::RomMetaInformation> romList;
	if( d->sqlQuery.isEmpty() )
	{
		romList = Gamefu::CollectionAccess::self()->getAllRomsList();
	}
	else
	{
		romList = Gamefu::CollectionAccess::self()->getRomsListFromQuery(d->sqlQuery);
	}

	if( romList.isEmpty() )
	{
		setStatus( i18n("A transaction is in progress, collection list view will not be filled.") );
		abort();
	}
	
	setProgressTotalSteps( romList.count() );
	QValueList<Gamefu::RomMetaInformation>::ConstIterator it, itEnd = romList.constEnd();
	for(it = romList.constBegin(); it != itEnd; ++it)
	{
		if( !isAborted() )
		{
			incrementProgress();
			kapp->postEvent(this, new ListedItemEvent(*it));
		}
		else
			break;
	}

	return !isAborted();
}

void ListCollectionJob::completeJob()
{
	emit listCollectionFinished();
}

void ListCollectionJob::customEvent(QCustomEvent *event)
{
	if(event->type() == ListedItemEventId)
	{
		ListedItemEvent *listEvent = static_cast<ListedItemEvent*>(event);
		
		if(listEvent)
		{
// 			kdDebug() << "Adding " << listEvent->romInformation().romName() << " to list view" << endl;
			d->collectionViewer->addItem( listEvent->romInformation() );
		}
	}
	else
		ThreadWeaver::DependentJob::customEvent(event);
}

#include "listcollectionjob.moc"
