/*
    abstractnotify.h - Abtract Notify proxy for NotifyManager.

    Copyright (c) 2006 by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#ifndef GAMEFUABSTRACTNOTIFY_H
#define GAMEFUABSTRACTNOTIFY_H

#include <qobject.h>

namespace KIO
{
	class Job;
}

namespace Gamefu {

/**
 * @brief Provide an interface for notification of progress and display messages
 * This is a abstract class that must be implemented to actually display something.
 *
 * Serve as a backend of Gamefu::NotifyManager
 *
 * @author Michaël Larouche <larouche@kde.org>
*/
class AbstractNotify : public QObject
{
	Q_OBJECT
protected:
	/**
	 * @brief c-tor
	 * Not used, declared protected so no one would creata a AbstractNotify instance directly.
	 */
	AbstractNotify(QObject *parent = 0, const char *name = 0);
	/**
	 * d-tor
	 */
	virtual ~AbstractNotify();

public slots:
	/**
	 * @brief Display a long message in a popup message.
	 * Display a passive popup.
	 * @param message message to display.
	 * @param type message type. See @ref Gamefu::NotifiyManager::MessageType enum.
	 */
	virtual void longMessage(const QString &message, int type) = 0;
	/**
	 * @brief Display a short message with a longer message description.
	 * A short message is displayed and you can popup a longer message.
	 * @param shortMessage short message to display.
	 * @param longMessage longMessage to display.
	 * @param type message type for longMessage. See @ref Gamefu::NotifyManager::MessageType enum.
	 */
	virtual void shortLongMessage(const QString &shortMessage, const QString &longMessage, int type) = 0;
	/**
	 * @brief Display a short message.
	 * A short message is displayed during a period of time.
	 * @param shortMessage shortMessage to display.
	 * @param longDelay if true, display the message during a greater period of time.
	 */
	virtual void shortMessage(const QString &shortMessage, bool longDelay) = 0;

	/**
	 * @brief Create a new progress operation.
	 * Display a progress bar with an abort button.
	 * @param owner The owner is used to identity the progress. Owner is like the key to update the progress bar.
	 * @param description Description of the current operation.
	 * @param totalSteps Total steps required to complete the operation.
	 * @param receiver QObject receiver for the abort button.
	 * @param abortSlot Slot for the abort button.
	 */
	virtual void createNewProgressOperation(QObject *owner, const QString &description, int totalSteps, QObject *receiver, const char *abortSlot) = 0;
	/**
	 * @brief Create a new progress operation for KIO::Job
	 * Same as above, but for KIO::Job.
	 * @param job given job
	 * @param description Description of the current operation
	 * @param totalSteps Total steps required to complete the operation.
	 */
	virtual void createNewProgressOperation(KIO::Job *job, const QString &description, int totalSteps) = 0;
	/**
	 * @brief Increment progress for the given QObject operation.
	 * @param owner Given QObject.
	 */
	virtual void incrementProgress(QObject *owner) = 0;
	/**
	 * @brief Set current progress for given QObject operation.
	 * @param owner Given QObject
	 * @param steps current progress.
	 */
	virtual void setProgress(QObject *owner, int steps) = 0;
	/**
	 * @brief Set a human-readable status message for the given QObject operation.
	 * @param owner Given QObject.
	 * @param status status message.
	 */
	virtual void setProgressStatus(QObject *owner, const QString &status) = 0;
};

}

#endif
