/*
    gamefupluginmanager.h - Gamefu plugin manager

    Copyright (c) 2006      by Michaël Larouche       <larouche@kde.org>

    *************************************************************************
    *                                                                       *
    * This program is free software; you can redistribute it and/or modify  *
    * it under the terms of the GNU General Public License as published by  *
    * the Free Software Foundation; either version 2 of the License, or     *
    * (at your option) any later version.                                   *
    *                                                                       *
    *************************************************************************
*/
#ifndef GAMEFUPLUGINMANAGER_H
#define GAMEFUPLUGINMANAGER_H

#include <qobject.h>
#include <kdemacros.h>

#include <qvaluelist.h>
#include <kplugininfo.h>
namespace Gamefu
{

class Plugin;

/**
 * @brief Manage Gamefu plugins.
 *
 * @author Michaël Larouche <larouche@kde.org>
 */
class KDE_EXPORT PluginManager : public QObject
{
	Q_OBJECT
public:
   ~PluginManager();

	/**
	 * @brief Get the single instance of the PluginManager.
	 * @return Singleton instance of PluginManager;
	 */
	static PluginManager *self();

	/**
	 * @brief Get plugin info for plugins into the given category.
	 *
	 * If category is empty, it return all available plugins.
	 * @param category Plugin category (ex: Plugins, System)
	 * @return plugin info list.
	 */
	QValueList<KPluginInfo *> availablePlugins(const QString &category = QString());

public slots:
	/**
	 * @brief Load all enabled plugins.
	 *
	 * Also used to reconfigure the plugins.
	 * New enabled plugins will be loaded and not enabled
	 * anymore will be unloaded.
	 */
	void load();

	/**
	 * @brief Shutdown the plugin manager.
	 */
	void shutdown();

signals:
	/**
	 * @brief A single Plugin is loaded.
	 *
	 * Used to plug GUI items into the main window.
	 * @param plugin Plugin loaded.
	 */
	void pluginLoaded(Gamefu::Plugin *plugin);
	/**
	 * @brief A Plugin is unloaded.
	 *
	 * Used to unplug GUI items from the main window.
	 * Don't store the resulting pointer because it will be soon deleted.
	 * @param plugin Plugin unloaded.
	 */
	void pluginUnloaded(Gamefu::Plugin *plugin);

	/**
	 * @brief All Plugin are loaded.
	 */
	void loadingFinished();

private slots:
	void createNextPlugin();
	void shutdownDone();

private:
	// Disable copy and constructor, this is a singleton
	PluginManager(QObject *parent = 0, const char *name = 0);
	PluginManager(const PluginManager &);
   PluginManager &operator=(const PluginManager &);

	void createPluginInternal(const QString &pluginName);
	void unloadPlugin(const QString &pluginName);

private:
	static PluginManager *s_self;

	class Private;
	Private *d;
};

}

#endif
