/*
    Copyright 2009  Stanislas Marquis <smarquis@chaosorigin.com>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "str.h"

const size_t STR_BUF_SZ = 16;   /* chunk of memory allocated */

void str_new( Str** s, const char* v )
{
    assert( *s == NULL );

    const size_t vlen = v ? strlen( v ) + 1 : 1;
    str_new2( s, v, vlen );
}

void str_new2( Str** s, const char* v, const size_t vlen )
{
    assert( *s == NULL );

    Trace( printf( "-- New str\n" ); )
    *s = Malloc( sizeof( Str ) );

    const size_t blen = ( ( ( vlen + 1 ) / STR_BUF_SZ ) + 1 ) * STR_BUF_SZ;

    (*s)->ptr = Malloc( sizeof( char ) * blen );

    (*s)->len = vlen;
    (*s)->buflen = blen;

    if ( v )
    {
        strncpy( (*s)->ptr, v, vlen );
        (*s)->ptr[ vlen ] = '\0';
    }
    else
        (*s)->ptr[0] = '\0';
}

void str_delete( Str** s )
{
    assert( *s );
    Trace( printf( "-- Delete str\n" ); )

    Free( (*s)->ptr, (*s)->buflen );
    (*s)->ptr = NULL;

    Free( *s, sizeof( Str ) );
    *s = NULL;
}

void str_set( Str* s, char* v, size_t len )
{
    assert( s );

    if ( !v )
    {
        v = (char*) "";
        len = 1;
    }
    else if ( len < 1 )
        len = strlen( v ) + 1;

    if ( len != s->len )
    {
        const size_t blen = ( ( len / STR_BUF_SZ ) + 1 ) * STR_BUF_SZ;
        if ( s->buflen != blen )
        {
            Trace( printf( "-- Realloc str\n" ); )
            s->ptr = Realloc( s->ptr, sizeof( char ) * blen, s->buflen );
            s->buflen = blen;
        }
        s->len = len;
    }
    else
    if ( strequ( v, s->ptr ) ) return;

    strncpy( s->ptr, v, len - 1 );
    s->ptr[ len - 1 ] = '\0';
}

void str_replace( Str* s, const char* s1, const char* s2 )
{
    assert( s );
    assert( s1 && s1[0] != '\0' );
    assert( s2 );

    char* ptr = strstr( s->ptr, s1 );
    if ( !ptr ) return;

    const int s1len = strlen( s1 );
    const int s2len = strlen( s2 );

    Str* dup = NULL;
    str_new( &dup, s->ptr );
    int cnt = 1;
    ptr = &dup->ptr[ ptr - s->ptr ];
    *ptr = '\0';
    ptr += s1len;

    while ( ( ptr = strstr( ptr, s1 ) ) )
    {
        *ptr = '\0';
        ptr += s1len;
        ++cnt;
    }

    const size_t len = s->len + ( ( s2len - s1len ) * cnt );
    if ( len != s->len )
    {
        const size_t blen = ( ( len / STR_BUF_SZ ) + 1 ) * STR_BUF_SZ;
        if ( s->buflen != blen )
        {
            Trace( printf( "-- Realloc str\n" ); )
            s->ptr = Realloc( s->ptr, sizeof( char ) * blen, s->buflen );
            s->buflen = blen;
        }
        s->len = len;
    }

    ptr = dup->ptr;
    char* ptr2 = s->ptr;
    int i = 0;
    for ( ; i < cnt; ++i )
    {
        while ( *ptr )
            *ptr2++ = *ptr++;
        if ( s2len )
            strncpy( ptr2, s2, s2len );
        ptr += s1len;
        ptr2 += s2len;
    }
    if ( *ptr )
        strcpy( ptr2, ptr );
    else
        *ptr2 = '\0';

    str_delete( &dup );
}

void str_xmlize( Str* s )
{
    assert( s );
    str_replace( s, "&", "&amp;" );
    str_replace( s, "<", "&lt;" );
    str_replace( s, ">", "&gt;" );
    str_replace( s, "\"", "&quot;" );
    str_replace( s, "'", "&apos;" );
}

Bool str_to_file( const Str* s, const char* path )
{
    assert( s );

    FILE *f = fopen( path, "w" );
    if ( !f )
        return False;

    fwrite( s->ptr, sizeof( char ), s->len - 1, f );
    if ( fclose( f ) != 0 )
        return False;
    return True;
}

Bool strequ( char* s1, char* s2 )
{
    if ( s1 == s2 ) return True;
    while ( 1 )
    {
        if ( *s1 != *s2++ ) return False;
        if ( !*s1++ ) return True;
    }
}

char* notadigit( char* s )
{
    for ( ; *s; ++s )
    {
        if ( !isdigit( *s ) )
            return s;
    }
    return NULL;
}

#ifndef NDEBUG

void str_unittest()
{
    Str* s1 = NULL;
    Str* s2 = NULL;

    str_new( &s1, "test 1" );
    str_new( &s2, "test 2" );

    assert( strcmp( s1->ptr, "test 1" ) == 0 );
    assert( strcmp( s2->ptr, "test 2" ) == 0 );
    assert( str_cmp( s1, s2 ) != 0 );
    assert( ! str_equ( s1, s2 ) );
    assert( s1->len == s2->len );

    str_delete( &s1 );
    str_delete( &s2 );
}

#endif /* !NDEBUG */
