/* 
 * Copyright (C) 2005 <mikelima@cirulla.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */


#include <stdlib.h>
#include <stdint.h>

#include <string>

#include <klocale.h>
#include <kgenericfactory.h>
#include <kdebug.h>

#include <qcstring.h>
#include <qfile.h>
#include <qregexp.h>

#include "kfile_tads.h"

static int16_t
readInt16LE(QFile &f)
{
    int16_t val = f.getch();
    val |= (f.getch() & 255) << 8;
    return val;
}

static int32_t
readInt32LE(QFile &f)
{
    int32_t val = f.getch();
    val |= (f.getch() & 255) << 8;
    val |= (f.getch() & 255) << 16;
    val |= (f.getch() & 255) << 24;
    return val;
}

typedef KGenericFactory<KTADSPlugin> TADSFactory;

K_EXPORT_COMPONENT_FACTORY(kfile_tads, TADSFactory("kfile_tads"))

KTADSPlugin::KTADSPlugin(
	QObject *parent, const char *name, const QStringList &args) : 
KFilePlugin(parent, name, args)
{
    // register our mimetypes
    QPtrList<KFileMimeTypeInfo> info;
    
    info.append(addMimeTypeInfo("application/x-tads"));
    info.append(addMimeTypeInfo("application/x-t3vm-image"));

    KFileMimeTypeInfo::GroupInfo *gameGroup;
    KFileMimeTypeInfo::GroupInfo *runtimeGroup;

    // info group
    for (KFileMimeTypeInfo *i = info.first(); i; i = info.next()) {
	gameGroup = addGroupInfo(i, "Game", i18n("Game Information"));
	addItemInfo(gameGroup, 
		"AuthorEmail", i18n("Author Email"), QVariant::String);
	addItemInfo(gameGroup, 
		"ByLine", i18n("Credits"), QVariant::String);
	addItemInfo(gameGroup, 
		"CopyingRules", i18n("Copying Rules"), QVariant::String);
	KFileMimeTypeInfo::ItemInfo *name = addItemInfo(gameGroup, 
		"GameName", i18n("Name"), QVariant::String);
	setHint(name, KFileMimeTypeInfo::Name);
	addItemInfo(gameGroup, 
		"Version", i18n("Version"), QVariant::String);
	addItemInfo(gameGroup, 
		"Language", i18n("Language"), QVariant::String);
	addItemInfo(gameGroup, 
		"LicenseType", i18n("License Type"), QVariant::String);
	addItemInfo(gameGroup, 
		"Profile", i18n("Presentation Profile"), QVariant::String);
	addItemInfo(gameGroup, 
		"ReleaseDate", i18n("Release Date"), QVariant::String);
	KFileMimeTypeInfo::ItemInfo *description = addItemInfo(gameGroup, 
		"Description", i18n("Description"), QVariant::String);
	setAttributes(description, 
		KFileMimeTypeInfo::MultiLine | KFileMimeTypeInfo::SqueezeText);
	setHint(description, KFileMimeTypeInfo::Description);

	runtimeGroup = addGroupInfo(i, "Runtime", i18n("Runtime Information"));
	addItemInfo(runtimeGroup, 
		"TADSVersion", i18n("TADS Version"), QVariant::String);
	addItemInfo(runtimeGroup, 
		"TMachine", i18n("T Machine"), QVariant::String);
	addItemInfo(runtimeGroup, 
		"CreationTime", i18n("Creation Time"), QVariant::String);
    }
}
enum FileFormat {
    Tads2Format,
    Tads3Format
};

class BlockHeader {
    QCString m_id;
    size_t m_length;
public:
    void read(QFile &f, FileFormat format);
    QCString id() { return m_id; }
    size_t size() { return m_length; }
};

void BlockHeader::read(QFile &f, FileFormat format)
{
    switch (format) {
    case Tads2Format:
	{
	    // Tads 2 section header contains one byte of ID size, n bytes of
	    // ID, 4 bytes indicating the end of the section, specified from
	    // the beginning of the file.
	    unsigned char idLength = f.getch();
	    
	    m_id.resize(idLength + 1);
	    f.readBlock(m_id.data(), idLength);
	    size_t currPos = f.at(); 
	    m_length = readInt32LE(f) - currPos - 4;
	}
	break;
    case Tads3Format:
	// Tads 3 section header is simpler.  ID is always 4 bytes, the section
	// lenght follows, and then there is 2 bytes of flags.
	m_id.resize(5);
	f.readBlock(m_id.data(), 4);
	m_length = readInt32LE(f); 
	readInt16LE(f); // Skip flags
	break;
    default:
        break;
    }
}

static uint
findTads2Resource(QFile &f, const char *resourceName)
{
    uint entries = readInt32LE(f);
    kdDebug(7034) << "Resource Table has " << entries << " entries" << endl;

    // Skip reserver word if reading a Tads2 resource section
    readInt32LE(f);
    
    uint resDataOffset;
    uint resDataLength;
    uint resNameLength;
    uint targetDataOffset = 0;
    uint targetDataLength = 0;
    for (uint i = 0; i < entries; i++) {
        resDataOffset = readInt32LE(f);
	resDataLength = readInt32LE(f);
	
	// Tads3 resource names are 1 byte vs. two for Tads2
	resNameLength = readInt16LE(f);
	QCString resName(resNameLength + 1);
	f.readBlock(resName.data(), resNameLength);
	kdDebug(7034) << resName << " " << 
	    resDataOffset << " " << resDataLength << endl; 
	if (strlen(resourceName) != resNameLength) {
	    continue;
	}
	if (memcmp(resourceName, resName.data(), resNameLength) == 0) {
	    targetDataLength = resDataLength;
	    targetDataOffset = resDataOffset;
	}
    }
    if (targetDataLength != 0) {
	f.at(f.at() + targetDataOffset);
    }
    return targetDataLength;
}
static int
findTads3Resource(QFile &f, const char *resourceName)
{
    size_t resourceBase = f.at();
    uint entries = readInt16LE(f);
    kdDebug(7034) << "Resource Table has " << entries << " entries" << endl;

    uint resDataOffset;
    uint resDataLength;
    uint resNameLength;
    for (uint i = 0; i < entries; i++) {
        resDataOffset = readInt32LE(f);
	resDataLength = readInt32LE(f);
	
	resNameLength = f.getch();
	QCString resName(resNameLength + 1);
	f.readBlock(resName.data(), resNameLength);
	
	// Tads3 resource names are "encrypted" 
	for (uint j = 0; j < resNameLength; j++) {
	    resName[j] ^= 0xff;
	}
	kdDebug(7034) << resName << " " << 
	    resDataOffset << " " << resDataLength << endl; 
	if (strlen(resourceName) != resNameLength) {
	    continue;
	}
	if (memcmp(resourceName, resName.data(), resNameLength) == 0) {
	    f.at(resourceBase + resDataOffset);
	    return resDataLength;
	}
    }
    return 0;
}

static void 
scanSections(QFile &f, FileFormat format, QMap<QString, QString> &properties) 
{
    // Skip the header
    BlockHeader header;
    const QCString MediaId(format == Tads2Format ? "HTMLRES" : "MRES");
    const QCString EofId(format == Tads2Format ? "$EOF" : "EOF ");
    do {
	header.read(f, format);
	if (header.id() == MediaId) {
	    int length;
	    kdDebug(7034) << header.size() << 
		    " bytes of " << MediaId << " found" << endl;
	    if (format == Tads2Format) length = findTads2Resource(f, "GameInfo.txt");
	    else if (format == Tads3Format) length = findTads3Resource(f, "GameInfo.txt");
	    else length = 0;
	    
	    if ((length > 0) && (length < 65536)) {
		kdDebug(7034) << 
		    "parsing Gameinfo.txt, length = " << length << endl;
		QByteArray resData(length);
		f.readBlock(resData.data(), length);
		QTextStream s(resData, IO_ReadOnly);
		s.setEncoding(QTextStream::UnicodeUTF8);
	    
		QRegExp comment("^#.*");
		QRegExp blank("^\\s*$");
		QRegExp continuation("^\\s+(.*)$");
		QRegExp nameValue("^([A-Z0-9\\.]*):(.*)$", false);
		nameValue.setMinimal(true);

		QString name;
		QString value;

		// Parse the stream and put the values in a map
		while (!s.atEnd()) {
		    QString line = s.readLine();
		    if (comment.search(line) != -1) {
			// It's a comment, skip it
			kdDebug(7034) << "comment found" << endl;
			continue;
		    } else if (comment.search(line) != -1) {
			// It's a blank line, skip it
			kdDebug(7034) << "blank line found" << endl;
			continue;
		    } else if (nameValue.search(line) != -1) {
			// It's a name/value pair 
			name = nameValue.cap(1).lower();
			value = nameValue.cap(2).simplifyWhiteSpace();
			kdDebug(7034) << "name[" << name << 
			    "]/value found" << endl;
			properties[name] = value;
		    } else if (continuation.search(line) != -1) {
			kdDebug(7034) << "continuation found" << endl;
			value += continuation.cap(1).simplifyWhiteSpace();
			if (!name.isEmpty()) {
			    properties[name] = value;
			}
		    } else {
			kdDebug(7034) << "no match for line " << line << endl;
		    }
		}
		break;
	    } else {
		kdDebug(7034) << "no GameInfo.txt (or too big)" << endl;
	    }
	    break;
	}
	if (header.id() == EofId) break;
	if (header.size() == 0) break;
	kdDebug(7034) << "skipping " << header.size() << 
		     " bytes of " << header.id() << endl;
	if (!f.at(f.at() + header.size())) {
	    // Seek failed. Either the file is incomplete or it is corrupted.
	    break;
	}
    } while (!f.atEnd());
}

bool 
KTADSPlugin::readInfo(KFileMetaInfo& info, uint what)
{
    if (info.path().isEmpty()) return false;

    QFile f(info.path());
    if (!f.open(IO_ReadOnly)) return false;

    // Read the TADS file header. 
    // 
    char data[69];
    if (f.readBlock(data, sizeof data) != sizeof data) return false;
    
    bool fastScan = (what & KFileMetaInfo::Fastest);
    
    QString virtualMachine;
    QString tadsVersion;
    QString creationTime;
    unsigned int fileFormat;

    QString gameName;
    QString byLine;
    QString authorEmail;
    QString desc;
    QString language;
    QString copyingRules;
    QString licenseType;
    QString releaseDate;
    QString version;
    QString profile;
    QMap<QString, QString> properties;
    
    if (memcmp(data, "TADS2 bin\n\r\032", 12) == 0) {
	virtualMachine = "Version 2";
	tadsVersion = QString::fromLatin1(&data[13], 7);
	creationTime = QString::fromLatin1(&data[22], 24);
	if (!fastScan || !f.isSequentialAccess()) {
	    if (f.at(48)) {
		scanSections(f, Tads2Format, properties);
	    }
	}
    } else if (memcmp(data, "T3-image\r\n\032", 11) == 0) {
	virtualMachine = "Version 3";
	fileFormat = data[11] + data[12] * 256;
	if (fileFormat == 1) {
	    creationTime = QString::fromLatin1(&data[45], 24);
	}
	if (!fastScan || !f.isSequentialAccess()) {
	    if (f.at(69)) {
		scanSections(f, Tads3Format, properties);
	    }
	}
    } else {
        return false;
    }

    // Find the information we were searching.
    gameName = properties["name"];
    byLine = properties["byline"];
    authorEmail = properties["authoremail"];
    desc = properties["desc"];
    copyingRules = properties["copyingrules"];
    licenseType = properties["licensetype"];
    language = properties["language"];
    copyingRules = properties["copyingrules"];
    releaseDate = properties["releasedate"];
    version = properties["version"];
    profile = properties["presentationprofile"];

    // If all is well, set the metadata
    KFileMetaInfoGroup gameGroup = appendGroup(info, "Game");
    KFileMetaInfoGroup runtimeGroup = appendGroup(info, "Runtime");
    appendItem(runtimeGroup, "TMachine", virtualMachine);
    if (!tadsVersion.isEmpty()) {
	appendItem(runtimeGroup, "TADSVersion", tadsVersion);
    }
    if (!creationTime.isEmpty()) {
	appendItem(runtimeGroup, "CreationTime", creationTime);
    }
    if (!gameName.isEmpty()) {
	appendItem(gameGroup, "GameName", gameName);
    }
    if (!releaseDate.isEmpty()) {
	appendItem(gameGroup, "ReleaseDate", releaseDate);
    }
    if (!version.isEmpty()) {
	appendItem(gameGroup, "Version", version);
    }
    if (!byLine.isEmpty()) {
	appendItem(gameGroup, "ByLine", byLine);
    }
    if (!authorEmail.isEmpty()) {
	appendItem(gameGroup, "AuthorEmail", authorEmail);
    }
    if (!desc.isEmpty()) {
	appendItem(gameGroup, "Description", QString("<qt>%1</qt>").arg(desc));
    }
    if (!copyingRules.isEmpty()) {
	appendItem(gameGroup, "CopyingRules", copyingRules);
    }
    if (!language.isEmpty()) {
	appendItem(gameGroup, "Language", language);
    }
    if (!licenseType.isEmpty()) {
	appendItem(gameGroup, "LicenseType", licenseType);
    }
    if (!profile.isEmpty()) {
	appendItem(gameGroup, "Profile", profile);
    }
    return true;
}

#include "kfile_tads.moc"
// vim:sw=4:
