/***************************************************************************
 *   Copyright 2009 Casper van Donderen <casper.vandonderen@gmail.com>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "renderstatemanager.h"

#include <QColor>

#ifdef Q_WS_WIN
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

Kolf::RenderStateManager::RenderStateManager()
	: m_currentViewport(top)
{
}

void Kolf::RenderStateManager::setDefaultRenderState()
{
	enableDepthState();
	disableBlendingState();
	//disableLightState();
	enableLightState();
	disableWireframeState();
	enableTextureState();
	setGlobalBackgroundColor(Qt::white);
	setGlobalDrawingColor(Qt::green);
	setViewport(front);
	setGlobalLight();
}

void Kolf::RenderStateManager::setGlobalBackgroundColor(float r, float g, float b, float a)
{
	glClearColor(r, g, b, a);
}

void Kolf::RenderStateManager::setGlobalBackgroundColor(const QColor& color)
{
	qreal r, g, b, a;
	color.getRgbF(&r, &g, &b, &a);
	glClearColor(r, g, b, a);
}

void Kolf::RenderStateManager::setGlobalDrawingColor(float r, float g, float b)
{
	glColor3f(r, g, b);
}

void Kolf::RenderStateManager::setGlobalDrawingColor(const QColor& color)
{
	qreal r, g, b;
	color.getRgbF(&r, &g, &b);
	glColor3f(r, g, b);
}

void Kolf::RenderStateManager::setGlobalLight()
{
	glShadeModel(GL_SMOOTH);
	GLfloat lightPosition[] = {1.0, 1.0, 1.0, 0.0};	
	GLfloat whiteLight[] = {1.0, 1.0, 1.0, 1.0};	
	glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, whiteLight);
	//glLightfv(GL_LIGHT0, GL_SPECULAR, whiteLight);
	glEnable(GL_LIGHT0);
}

void Kolf::RenderStateManager::enableBlendingState()
{
	glEnable(GL_BLEND);
}

void Kolf::RenderStateManager::enableDepthState()
{
	glEnable(GL_DEPTH_TEST);
}

void Kolf::RenderStateManager::enableLightState()
{
	glEnable(GL_LIGHTING);
}

void Kolf::RenderStateManager::enableWireframeState()
{
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
}

void Kolf::RenderStateManager::enableTextureState()
{
	glEnable(GL_TEXTURE_2D);
}

void Kolf::RenderStateManager::disableBlendingState()
{
	glDisable(GL_BLEND);
}

void Kolf::RenderStateManager::disableDepthState()
{
	glDisable(GL_DEPTH_TEST);
}

void Kolf::RenderStateManager::disableLightState()
{
	glDisable(GL_LIGHTING);
}

void Kolf::RenderStateManager::disableWireframeState()
{
	glPolygonMode(GL_FRONT, GL_POLYGON);
}

void Kolf::RenderStateManager::disableTextureState()
{
	glDisable(GL_TEXTURE_2D);
}

void Kolf::RenderStateManager::setTopViewport()
{
	gluLookAt(500, 500, 2000, 500, 500, 0, 0, 1, 0);
	m_currentViewport = top;
}

///After this some placeholders to later be able to switch camera viewports. Not completely working yet....
void Kolf::RenderStateManager::setFrontPerspectiveViewport()
{
	gluLookAt(500, -1000, 1000, 500, 500, 0, 0, 0, 1);
	m_currentViewport = front;
}

void Kolf::RenderStateManager::setRightPerspectiveViewport()
{
	gluLookAt(2000, 500, 1000, 500, 500, 0, 0, 0, 1);
	m_currentViewport = right;
}

void Kolf::RenderStateManager::setBackPerspectiveViewport()
{
	gluLookAt(500, 2000, 1000, 500, 500, 0, 0, 0, 1);
	m_currentViewport = back;
}

void Kolf::RenderStateManager::setLeftPerspectiveViewport()
{
	gluLookAt(-1000, 500, 1000, 500, 500, 0, 0, 0, 1);
	m_currentViewport = left;
}

void Kolf::RenderStateManager::setViewport(int direction)
{
	switch ( direction )
	{
		case top:
			setTopViewport();
			break;
		case front:
			setFrontPerspectiveViewport();
			break;
		case back:
			setBackPerspectiveViewport();
			break;
		case left:
			setLeftPerspectiveViewport();
			break;
		case right:
			setRightPerspectiveViewport();
			break;
		default :
			break;
	}

}

void Kolf::RenderStateManager::changeViewport(int w, int h, int direction)
{
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	double aspect = (double)w/h;
	gluPerspective(30, aspect, 1, 3000);	
	setViewport(direction);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}
