/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_TEXTUREPOOL_H
#define KOLF_TEXTUREPOOL_H

#include <QMap>
#include <QImage>
#include <QObject>

#ifdef Q_WS_WIN
#include <windows.h>
#endif

#include <GL/gl.h>
class QGLContext;

namespace Kolf
{
	/**
	 * \class TexturePool
	 * \since 2.0
	 *
	 * This singleton class implements a key-based global storage for textures, and a texture binding strategy for QGLContexts.
	 */
	class TexturePool : public QObject
	{
		Q_OBJECT
		private:
			TexturePool();
			~TexturePool();
			Q_DISABLE_COPY(TexturePool)
		public:
			static TexturePool* instance();

			///Puts a new texture into the pool. If a texture with this \a key is already in the pool, the old texture is removed and replaced with the new texture.
			void addTexture(const QByteArray& key, const QImage& image);
			///Removes the texture with the given \a key from the pool (and removes it from all OpenGL contexts that uses it previously).
			void removeTexture(const QByteArray& key);
			///Makes the texture available for usage within the given \a context.
			///\returns A valid texture ID, or 0 if the given texture \a key is invalid.
			///\note You still have to call glBindTexture() on the texture ID you get from this function!
			GLuint prepareTexture(const QByteArray& key, QGLContext* context);
		public Q_SLOTS:
			///For any QGLContext that has ever bound textures from this pool, call this before the \a context is destroyed. The texture pool needs to update its internal data structures before the context is deleted.
			void detachContext(QGLContext* context);
		private:
			struct ContextData
			{
				QMap<QByteArray, GLuint> textures;
			};
			QMap<QByteArray, QImage> m_textureData; ///< Holds all known textures.
			QMap<QGLContext*, ContextData> m_contextData; ///< Holds information about which contexts have already loaded which textures.
	};
}

#endif // KOLF_TEXTUREPOOL_H
