/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "managedobject.h"
#include "object.h"
#include "objectregistry.h"
#include "scene.h"

#include <KConfigGroup>

Kolf::ManagedObject::ManagedObject(Kolf::Object* object, Kolf::Scene* scene)
	: m_object(object)
	, m_scene(scene)
	, m_currentStateIndex(0) //the according state is created below
	, m_blockPushRequests(false)
{
	//initialize history with the pre-creation state
	m_states << new Kolf::ManagedObject::State;
}

Kolf::ManagedObject::ManagedObject(Kolf::ManagedObject::State* state, Kolf::Scene* scene)
	: m_object(0)
	, m_scene(scene)
	, m_currentStateIndex(1)
	, m_blockPushRequests(false)
{
	//initialize history with the pre-creation state and the current state
	m_states << new Kolf::ManagedObject::State << state;
}

Kolf::ManagedObject::~ManagedObject()
{
	qDeleteAll(m_states);
	delete m_object;
}

Kolf::Object* Kolf::ManagedObject::object() const
{
	Kolf::ManagedObject::State* currentState = m_states[m_currentStateIndex];
	if (!m_object && !currentState->isEmpty())
	{
		m_object = Kolf::ObjectRegistry::createObject(currentState->className(), m_scene);
		if (m_object)
		{
			currentState->applyToObject(m_object);
			Kolf::ManagedObject* mo = const_cast<Kolf::ManagedObject*>(this);
			emit mo->objectCreated(mo);
		}
	}
	return m_object;
}

int Kolf::ManagedObject::stateCount() const
{
	return m_states.count();
}

int Kolf::ManagedObject::currentState() const
{
	return m_currentStateIndex;
}

Kolf::ManagedObject::State* Kolf::ManagedObject::state(int index) const
{
	return m_states.value(index);
}

void Kolf::ManagedObject::serialize(KConfigGroup* group)
{
	//TODO: For now, only the current state of the object is saved.
	m_states[m_currentStateIndex]->serialize(group);
}

Kolf::ManagedObject* Kolf::ManagedObject::deserialize(KConfigGroup* group, Kolf::Scene* scene)
{
	Kolf::ManagedObject::State* state = Kolf::ManagedObject::State::deserialize(group);
	return state ? new Kolf::ManagedObject(state, scene) : 0;
}

Kolf::ManagedObjectStateRef Kolf::ManagedObject::pushState(bool allowStateMerging)
{
	//FIXME: The checks in isMediacySignificant might require knowledge of the ManagedObjectGroup's base state index.
	//deny any push requests if activateState() is active
	if (m_blockPushRequests)
		return Kolf::ManagedObjectStateRef();
	//delete all states after the current state
	while (m_states.count() > m_currentStateIndex + 1)
		delete m_states.takeAt(m_currentStateIndex + 1);
	//create new state
	Kolf::ManagedObject::State* state = Kolf::ManagedObject::State::fromObject(m_object);
	//do not append state if both current and new state are empty
	if (m_states[m_currentStateIndex]->isEmpty() && state->isEmpty())
		return Kolf::ManagedObjectStateRef();
	//do not append state if the difference to the previous state is not significant, but merge it with the previous state instead
	if (m_currentStateIndex != 0 && allowStateMerging)
	{
		Kolf::ManagedObject::State* previousState1 = m_states[m_currentStateIndex - 1];
		Kolf::ManagedObject::State* previousState2 = m_states[m_currentStateIndex];
		if (!Kolf::ManagedObject::State::isMediacySignificant(previousState1, previousState2, state))
		{
			m_states[m_currentStateIndex] = state;
			delete previousState2;
			return Kolf::ManagedObjectStateRef(); //no new state has been added
		}
	}
	//append state
	m_states << state;
	++m_currentStateIndex;
	return Kolf::ManagedObjectStateRef(this, m_currentStateIndex);
}

void Kolf::ManagedObject::activateState(int index)
{
	//can we move to this state?
	if (index < 0 || index >= m_states.count())
		return;
	//do we need to move to this state?
	if (index == m_currentStateIndex)
		return;
	//deny any calls to pushState() during the transition, or the changes in the object's properties could lead to new states being pushed
	m_blockPushRequests = true;
	//perform transition
	Kolf::ManagedObject::State* fromState = m_states[m_currentStateIndex];
	Kolf::ManagedObject::State* toState = m_states[index];
	m_currentStateIndex = index;
	if (fromState->isEmpty())
	{
		if (!toState->isEmpty())
		{
			//create object
			m_object = Kolf::ObjectRegistry::createObject(toState->className(), m_scene);
			if (m_object)
				toState->applyToObject(m_object);
			emit objectCreated(this);
		}
	}
	else // !fromState->isEmpty()
	{
		if (toState->isEmpty())
		{
			//delete object
			emit objectAboutToBeDeleted(this);
			delete m_object;
			m_object = 0;
		}
		else
			//apply new state to object
			toState->applyToObject(m_object);
	}
	m_blockPushRequests = false;
}

#include "managedobject.moc"
