/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_MANAGEDOBJECTGROUP_H
#define KOLF_MANAGEDOBJECTGROUP_H

#include "managedobject.h"

namespace Kolf
{
	/**
	 * \class ManagedObjectGroup
	 * \since 2.0
	 *
	 * This class provides state tracking for multiple Kolf::Object instances, with an interface similar to QUndoStack.
	 * \sa Kolf::ManagedObject
	 */
	class ManagedObjectGroup : public QObject
	{
		Q_OBJECT
		Q_DISABLE_COPY(ManagedObjectGroup)
		public:
			///Creates a new managed object group.
			ManagedObjectGroup(Kolf::Scene* scene);
			///Deletes this managed object group, as well as all managed objects.
			virtual ~ManagedObjectGroup();

			///Returns whether there is a state after the current state, to which the group can be moved by a redo operation.
			bool canRedo() const;
			///Returns whether there is a state before the current state, to which the group can be moved by a undo operation.
			bool canUndo() const;
			///Returns whether the group is in the clean state (i.e., the state that was marked as saved).
			bool isClean() const;
			///Returns the count of managed objects in this managed object group.
			int managedObjectCount() const;
			///Returns the n-th managed object in this managed object group, or a null pointer if the index n is invalid.
			Kolf::ManagedObject* managedObject(int index) const;

			///Adds a new managed object to this group. If you do not have a Kolf::ManagedObject instance for your Kolf::Object instance, use the other overload, which includes checks for safely wrapping a Kolf::Object into a Kolf::ManagedObject.
			///\note The managed object group takes ownership of the Kolf::ManagedObject instance.
			void addObject(Kolf::ManagedObject* managedObject);
			///\overload
			Kolf::ManagedObject* addObject(Kolf::Object* object);
			///Deletes the given managed object, and the Kolf::Object instance inside it.
			///\warning This operation deletes the complete history of the given managed object, as if it was never there. This operation is irreversible, and in no way identical to removing the object from the scene (e.g. in the editor). In this case, you want to add an empty state to the managed object.
			void removeObject(Kolf::ManagedObject* managedObject);
			///\overload
			void removeObject(Kolf::Object* object);
			///Removes all managed objects from this group, thereby effectively resetting this group to the state it had immediately after its creation.
			///\warning This operation is irreversible. See removeObject() for details.
			void clear();

			///Saves the current state of all contained objects and such into the KConfigGroup.
			void serialize(KConfigGroup* group);
			///Loads the serialized managed object group into this object.
			///\note It is highly recommended to use this method only with empty groups.
			void deserialize(KConfigGroup* group);
		public Q_SLOTS:
			///Saves the \a object's current state in the history of the according managed object. This method does nothing if the given \a object is not managed by this managed object group.
			void pushState(Kolf::ManagedObject* managedObject);
			///\overload
			void pushState(Kolf::Object* object);
			///Indicates that the current state of this managed object group is the "base state", i.e. no state before this state can be accessed through the user interface.
			void markAsBaseState();

			//QUndoStack interface imitation
			void redo();
			void undo();
			void setClean();
		Q_SIGNALS:
			///This signal forwards the objectCreated() signals of all Kolf::ManagedObject instances in this group.
			void objectCreated(Kolf::ManagedObject* managedObject);
			///This signal forwards the objectAboutToBeDeleted() signals of all Kolf::ManagedObject instances in this group.
			void objectAboutToBeDeleted(Kolf::ManagedObject* managedObject);

			//QUndoStack interface imitation
			void cleanChanged(bool clean);
			void canRedoChanged(bool canRedo);
			void canUndoChanged(bool canUndo);
		private Q_SLOTS:
			void slotObjectCreated(Kolf::ManagedObject* managedObject);
			void objectUpdated();
		private:
			///Removes all states after the current state. (This operation must be performed before any new states are created and appended to the state order.)
			void trimStateList();

			Kolf::Scene* m_scene;
			QList<Kolf::ManagedObject*> m_managedObjects;
			QList<Kolf::ManagedObjectStateRef> m_stateOrder;
			int m_currentStateIndex, m_cleanStateIndex, m_baseStateIndex;
	};
}

#endif // KOLF_MANAGEDOBJECTGROUP_H
