/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_MATH_SIZEF_H
#define KOLF_MATH_SIZEF_H

#include <QDataStream>
#include <QDebug>
#include <QMetaType>

namespace Kolf
{
	/**
	 * \class SizeF
	 * \since 2.0
	 * This is a three-dimensional equivalent of QSizeF. The interface is the same as for QSizeF, with the following exceptions:
	 * \li Z dimensions have been added everywhere.
	 * \li Instead of "width" and "height", the terms "xSpread", "ySpread" and "zSpread" are used.
	 * \li The transpose() and toSize() methods are not available.
	 */
	class SizeF
	{
		private:
			qreal m_xs, m_ys, m_zs; //names are short forms for "xSpread" and so on
		public:
			inline SizeF()
				: m_xs(0), m_ys(0), m_zs(0)
			{
			}
			inline SizeF(const Kolf::SizeF& size)
				: m_xs(size.m_xs), m_ys(size.m_ys), m_zs(size.m_zs)
			{
			}
			inline SizeF(qreal xSpread, qreal ySpread, qreal zSpread)
				: m_xs(xSpread), m_ys(ySpread), m_zs(zSpread)
			{
			}

			inline bool isEmpty() const
			{
				return m_xs <= 0 || m_ys <= 0 || m_zs <= 0;
			}
			inline bool isNull() const
			{
				return m_xs == 0 && m_ys == 0 && m_zs == 0;
			}
			inline bool isValid() const
			{
				return m_xs >= 0 && m_ys >= 0 && m_zs >= 0;
			}

			inline qreal& rxSpread()
			{
				return m_xs;
			}
			inline qreal& rySpread()
			{
				return m_ys;
			}
			inline qreal& rzSpread()
			{
				return m_zs;
			}
			inline qreal xSpread() const
			{
				return m_xs;
			}
			inline qreal ySpread() const
			{
				return m_ys;
			}
			inline qreal zSpread() const
			{
				return m_zs;
			}

			inline void setXSpread(qreal xSpread)
			{
				m_xs = xSpread;
			}
			inline void setYSpread(qreal ySpread)
			{
				m_ys = ySpread;
			}
			inline void setZSpread(qreal zSpread)
			{
				m_zs = zSpread;
			}

			inline void scale(qreal xSpread, qreal ySpread, qreal zSpread, Qt::AspectRatioMode mode)
			{
				switch (mode)
				{
					case Qt::IgnoreAspectRatio:
						m_xs = xSpread;
						m_ys = ySpread;
						m_zs = zSpread;
						break;
					case Qt::KeepAspectRatio:
						*this *= qMin(qMin(xSpread / m_xs, ySpread / m_ys), zSpread / m_zs);
						break;
					case Qt::KeepAspectRatioByExpanding:
						*this *= qMax(qMax(xSpread / m_xs, ySpread / m_ys), zSpread / m_zs);
						break;
				}
			}
			inline void scale(const Kolf::SizeF& size, Qt::AspectRatioMode mode)
			{
				scale(size.m_xs, size.m_ys, size.m_zs, mode);
			}
			inline Kolf::SizeF boundedTo(const Kolf::SizeF& size) const
			{
				Kolf::SizeF newSize(*this);
				if (m_xs > size.m_xs)
					newSize.rxSpread() = size.m_xs;
				if (m_ys > size.m_ys)
					newSize.rySpread() = size.m_ys;
				if (m_zs > size.m_zs)
					newSize.rzSpread() = size.m_zs;
				return newSize;
			}
			inline Kolf::SizeF expandedTo(const Kolf::SizeF& size) const
			{
				Kolf::SizeF newSize(*this);
				if (m_xs < size.m_xs)
					newSize.rxSpread() = size.m_xs;
				if (m_ys < size.m_ys)
					newSize.rySpread() = size.m_ys;
				if (m_zs < size.m_zs)
					newSize.rzSpread() = size.m_zs;
				return newSize;
			}

			inline Kolf::SizeF& operator=(const Kolf::SizeF& size)
			{
				m_xs = size.m_xs;
				m_ys = size.m_ys;
				m_zs = size.m_zs;
				return *this;
			}
			inline Kolf::SizeF& operator+=(const Kolf::SizeF& size)
			{
				m_xs += size.m_xs;
				m_ys += size.m_ys;
				m_zs += size.m_zs;
				return *this;
			}
			inline Kolf::SizeF& operator-=(const Kolf::SizeF& size)
			{
				m_xs -= size.m_xs;
				m_ys -= size.m_ys;
				m_zs -= size.m_zs;
				return *this;
			}
			inline Kolf::SizeF& operator*=(qreal factor)
			{
				m_xs *= factor;
				m_ys *= factor;
				m_zs *= factor;
				return *this;
			}
			inline Kolf::SizeF& operator/=(qreal divisor)
			{
				m_xs /= divisor;
				m_ys /= divisor;
				m_zs /= divisor;
				return *this;
			}

			inline const Kolf::SizeF operator+(const Kolf::SizeF& size) const
			{
				return Kolf::SizeF(m_xs + size.m_xs, m_ys + size.m_ys, m_zs + size.m_zs);
			}
			inline const Kolf::SizeF operator-(const Kolf::SizeF& size) const
			{
				return Kolf::SizeF(m_xs - size.m_xs, m_ys - size.m_ys, m_zs - size.m_zs);
			}
			inline const Kolf::SizeF operator*(qreal factor) const
			{
				return Kolf::SizeF(m_xs * factor, m_ys * factor, m_zs * factor);
			}
			inline const Kolf::SizeF operator/(qreal divisor) const
			{
				return Kolf::SizeF(m_xs / divisor, m_ys / divisor, m_zs / divisor);
			}

			inline bool operator==(const Kolf::SizeF& size) const
			{
				return m_xs == size.m_xs && m_ys == size.m_ys && m_zs == size.m_zs;
			}
			inline bool operator!=(const Kolf::SizeF& size) const
			{
				return m_xs != size.m_xs || m_ys != size.m_ys || m_zs != size.m_zs;
			}
	};
}

inline const Kolf::SizeF operator*(qreal factor, const Kolf::SizeF& size)
{
	return size * factor;
}

inline QDataStream& operator<<(QDataStream& stream, const Kolf::SizeF& size)
{
	stream << double(size.xSpread()) << double(size.ySpread()) << double(size.zSpread());
	return stream;
}
inline QDataStream& operator>>(QDataStream& stream, Kolf::SizeF& size)
{
	stream >> size.rxSpread();
	stream >> size.rySpread();
	stream >> size.rzSpread();
	return stream;
}
inline QDebug operator<<(QDebug debug, const Kolf::SizeF& size)
{
	debug.nospace() << "Kolf::SizeF(" << size.xSpread() << ", " << size.ySpread() << ", " << size.zSpread() << ")";
	return debug.space();
}

Q_DECLARE_METATYPE(Kolf::SizeF)

#endif // KOLF_MATH_SIZEF_H
