/***************************************************************************
 *   Copyright 2008-2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_OBJECT_H
#define KOLF_OBJECT_H

#include <QColor>
#include <QGraphicsItem>
#include <QObject>

#include "geometry.h"
#include "math-pointf.h"

namespace Kolf
{
	class Ball;
	class Overlay;

	/**
	 * \class Object
	 * \since 2.0
	 *
	 * Abstract base class for any visible object on Kolf's game views. It provides property, overlay, and geometry management.
	 *
	 * For the object representation, subclasses have to reimplement the createOverlay() method of this class. Furthermore, to participate in collision, one has to create a Kolf::Geometry for this object. (Note that Kolf::Geometry instances are automatically bound to the object when they're created.)
	 *
	 * All properties of the object (such as geometry parameters) have to be exported through the Q_PROPERTY macro. By default, the Object is rendered empty; add subitems or reimplement boundingRect() and paint() to make the object appear on screen.
	 * \sa Kolf::Geometry, Kolf::Overlay, BIND_KOLF_OBJECT
	 */
	class Object : public QObject, public QGraphicsItem
	{
		Q_OBJECT
		Q_PROPERTY(QColor Color READ color WRITE setColor)
		Q_PROPERTY(Kolf::PointF Position READ position WRITE setPosition)
		Q_PROPERTY(int Rotation READ rotation WRITE setRotation)
		//velocity is not a property because it should not be accessible through the editor
		public:
			///Defines how an object participates in the physical simulation.
			///\sa simulationFlags()
			enum SimulationFlag
			{
				NoSimulation = 0x0, ///< The object is immovable and does not interact with any other objects.
				KinematicSimulation = 0x1, ///< The object moves according to its kinematic state.
				DynamicSimulation = 0x2, ///< Other objects can collide with the geometry of this object.
				CustomInteraction = 0x10 ///< The behavior for resolving collisions with other movable objects is provided by the Kolf::Object::doCustomInteraction() method. You have to use this together with DynamicSimulation.
			};
			Q_DECLARE_FLAGS(SimulationFlags, SimulationFlag)
			///Defines some event types that Kolf::Object subclasses might handle.
			///\sa objectEvent()
			enum EventType
			{
				PositionChangeEvent, ///< The object's position has changed. The new position as a Kolf::PointF is given in the second argument of the objectEvent() method.
				RotationChangeEvent, ///< The object's rotation has changed. The new rotation is given in degrees in the second argument of the objectEvent() method.
				VelocityChangeEvent, ///< The object's velocity has changed. The new velocity as a Kolf::VectorF is given in the second argument of the objectEvent() method.
				ColorChangeEvent     ///< The object's color has changed. The new color as a QColor instance is given in the second argument of the objectEvent() method.
			};

			Object(dWorldID worldID, dSpaceID spaceID);
			virtual ~Object();

			///Returns a reference to the object's geometry (or 0 if the object has no geometry, e.g. because it does not participate in collision detection).
			const Kolf::Geometry* geometry() const;
			///Returns a reference to the overlay for this object. The overlay is created once it is accessed for the first time (unless \a createIfNecessary is set to false, which causes the method to return 0 if no overlay has been created yet). Subclasses should also use this method to get to the own overlay.
			Kolf::Overlay* overlay(bool createIfNecessary = true);

			///Returns whether the object's properties can be manipulated in the editor interface. The default is true, as most objects are expected to be editable.
			virtual bool isEditable() const;
			///Returns whether this object is rotateable. The default is true, but reimplement this as a function that returns false if it does not make any sense for your object to be rotated (e.g. for perfectly round objects, or objects that are rotateable by their nature).
			virtual bool isRotateable() const;

			///Returns the object's color.
			QColor color() const;
			///Returns the object's position, i.e. the position of the reference point of this object (which is also the base position of the object's geometry). The topology of the reference point depends on the object type and implementation.
			Kolf::PointF position() const;
			///Returns the object's rotation angle (around the Z axis) in degrees.
			int rotation() const;
			///Returns the object's velocity (more exactly: the velocity of the object's reference point).
			Kolf::VectorF velocity() const;
			///Sets the object's color.
			void setColor(const QColor& color);
			///Sets the object's position, i.e. the position of the reference point of this object (which is also the base position of the object's geometry). The topology of the reference point depends on the object type and implementation.
			void setPosition(const Kolf::PointF& position);
			///Sets the object's rotation angle (around the Z axis) which has to be given in degrees.
			void setRotation(int angle);
			///Sets the object's velocity (more exactly: the velocity of the object's reference point).
			void setVelocity(const Kolf::VectorF& velocity);

			///Calling this function will set this object in update-blocking mode: Updates on the objects' properties will not be propagated to the Kolf::ManagedObject instance for this object.
			void blockUpdates();
			///Stop blocking updates for this object. If the parameter is set to true, the changes that were made to the object's properties during the blocking mode was active, will be propagated.
			///\see blockUpdates
			void unblockUpdates(bool propagateChangesAsUpdate = false);

			///Defines how an object participates in the physical simulation. The default is Kolf::Object::DynamicSimulation, as we expect most objects to be immovable, while still participating in collisions.
			///\sa SimulationFlag
			virtual Kolf::Object::SimulationFlags simulationFlags() const;
			///If the simulationFlags() contain CustomInteraction, this method will have to provide the custom interaction. The default implementation does nothing (because the CustomInteraction flag is not set by default).
			virtual void doCustomInteraction(Kolf::Object* object);
			///The physics engine calls this method to prepare the object for the following simulation step.
			void prepareKinematicState();
			///The physics engine calls this method after calculating the next frame, to let the objects update their representation and generate objectEvents for the changed kinematic state (this includes position and velocity).
			void updateKinematicState();

			///Returns all subobjects of this object. The object-subobject relationship is a merely hierarchical one, the subobject's coordinate system is not related to that of the parent object.
			QList<Kolf::Object*> subObjects() const;
			///Adds a new subobject to this object. This object will take ownership of the subobject. Note that the properties of subobjects are regarded as being managed completely by the parent object: For instance, they are not loaded and saved directly by the ManagedObjectGroup, and they cannot be edited directly because their overlays are not activated.
			void addSubObject(Kolf::Object* object);
			///Removes the given subobject.
			void removeSubObject(Kolf::Object* object);

			///Creates the 3D-geometry of the object, placed in a separate method to allow a compile without OpenGL (don't know how to change the compiler options to skip when no gl.h found) and to skip the creation of the arrays when the objects become complex and only the 2D-view is active.
			virtual void createSpatialGeometry();
			///Returns the vertex and normal arrays
			virtual float* getNormalGeometry();
			virtual float* getVertexGeometry();
			void rotateNormalArray(int numItems, float* array);
			void rotateVertexArray(int numItems, float* array);
			//Counter to keep the number of items in the arrays.
			int numVerts;

			virtual QRectF boundingRect() const;
			virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget = 0);
		Q_SIGNALS:
			///This signal is emitted whenever the value of an object's property changes.
			void updated();
			///It's so obvious that it hurts.
			void subObjectAdded(Kolf::Object* subObject);
			///It's so obvious that it hurts.
			void subObjectAboutToBeRemoved(Kolf::Object* subObject);
		protected:
			///The base class uses this event handler to propagate several internal events to the subclass implementation. See the Kolf::Object::EventType enumeration for which events are defined, and which data is passed with them. The default implementation does nothing.
			virtual void objectEvent(int type, const QVariant& data);
			///Creates the optimal overlay for this object. The implementation does not have to propagate its properties to the overlay, as the overlay is updated just after it has been created.
			virtual Kolf::Overlay* createOverlay() = 0;
			///This function should be called whenever the value of an object's property changes. This will most prominently cause the overlay to be updated (if it exists) and save the new state in a Kolf::ManagedObject, if applicable.
			void propagateUpdate();
		private:
			friend class Geometry;
			friend class ObjectRegistry;

			dBodyID m_bodyID;
			dWorldID m_worldID;
			dSpaceID m_spaceID;
			Kolf::Geometry* m_geometry;
			Kolf::Overlay* m_overlay;
			QList<Kolf::Object*> m_subObjects;
			///The QGraphicsItem does not allow to get the rotation angle, we have to cache it here to properly rotate the QGraphicsItem part if the rotation angle changes in ODE or through setRotation.
			int m_2DRotationAngle;
			bool m_blockUpdates;
			QColor m_color;
	};
}

Q_DECLARE_OPERATORS_FOR_FLAGS(Kolf::Object::SimulationFlags)

#endif // KOLF_OBJECT_H
