/***************************************************************************
 *   Copyright 2008-2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "game-normal.h"
#include "component-basicviews.h"
#include "component-scorecard.h"
#include "course.h"
#include "player-local.h"
#include "../elements/ball.h"
#include "../elements/putter.h"
#include "../elements/scene.h"

#include <QApplication>
#include <QTimer>
#include <KAction>
#include <KActionCollection>
#include <KIcon>
#include <KLocalizedString>

//BEGIN Kolf::NormalGameUI

Kolf::NormalGameUI::NormalGameUI()
	: Kolf::GameUI(QLatin1String("kolf-normal"))
	, m_scoreCard(new Kolf::ScoreCardComponent)
{
	//action: skip course
	const KIcon skipIcon(QApplication::isLeftToRight() ? "go-next-view" : "go-previous-view");
	KAction* skipAct = new KAction(skipIcon, i18nc("Title of toolbar action", "Skip hole"), actionCollection());
	skipAct->setToolTip(i18n("Clear score for this hole and proceed to the next hole"));
	actionCollection()->addAction("kolf_level_skip", skipAct);
	connect(skipAct, SIGNAL(triggered()), this, SIGNAL(actionSkipHole()));
	//dock widget: score card
	addDockWidget(m_scoreCard);
	//page widgets: basic views
	addPageWidget(new Kolf::View2DComponent);
	addPageWidget(new Kolf::View3DComponent);
}

void Kolf::NormalGameUI::lateInit(Kolf::NormalGame* game)
{
	m_scoreCard->setGame(game);
}

//END Kolf::NormalGameUI

//BEGIN Kolf::NormalGame

Kolf::NormalGame::NormalGame(Kolf::Course* course)
	: Kolf::Game(course, new Kolf::NormalGameUI)
	, m_currentPlayer(-1)
{
	connect(gameUI(), SIGNAL(actionSkipHole()), this, SLOT(actionSkipHole()));
	reinterpret_cast<Kolf::NormalGameUI*>(gameUI())->lateInit(this);
}

Kolf::NormalGame::~NormalGame()
{
	//TODO: necessary?
	currentHoleAboutToBeChanged(); //deinitialize everything
}

Kolf::LocalPlayer* Kolf::NormalGame::addPlayer()
{
	Kolf::LocalPlayer* newPlayer = new Kolf::LocalPlayer(QString(), course()->holeCount());
	Kolf::Game::addPlayer(newPlayer);
	return newPlayer;
}

bool Kolf::NormalGame::startGame()
{
	if (players().count() == 0 || course()->holeCount() <= 0)
		return false;
	gameUI()->setWindowTitle(course()->title());
	//start with first hole and first player
	setCurrentHole(0);
	return true;
}

void Kolf::NormalGame::finishMove()
{
	//all balls are standing still - turn control to next player
	emit playerMoveEnded(m_currentPlayer);
	QTimer::singleShot(500, this, SLOT(advancePlayer()));
}

void Kolf::NormalGame::advanceHole()
{
	setCurrentHole(currentHole() + 1);
}

//TODO: base play order on success at last hole (like in Kolf 1)
void Kolf::NormalGame::advancePlayer()
{
	QList<Kolf::Player*> players = this->players();
	//find next player (i.e. a player who's ball is not in a hole yet)
	const int oldPlayer = m_currentPlayer;
	int newPlayer = m_currentPlayer;
	do
	{
		newPlayer = (newPlayer + 1) % players.count();
		if (newPlayer == oldPlayer && players[newPlayer]->ball()->isInHole())
		{
			//looped through all players -> no player can play -> all players are in the hole -> proceed to the next hole
			QTimer::singleShot(500, this, SLOT(advanceHole()));
			return;
		}
	}
	while (players[newPlayer]->ball()->isInHole());
	//turn control to this player
	m_currentPlayer = newPlayer;
	emit playerMoveStarted(newPlayer);
	players[newPlayer]->giveControl();
}

void Kolf::NormalGame::currentHoleAboutToBeChanged()
{
	if (currentScene())
		disconnect(currentScene()->physicsEngine(), SIGNAL(turnEnded()), this, SLOT(finishMove()));
}

void Kolf::NormalGame::currentHoleChanged()
{
	Kolf::Scene* scene = currentScene();
	if (scene)
	{
		connect(scene->physicsEngine(), SIGNAL(turnEnded()), this, SLOT(finishMove()));
		//start with first player
		m_currentPlayer = -1;
		QTimer::singleShot(500, this, SLOT(advancePlayer()));
	}
	else
		emit gameEnded();
}

void Kolf::NormalGame::actionSkipHole()
{
	//FIXME: disable action when balls are moving (requires a Kolf::Scene::turnStarted signal)
	//reset scores on this hole
	foreach (Kolf::Player* player, players())
		player->setCurrentScore(0);
	//go to next hole
	setCurrentHole(currentHole() + 1);
}

//END Kolf::NormalGame

#include "game-normal.moc"
