/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_GAMEUI_H
#define KOLF_GAMEUI_H

#include <QMap>
#include <QObject>
class KActionCollection;
class KTabWidget;
class KXMLGUIClient;
class KXmlGuiWindow;

namespace Kolf
{
	/**
	 * \struct GameUIParent
	 * \since 2.0
	 *
	 * This type describes an environment into which a Kolf::GameUI can be inserted. The KXmlGuiWindow will grab the action collection and the dock widgets of the GameUI, and the KTabWidget grabs all tab widgets defined by the GameUI.
	 */
	struct GameUIParent
	{
		public:
			GameUIParent(KXmlGuiWindow* window, KTabWidget* tabWidget);

			KXmlGuiWindow* window;
			KTabWidget* tabWidget;
	};

	class GameUIDockWidget;
	class GameUIPageWidget;
	class Scene;

	/**
	 * \class GameUI
	 * \since 2.0
	 *
	 * This class describes the interface of a game type. This interface can consist of actions (stored in a KXMLGUIClient's action collection) and  dock widgets. Events should be forwarded to the corresponding Kolf::Game instance through signal-slot connections.
	 */
	class GameUI : public QObject
	{
		Q_OBJECT
		public:
			virtual ~GameUI();

			///Returns the UI parent that is currently displaying this interface, or 0 if this interface is not shown currently.
			Kolf::GameUIParent uiParent() const;
			///Insert this interface into the given \a uiParent. This will remove the interface from any earlier parents.
			void setUIParent(const Kolf::GameUIParent& uiParent);
			///Remove this interface from its current UI parent.
			void unsetUIParent();
		protected:
			///Creates a new GameUI instance. The given \a interfaceName corresponds to the name of the XMLGUI description file for this UI (if the interface name is "abc", the XML file name has to be "abcui.rc").
			GameUI(const QString& interfaceName);

			///Returns the action collection of this interface. Use this to add actions to the GUI parent's toolbar and menus.
			///\note Changes will appear in the UI parent immediately, if a UI parent is currently assigned to this UI.
			KActionCollection* actionCollection();
			///Adds the given \a dockWidget to the interface. (This will do nothing if a null pointer is given.)
			///\note Changes will appear in the UI parent immediately, if a UI parent is currently assigned to this UI. The GameUI instance will take ownership of the dock widget, until it is manually removed from the UI with the removeDockWidget() method.
			void addDockWidget(Kolf::GameUIDockWidget* dockWidget);
			///Removes the given \a dockWidget from the interface. (This will do nothing if the given pointer does not belong to a dock widget that currently belongs to the interface.)
			///\note Changes will appear in the UI parent immediately, if a UI parent is currently assigned to this UI. It is especially safe to delete the \a dockWidget after this operation.
			void removeDockWidget(Kolf::GameUIDockWidget* dockWidget);
			///Adds the given \a pageWidget to the interface. It will appear in the central tab area.
			///\note Changes will appear in the UI parent immediately, if a UI parent is currently assigned to this UI. The GameUI instance will take ownership of the page widget, until it is manually removed from the UI with the removePageWidget() method.
			void addPageWidget(Kolf::GameUIPageWidget* pageWidget);
			///Removes the given \a widget from the interface.
			///\note Changes will appear in the UI parent immediately, if a UI parent is currently assigned to this UI. It is especially safe to delete the \a pageWidget after this operation.
			void removePageWidget(Kolf::GameUIPageWidget* pageWidget);
		public Q_SLOTS:
			///Sets the window title.
			///Do not include the application name (this is added automatically), nor some "[Changed]" flag (use setClean() instead).
			void setWindowTitle(const QString& title);
			///Sets whether all changes have been saved to disk. (A "[Changed]" flag will be displayed in the window title when the game is not clean.)
			void setClean(bool clean);
		Q_SIGNALS:
			///This internal signal transfers scene change events. It is automatically connected to the setScene() slots of any page and dock widgets in this game UI.
			void sceneChanged(Kolf::Scene* scene);
		private:
			Kolf::GameUIParent m_uiParent;
			KXMLGUIClient* m_uiClient;
			QList<Kolf::GameUIDockWidget*> m_dockWidgets;
			QList<Kolf::GameUIPageWidget*> m_pageWidgets;
			QString m_windowTitle;
			bool m_clean;
	};
}

#endif // KOLF_GAMEUI_H
