/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "scoremodel.h"
#include "course.h"
#include "game.h"
#include "player.h"

#include <QApplication>
#include <KColorScheme>
#include <KIcon>
#include <KLocalizedString>

Kolf::ScoreModel::ScoreModel()
	: m_game(0)
{
}

Kolf::Game* Kolf::ScoreModel::game() const
{
	return m_game;
}

void Kolf::ScoreModel::setGame(Kolf::Game* game)
{
	if (m_game == game)
		return;
	if (m_game)
	{
		disconnect(m_game, 0, this, 0);
		foreach (Kolf::Player* player, m_game->players())
			disconnect(player, 0, this, 0);
	}
	m_game = game;
	if (game)
	{
		connect(game, SIGNAL(currentHoleChanged(int, int)), this, SLOT(currentHoleChanged(int, int)));
		connect(game, SIGNAL(playerAdded(Kolf::Player*)), this, SLOT(playerAdded(Kolf::Player*)));
		foreach (Kolf::Player* player, game->players())
			connect(player, SIGNAL(scoreChanged(Kolf::Player*, int)), this, SLOT(scoreChanged(Kolf::Player*, int)));
	}
	reset();
}

int Kolf::ScoreModel::rowCount(const QModelIndex& parent) const
{
	if (!m_game)
		return 0;
	if (parent.isValid())
		return 0;
	return m_game->course()->holeCount() + 1; //"+1" because the last row contains the total score
}

int Kolf::ScoreModel::columnCount(const QModelIndex& parent) const
{
	if (!m_game)
		return 0;
	if (parent.isValid())
		return 0;
	return m_game->players().count() + 1; //"+1" because of the parPlayer
}

Qt::ItemFlags Kolf::ScoreModel::flags(const QModelIndex& index) const
{
	Q_UNUSED(index)
	return Qt::ItemIsEnabled;
}

QVariant Kolf::ScoreModel::data(const QModelIndex& index, int role) const
{
	if (!m_game)
		return QVariant();
	if (index.parent().isValid())
		return QVariant();
	const int holeIndex = index.row();
	const int playerIndex = index.column() - 1;
	//column selects player
	const Kolf::Player* const player = (playerIndex == -1) ? m_game->parPlayer() : m_game->players()[playerIndex];
	//variable declarations (cannot be done in switch statement, at least in gcc)
	KColorScheme::BackgroundRole bgRole = KColorScheme::NormalBackground;
	//retrieve data
	switch (role)
	{
		case Qt::DisplayRole:
			//show score
			if (holeIndex == m_game->course()->holeCount())
				return player->totalScore();
			else
			{
				if (player == m_game->parPlayer() || holeIndex >= m_game->currentHole())
					return player->score(holeIndex);
				else
				{
					//player has finished this hole -> show difference to par together with score
					return i18nc("used as \"score (difference)\" on the score card", "%1 (%2)",
						player->score(holeIndex),
						player->score(holeIndex) - m_game->parPlayer()->score(holeIndex)
					);
				}
			}
		case Qt::BackgroundRole:
			//mark current hole
			if (holeIndex == m_game->currentHole())
				bgRole = KColorScheme::NeutralBackground;
			//create brush
			return KStatefulBrush(KColorScheme::View, bgRole).brush(QPalette::Active);
		default:
			return QVariant();
	}
}

QVariant Kolf::ScoreModel::headerData(int section, Qt::Orientation orientation, int role) const
{
	if (!m_game)
		return QVariant();
	switch (role)
	{
		case Qt::DisplayRole:
			if (orientation == Qt::Vertical)
			{
				if (section == m_game->course()->holeCount())
					return i18nc("as in: total score", "Total");
				else
					return i18n("Hole %1", section + 1); //"+1" because user-visible hole numbers start at 1
			}
			else //orientation == Qt::Horizontal
			{
				if (section == 0)
					return m_game->parPlayer()->name();
				else
					return m_game->players()[section - 1]->name();
			}
		case Qt::DecorationRole:
			//mark current hole
			if (orientation == Qt::Vertical && section == m_game->currentHole())
				return KIcon(QApplication::isLeftToRight() ? "arrow-right" : "arrow-left");
			else
				return QVariant();
		case Qt::BackgroundRole:
			//show player color
			if (orientation == Qt::Horizontal && section != 0) //section == 0 is the par player
				return QColor::fromHsv(m_game->players()[section - 1]->hue(), 255, 255);
		default:
			return QVariant();
	}
}

void Kolf::ScoreModel::playerAdded(Kolf::Player* player)
{
	const int playerCount = m_game->players().count();
	beginInsertColumns(QModelIndex(), playerCount, playerCount);
	endInsertColumns();
	connect(player, SIGNAL(scoreChanged(Kolf::Player*, int)), this, SLOT(scoreChanged(Kolf::Player*, int)));
}

void Kolf::ScoreModel::currentHoleChanged(int newHoleIndex, int oldHoleIndex)
{
	const int holeCount = m_game->course()->holeCount();
	const int playerCount = m_game->players().count();
	//report header data changes
	if (oldHoleIndex >= 0 && oldHoleIndex < holeCount)
		emit headerDataChanged(Qt::Vertical, oldHoleIndex, oldHoleIndex);
	if (newHoleIndex >= 0 && newHoleIndex < holeCount)
		emit headerDataChanged(Qt::Vertical, newHoleIndex, newHoleIndex);
	//report data change for the data row of the old hole (background brush has changed)
	QModelIndex startIndex = index(oldHoleIndex, 0);
	QModelIndex endIndex = index(oldHoleIndex, playerCount + 1);
	emit dataChanged(startIndex, endIndex);
	//report data change for the data row of the new hole (background brush has changed)
	startIndex = index(newHoleIndex, 0);
	endIndex = index(newHoleIndex, playerCount + 1);
	emit dataChanged(startIndex, endIndex);
}

void Kolf::ScoreModel::scoreChanged(Kolf::Player* player, int holeIndex)
{
	//find player
	int playerIndex = m_game->players().indexOf(player);
	if (playerIndex < 0)
		return;
	//report data change for the single score
	QModelIndex dataIndex = index(holeIndex, playerIndex + 1);
	emit dataChanged(dataIndex, dataIndex);
	//report data change for the total score
	dataIndex = index(m_game->course()->holeCount() + 1, playerIndex + 1);
	emit dataChanged(dataIndex, dataIndex);
}

#include "scoremodel.moc"
