/***************************************************************************
 *   Copyright 2009 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "newgame-playerlist.h"
#include "newgame-playerwidget.h"
#include "../engine/game-normal.h"
#include "../engine/player-local.h"

#include <QGridLayout>
#include <QPushButton>
#include <QSpacerItem>
#include <KIcon>
#include <KLocalizedString>

Kolf::NewGamePlayerList::NewGamePlayerList()
	: m_layout(new QGridLayout)
	, m_bottomSpacer(new QSpacerItem(1, 1, QSizePolicy::Expanding, QSizePolicy::Expanding))
{
	//create consistent layout state
	m_layout->addItem(m_bottomSpacer, 0, 0);
	setLayout(m_layout);
	//create first two players
	addPlayer();
	addPlayer();
}

QColor Kolf__NewGamePlayerList__defaultColorFromRow(int row)
{
	//default colors are repeated every twelve players
	if (row >= 12)
		return Kolf__NewGamePlayerList__defaultColorFromRow(row - 12);
	int hue;
	//red/green/blue for the first three players (hue = 0 or 120 or 240)
	if (row < 3)
		hue = 120 * row;
	//yellow/turquoise/violet for the next three players (hue = 60 or 180 or 300)
	else if (row < 6)
		hue = 120 * (row - 3) + 60;
	//the colors between these colors for the last six players (hue = 30 or 90 or 150 or 210 or 270 or 330)
	else
		hue = 60 * (row - 6) + 30;
	//very strong, saturated default ball colors
	return QColor::fromHsv(hue, 255, 255);
}

void Kolf::NewGamePlayerList::addPlayer()
{
	//add new player
	int newRow = m_playerWidgets.count();
	Kolf::NewGamePlayerWidget* playerWidget = new Kolf::NewGamePlayerWidget;
	playerWidget->setDefaultName(i18n("Player %1", newRow + 1));
	playerWidget->setDefaultColor(Kolf__NewGamePlayerList__defaultColorFromRow(newRow));
	m_playerWidgets << playerWidget;
	QPushButton* removeButton = new QPushButton(KIcon("list-remove-user"), i18n("Remove this player"));
	connect(removeButton, SIGNAL(clicked()), this, SLOT(removePlayer()));
	m_removeButtons << removeButton;
	//add widgets to layout, and move spacer down
	m_layout->removeItem(m_bottomSpacer);
	m_layout->addWidget(playerWidget, newRow, 0);
	m_layout->addWidget(removeButton, newRow, 1);
	m_layout->addItem(m_bottomSpacer, newRow + 1, 0);
	//removePlayer() has disabled the remove button of the first player if he was the only one -> reenable it
	m_removeButtons.at(0)->setEnabled(true);
}

void Kolf::NewGamePlayerList::removePlayer()
{
	QPushButton* removeButton = qobject_cast<QPushButton*>(sender());
	if (!removeButton)
		return;
	//remove player from lists
	int index = m_removeButtons.indexOf(removeButton);
	Kolf::NewGamePlayerWidget* playerWidget = m_playerWidgets.takeAt(index);
	m_removeButtons.removeAt(index);
	//schedule widget deletion (do not delete directly because it's not nice to delete a QObject in its event handler)
	playerWidget->deleteLater();
	removeButton->deleteLater();
	//move widgets below this index up in layout
	for (int i = index; i < m_playerWidgets.count(); ++i)
	{
		Kolf::NewGamePlayerWidget* otherPlayerWidget = m_playerWidgets.at(i);
		QPushButton* otherRemoveButton = m_removeButtons.at(i);
		m_layout->removeWidget(otherPlayerWidget);
		m_layout->removeWidget(otherRemoveButton);
		m_layout->addWidget(otherPlayerWidget, i, 0);
		m_layout->addWidget(otherRemoveButton, i, 1);
	}
	m_layout->removeItem(m_bottomSpacer);
	m_layout->addItem(m_bottomSpacer, m_playerWidgets.count(), 0);
	//if only one player is left, disable his remove button
	if (m_removeButtons.count() == 1)
		m_removeButtons.at(0)->setEnabled(false);
}

void Kolf::NewGamePlayerList::writePlayers(Kolf::NormalGame* game)
{
	foreach (Kolf::NewGamePlayerWidget* playerWidget, m_playerWidgets)
	{
		Kolf::LocalPlayer* player = game->addPlayer();
		player->setName(playerWidget->name());
		player->setHue(playerWidget->color().hue());
	}
}

#include "newgame-playerlist.moc"
