/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_MANAGEDOBJECT_H
#define KOLF_MANAGEDOBJECT_H

#include "utils-undostack.h"

#include <QMap>
#include <QPointer>
class KConfigGroup;

namespace Kolf
{
	class Object;
	class ObjectFactory;

	/**
	 * \class ManagedObject
	 * \since 2.0
	 *
	 * This is a layer over Kolf::Object, which provides state management.
	 */
	class ManagedObject : public Utils::UndoStack
	{
		Q_OBJECT
		private:
			//forward declaration of inner class (for declaration, see below this class declaration)
			class State;
		public:
			virtual ~ManagedObject();

			///Returns the Kolf::Object instance which is managed by this Kolf::ManagedObject instance, or 0 if the object is not existing in the current state.
			Kolf::Object* object() const;
			///Returns the \a index-th state of this managed object. Note that index == 0 might give an empty pre-creation state, and index == stateCount() might give a post-deletion state if this object has been deleted (or was deleted, but this was undone).
			Kolf::ManagedObject::State* state(int index) const;

			///Writes this state snapshot into the given KConfigGroup.
			void serialize(KConfigGroup* group);
			///Read a state snapshot from the given KConfigGroup. A null pointer is returned if the format of this group is empty.
			static Kolf::ManagedObject* deserialize(KConfigGroup* group, Kolf::ObjectFactory* objectFactory);

			///Sets whether pushState() may be aborted if the state is not significantly different from the current one. (The new state will then overwrite the current one instead.)
			void setAllowStateMerging(bool allowStateMerging);
		public Q_SLOTS:
			///Saves the object's current state in the managed object's history. If the current state of the object is not the last one in this history, all following states will be dropped by inserting the new state (as one expects from an undo stack).
			void pushState();
		Q_SIGNALS:
			///This signal is emitted after the Kolf::Object instance of this managed object has been reincarnated through an undo or redo operation.
			void objectCreated(Kolf::ManagedObject* object);
			///This signal is emitted after the Kolf::Object instance of this managed object is about to be deleted through an undo or redo operation.
			void objectAboutToBeDeleted(Kolf::ManagedObject* object);
		protected:
			virtual void applyState(int index);
			virtual void removeStatesAbove(int index);
		private:
			class State;
			///Used by Kolf::ManagedObject::deserialize.
			ManagedObject(Kolf::ManagedObject::State* state, Kolf::ObjectFactory* objectFactory);

			mutable QPointer<Kolf::Object> m_object; //may need to be recreated in the const method object()
			Kolf::ObjectFactory* m_objectFactory;
			QList<Kolf::ManagedObject::State*> m_states;
			bool m_allowStateMerging;
	};
}

/**
 * \class ManagedObject::State
 * \since 2.0
 *
 * This object saves the complete state of a Kolf::Object, that is: the class name and the values of all QMetaProperties. It is internally used by Kolf::ManagedObject and Kolf::ManagedObjectGroup.
 *
 * \sa Kolf::ManagedObject
 */
class Kolf::ManagedObject::State
{
	Q_DISABLE_COPY(State)
	public:
		///Constructs an empty state snapshot.
		///\sa isEmpty()
		State();
		///Creates a new state snapshot of the given \a object, or an empty state snapshot if a null pointer is given.
		static Kolf::ManagedObject::State* fromObject(Kolf::Object* object);

		///Returns the class name for this state.
		QByteArray className() const;
		///Returns whether this state snapshot is empty. An empty state snapshot is regularly used to represent objects that were deleted or have not yet been created.
		bool isEmpty() const;
		///Finds whether the difference between the given three states of one object (specified in chronological order) is enough to justify that all three states are saved in a managed object.
		///The relevant condition is that some properties change between \a previousState and \a mediacyState which do not change between \a mediacyState and \a followingState. Of course, something _has_ to change between \a mediacyState and \a followingState, too.
		static bool isMediacySignificant(Kolf::ManagedObject::State* previousState, Kolf::ManagedObject::State* mediacyState, Kolf::ManagedObject::State* followingState);

		///Applies this state to the given \a object. If the object's class is the same as the object which this state snapshot was taken from, the object will afterwards be in an exactly identical state.
		void applyToObject(Kolf::Object* object) const;

		///Writes this state snapshot into the given KConfigGroup.
		void serialize(KConfigGroup* group);
		///Read a state snapshot from the given KConfigGroup into a new Kolf::ManagedObject::State snapshot. A null pointer is returned if the format of this group is empty.
		static Kolf::ManagedObject::State* deserialize(KConfigGroup* group);
	private:
		///Creates a new state snapshot of the given \a object.
		///\warning This method assumes the validity of the given pointer.
		State(Kolf::Object* object);
		///Helper function for Kolf::ManagedObject::State::isMediacySignificant.
		static QList<QByteArray> findChangingProperties(Kolf::ManagedObject::State* previousState, Kolf::ManagedObject::State* nextState);

		QByteArray m_className;
		mutable QMap<QByteArray, QVariant> m_propertyValues; //mutability allows efficient QVariant casting in const methods
};

#endif // KOLF_MANAGEDOBJECT_H
