/***************************************************************************
 *   Copyright 2008, 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef KOLF_OBJECT_H
#define KOLF_OBJECT_H

#include <QGraphicsItem>

class b2Body;
class b2World;

namespace Kolf
{
	class ManagedObject;
	class ObjectFactory;
	class Overlay;
	class Scene;
	class Shape;

	/**
	 * \class Object
	 * \since 2.0
	 *
	 * Abstract base class for any visible and invisible object on Kolf's game views. It provides property, overlay, and geometry management.
	 *
	 * Subclass implementations have to:
	 *  \li add subitems or reimplement boundingRect() and paint() to make the object appear on screen
	 *  \li add Kolf::Shape instances to participate in dynamical and collision simulation
	 *  \li implement createOverlay() to provide an editing interface
	 *  \li export all editable properties of the object (such as geometry parameters) with the Q_PROPERTY macro, and call propagateUpdate() when one of these properties changes
	 *
	 * \sa REGISTER_KOLF_OBJECT
	 */
	class Object : public QObject, public QGraphicsItem //NOTE: We are *not* subclassing QGraphicsObject to keep the amount of properties that is written into level files reasonably small.
	{
		Q_OBJECT
		Q_PROPERTY(QPointF Position READ pos WRITE setPos)
		Q_PROPERTY(qreal Rotation READ rotation WRITE setRotation)
		Q_INTERFACES(QGraphicsItem)
		//velocity is not a property because it should not be accessible through the editor
		public:
			enum SimulationFlag
			{
				CollisionFlag = 1 << 0,
				KinematicSimulationFlag = 1 << 1,
				DynamicSimulationFlag = 1 << 2
			};
			enum SimulationType
			{
				///The object is immovable.
				NoSimulation = 0,
				///The object is immovable, but other objects can interact with it.
				CollisionSimulation = CollisionFlag,
				///The object moves according to its kinematic state.
				KinematicSimulation = CollisionSimulation | KinematicSimulationFlag,
				///Other objects can collide with the shape of this object.
				DynamicSimulation = KinematicSimulation | DynamicSimulationFlag
			};
			enum EventType
			{
				PositionChangeEvent,
				RotationChangeEvent,
				VelocityChangeEvent
			};

			///This constructor should only be used internally by subclasses. All subclasses are expected to offer a structurally equivalent constructor.
			Object(b2World* world);
			virtual ~Object();

			///Returns a Kolf::ObjectFactory which can be used to create more objects on the same scene, or a null pointer if the object is not located on a Kolf::Scene at the moment. (This will not happen unless you manually mess around with the QGraphicsItem::scene().)
			Kolf::ObjectFactory* objectFactory() const;

			//====== physics simulation ======
			///Returns a list of all Kolf::Shape instances attached to this object.
			QList<Kolf::Shape*> shapes() const;
			///Returns all subobjects of this object. The object-subobject relationship is a merely hierarchical one, the subobject's coordinate system is not related to that of the parent object.
			QList<Kolf::Object*> subObjects() const;
			///Returns how this object participates in physical simulations.
			Kolf::Object::SimulationType simulationType() const;
			///Returns the object's velocity (more exactly: the velocity of the object's reference point).
			QPointF velocity() const;
			///Configure how this object will participate in physical simulation.
			void setSimulationType(Kolf::Object::SimulationType type);
			///Sets the object's velocity (more exactly: the velocity of the object's reference point).
			void setVelocity(const QPointF& velocity);
			///Works exactly like QGraphicsItem::setRotation.
			///\warning Always use this method instead of QGraphicsItem::setRotation.
			void setRotation(qreal rotation);

			//====== editor management ======
			///Returns a reference to the overlay for this object. The overlay is created once it is accessed for the first time (unless \a createIfNecessary is set to false, which causes the method to return 0 if no overlay has been created yet). Subclasses should also use this method to get to the own overlay.
			Kolf::Overlay* overlay(bool createIfNecessary = true);
			///Returns whether the object's properties can be manipulated in the editor interface. The default is true, as most objects are expected to be editable.
			///\warning Any reimplementation has to return a value which only depends on the object's type, not on its current state.
			virtual bool isEditable() const;
			///Returns whether this object's overlay should add a handle for rotating automatically. The default is false.
			///\warning Any reimplementation has to return a value which only depends on the object's type, not on its current state.
			virtual bool showRotatorHandleOnOverlay() const;

			//====== QGraphicsItem reimplementation ======
			virtual QRectF boundingRect() const;
			virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget = 0);
		Q_SIGNALS:
			///This signal is emitted whenever the value of an object's property changes.
			void updated();
			///Similar to QObject::destroyed, but this signal is called just before the Kolf::Object destructor.
			void aboutToBeDestroyed(Kolf::Object* object);
		protected:
			//====== physics simulation ======
			b2Body* body() const;
			///Add a new Kolf::Shape to this object.
			void addShape(Kolf::Shape* shape);
			///Adds a new subobject to this object. This object will take ownership of the subobject. Note that the properties of subobjects are regarded as being managed completely by the parent object: For instance, they are not loaded and saved directly by the Kolf::ManagedObjectGroup, and they cannot be edited directly because their overlays are not activated.
			void addSubObject(Kolf::Object* object);
			///Removes the given subobject without deleting it.
			void removeSubObject(Kolf::Object* object);

			friend class Kolf::Scene; //for the following two methods
			///The physics engine calls this method to prepare the object for the following simulation step.
			void prepareKinematicState();
			///The physics engine calls this method after calculating the next frame, to let the objects update their representation and generate objectEvents for the changed kinematic state (this includes position and velocity).
			void updateKinematicState();

			//====== property management ======
			virtual QVariant itemChange(QGraphicsItem::GraphicsItemChange change, const QVariant& value);
			friend class Kolf::Overlay;
			///The base class uses this event handler to propagate several internal events to the subclass implementation. For events that indicate property changes, \a data will contain the new value of that property. The default implementation does nothing.
			virtual void objectEvent(Kolf::Object::EventType type, const QVariant& data);
			friend class Kolf::ManagedObject;
			///Calling this function will set this object in update-blocking mode: Updates on the objects' properties will not be propagated to the Kolf::ManagedObject instance for this object.
			void blockUpdates();
			///Stop blocking updates for this object. If the parameter is set to true, the changes that were made to the object's properties during the blocking mode was active, will be propagated.
			///\see blockUpdates
			void unblockUpdates(bool propagateChangesAsUpdate = false);

			//====== editor management ======
			///Creates the optimal overlay for this object. The implementation does not have to propagate its properties to the overlay, as the overlay is updated just after it has been created.
			///\warning Do not actually call this function from subclass implementations. Use overlay() instead.
			virtual Kolf::Overlay* createOverlay() = 0;
			///This function should be called whenever the value of an object's property changes. This will most prominently cause the overlay to be updated (if it exists) and save the new state in a Kolf::ManagedObject, if applicable.
			void propagateUpdate();
		private:
			b2Body* m_body;
			Kolf::Overlay* m_overlay;
			QList<Kolf::Shape*> m_shapes;
			QList<Kolf::Object*> m_subObjects;
			bool m_blockUpdates;
			Kolf::Object::SimulationType m_type;
			QPointF m_velocity;
			bool m_velocityChanged; //tells whether velocity has been changed programmatically after last simulation step
			//The rationale behind this variable is that Box2D sometimes has very low velocities (~1e-10) during contacts. The QPointF conversion could create rounding errors on some platforms.
	};
}

#endif // KOLF_OBJECT_H
