/***************************************************************************
 *   Copyright 2009, 2010 Stefan Majewsky <majewsky@gmx.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "scene.h"
#include "object.h"
#include "objectfactory.h"
#include "overlay.h"
#include "scene-contactlistener.h"

#include <QGraphicsSceneEvent>
#include <Box2D/Dynamics/b2World.h>

static const int FrameRate = 50.0;
static const qreal SecondsPerStep = 1.0 / 50.0;
static const int MillisecondsPerStep = 1000.0 * SecondsPerStep;

Kolf::Scene::Scene(QObject* parent)
	: QGraphicsScene(parent)
	, m_world(new b2World(b2Vec2(0, 0), true)) //parameters: no gravity, objects are allowed to sleep
	, m_contactListener(new Kolf::ContactListener)
	, m_objectFactory(new Kolf::ObjectFactory(this, m_world))
	, m_simulationTimerId(-1)
	, m_overlaysActive(false)
{
	m_world->SetContactListener(m_contactListener);
	connect(m_objectFactory, SIGNAL(objectCreated(Kolf::Object*)), this, SLOT(addObject(Kolf::Object*)));
}

Kolf::Scene::~Scene()
{
	delete m_contactListener;
	delete m_objectFactory;
	qDeleteAll(m_objects);
	delete m_world; //this has to happen after the objects, because these use references to bodies in the world
}

bool Kolf::Scene::areOverlaysActive() const
{
	return m_overlaysActive;
}

bool Kolf::Scene::isSimulationActive() const
{
	return m_simulationTimerId >= 0;
}

Kolf::ObjectFactory* Kolf::Scene::objectFactory() const
{
	return m_objectFactory;
}

QList<Kolf::Object*> Kolf::Scene::objects() const
{
	return m_objects;
}

void Kolf::Scene::setOverlaysActive(bool overlaysActive)
{
	if (m_overlaysActive == overlaysActive)
		return;
	m_overlaysActive = overlaysActive;
	foreach (Kolf::Object* object, m_objects)
		if (object->isEditable())
		{
			Kolf::Overlay* overlay = object->overlay();
			overlay->setState(Kolf::Overlay::Passive);
			overlay->setVisible(overlaysActive);
			connect(overlay, SIGNAL(stateChanged()), this, SLOT(overlayStateChanged()));
		}
}

void Kolf::Scene::startSimulation()
{
	if (m_simulationTimerId < 0)
		m_simulationTimerId = startTimer(MillisecondsPerStep);
}

void Kolf::Scene::stopSimulation()
{
	if (m_simulationTimerId >= 0)
	{
		killTimer(m_simulationTimerId);
		m_simulationTimerId = -1;
	}
}

void Kolf::Scene::addObject(Kolf::Object* object)
{
	//check if this object has been created as subobject of another object
	if (object->parent())
		return;
	//add object
	if (!m_objects.contains(object))
		m_objects << object;
	connect(object, SIGNAL(aboutToBeDestroyed(Kolf::Object*)), this, SLOT(removeObject(Kolf::Object*)));
	//after having added object to scene, restore position afterwards
	const QPointF pos = object->pos();
	addItem(object);
	object->setPos(pos);
	//if in editing mode, show overlay
	if (m_overlaysActive && object->isEditable())
	{
		Kolf::Overlay* overlay = object->overlay();
		connect(overlay, SIGNAL(stateChanged()), this, SLOT(overlayStateChanged()));
		overlay->setState(Kolf::Overlay::Passive);
		overlay->setVisible(true);
	}
}

void Kolf::Scene::removeObject(Kolf::Object* object)
{
	m_objects.removeAll(object);
	disconnect(object);
	if (m_overlaysActive && object->isEditable())
	{
		Kolf::Overlay* overlay = object->overlay();
		disconnect(overlay);
		overlay->setState(Kolf::Overlay::Passive);
		overlay->setVisible(false);
	}
	removeItem(object);
}

void Kolf::Scene::overlayStateChanged()
{
	Kolf::Overlay* overlay = qobject_cast<Kolf::Overlay*>(sender());
	if (!overlay)
		return;
	//ensure that only one overlay is active
	if (overlay->state() >= Kolf::Overlay::Active)
		foreach (Kolf::Object* otherObject, m_objects)
			if (otherObject->isEditable())
			{
				Kolf::Overlay* otherOverlay = otherObject->overlay();
				if (otherOverlay != overlay && otherOverlay->state() >= Kolf::Overlay::Active)
					otherOverlay->setState(Kolf::Overlay::Passive);
			}
}

void Kolf::Scene::mousePressEvent(QGraphicsSceneMouseEvent* event)
{
	QGraphicsScene::mousePressEvent(event);
	if (m_overlaysActive)
	{
		//if no item wanted the event (i.e. no click on overlays), interpret as click on scene, and deactivate all overlays
		if (!event->isAccepted())
		{
			event->accept();
			foreach (Kolf::Object* object, m_objects)
				if (object->isEditable())
				{
					Kolf::Overlay* overlay = object->overlay();
					if (overlay->state() >= Kolf::Overlay::Active)
						overlay->setState(Kolf::Overlay::Passive);
				}
		}
	}
}

void Kolf::Scene::timerEvent(QTimerEvent* event)
{
	if (event->timerId() == m_simulationTimerId)
	{
		//prepare simulation
		foreach (Kolf::Object* object, m_objects)
			object->prepareKinematicState();
		//step world
		m_contactListener->clear();
		m_world->Step(SecondsPerStep, 10, 10); //parametes 2/3 = iteration counts (TODO: optimize before first release)
		//conclude simulation
		foreach (Kolf::Object* object, m_objects)
			object->updateKinematicState();
		//call contact callbacks
		foreach (Kolf::ContactCallback callback, Kolf::knownContactCallbacks()) //krazy:exclude=foreach
			foreach (const Kolf::ContactData& data, m_contactListener->encounteredContacts())
				callback(data);
	}
	else
		QGraphicsScene::timerEvent(event);
}

#include "scene.moc"
