/***************************************************************************
 *   Copyright (C) 2005 by Albert Astals Cid <tsdgeos@terra.es>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "chatwidget.h"
#include "settings.h"
#include "protocol.h"

#include <QList>
#include <QTime>

#include <kdebug.h>
#include <kglobal.h>
#include <klocale.h>

chatWidget::chatWidget(QWidget *parent) : QWidget(parent), m_ready(false)
{
	setupUi(this);
	setMinimumSize(100, 100);

	connect(sendButton, SIGNAL(clicked()), this, SLOT(sendMessage()));
	connect(messageLine, SIGNAL(returnPressed()), this, SLOT(sendMessage()));
	connect(messageLine, SIGNAL(textChanged(const QString&)), this, SLOT(textEditChanged(const QString&)));
}

void chatWidget::setEnabled(bool enable)
{
	sendButton->setEnabled(enable ? !messageLine->text().trimmed().isEmpty() : false);
	messageLine->setEnabled(enable);
}

void chatWidget::setIAm(const QString &nick)
{
	m_nick = nick;
}

const QStringList &chatWidget::ignoredNicks() const
{
	return m_ignoredPlayers;
}

void chatWidget::setIgnoredNicks(const QStringList &nicks)
{
	m_ignoredPlayers = nicks;
}

void chatWidget::connected(const QString &languageCode)
{
	KLocale *kloc = KGlobal::locale();
	QString lang = kloc->languageCodeToName(languageCode);
	addMessageWithTime(i18n("<font color=\"red\">* <b>%1</b> will be the language of the game</font>", lang));
}

void chatWidget::addPlayer(const QString &nick, int, bool isJoiningGame)
{
	if (isJoiningGame) addMessageWithTime(i18n("<font color=\"blue\">* <b>%1</b> has joined the game</font>", nick));
	else addMessageWithTime(i18n("<font color=\"blue\">* <b>%1</b> has entered the server</font>", nick));
}

void chatWidget::addMessage(const QString &nick, const QString &message)
{
	if (m_ignoredPlayers.indexOf(nick) == -1)
		addMessageWithTime(i18nc("%1 is the nick, %2 is the message", "&lt;<font color=\"#3176B3\">%1</font>&gt; %2", nick, message));
}

void chatWidget::startGame()
{
	addMessageWithTime(i18n("<font color=\"blue\">* The game has started</font>"));
}

void chatWidget::nextPlayer(const QString &nick, bool isItMe)
{
	if (isItMe) addMessageWithTime(i18n("<font color=\"blue\">* It is your turn</font>"));
	else addMessageWithTime(i18n("<font color=\"blue\">* It is the turn of %1</font>", nick));
	m_currentPlayer = nick;
}

void chatWidget::sendMessage()
{
	QString message = messageLine->text().trimmed();
	if (message.isEmpty()) return;
	emit typedMessage(message);
	addMessageWithTime(i18nc("%1 is the nick, %2 is the message", "&lt;%1&gt; %2", m_nick, message));
	messageLine->clear();
}

void chatWidget::textEditChanged(const QString &text)
{
	sendButton->setEnabled(!text.trimmed().isEmpty());
}

void chatWidget::addMessageWithTime(const QString &text)
{
	if (Settings::showTime())
	{
		QString timestring = KGlobal::locale()->formatTime(QTime::currentTime(), true);
		textEdit->append(QString("<font color=\"#709070\">[%1]</font> %2").arg(timestring, text));
	}
	else
		textEdit->append(text);
}

void chatWidget::gameEnded(const QString &nickFinisher, const QString &nickWinner)
{
	addMessageWithTime(i18n("<font color=\"blue\">* %1 has used all his tiles. The game has ended!<br><b>%2</b> has won!</font>", nickFinisher, nickWinner));
}

void chatWidget::removePlayer(const QString &nick, const QString &game)
{
	if (game.isEmpty())
	{
		addMessageWithTime(i18n("<font color=\"blue\">* <b>%1</b> has left the game</font>", nick));
	}
	else
	{
		addMessageWithTime(i18n("<font color=\"blue\">* <b>%1</b> has entered the game named %2</font>", nick, game));
	}
}

void chatWidget::serverDown()
{
	addMessageWithTime(i18n("<font color=\"blue\">* Disconnected from server</font>"));
}

void chatWidget::tilesRejected(int reason, const QString &nick)
{
	// TODO improve that messages !!!
	if (reason == USES_EXCLUSIVE_TILES)
	{
		if (nick == m_nick) addMessageWithTime(i18n("<font color=\"red\">* You have used a word that contains tiles to make up a exclusive tile, which is forbidden.</font>"));
		else addMessageWithTime(i18n("<font color=\"red\">* %1 has used a word that contains tiles to make up a exclusive tile, which is forbidden.</font>", nick));
	}
	else if (reason == LOST_VOTING)
	{
		if (nick == m_nick) addMessageWithTime(i18n("<font color=\"red\">* Your word was not accepted by the other players</font>"));
		else addMessageWithTime(i18n("<font color=\"red\">* %1's word was not accepted by the other players</font>", nick));
	}
}

void chatWidget::addScoreCurrentPlayer(const QStringList &wordList, int score)
{
	if (m_currentPlayer == m_nick) addMessageWithTime(i18np("<font color=\"blue\">* You made 1 point creating the words %2","<font color=\"blue\">* You made %1 points creating the words %2", score, wordList.join(", ")));
	else addMessageWithTime(i18np("<font color=\"blue\">* %2 made 1 point creating the words %3","<font color=\"blue\">* %2 made %1 points creating the words %3", score, m_currentPlayer, wordList.join(", ")));
}

void chatWidget::someonePassedHisTurn(const QString &nick)
{
	if (nick == m_nick) addMessageWithTime(i18n("<font color=\"blue\">* You skipped your turn</font>"));
	else addMessageWithTime(i18n("<font color=\"blue\">* %1 skipped their turn</font>", nick));
}

void chatWidget::someoneExchangedHisTiles(const QString &nick, int nTiles)
{
	if (nick == m_nick) addMessageWithTime(i18np("<font color=\"blue\">* You exchanged %1 tile</font>", "<font color=\"blue\">* You exchanged %1 tiles</font>", nTiles));
	else addMessageWithTime(i18np("<font color=\"blue\">* %2 exchanged %1 tile</font>", "<font color=\"blue\">* %2 exchanged %1 tiles</font>", nTiles, nick));
}

#include "chatwidget.moc"
