/***************************************************************************
 *   Copyright (C) 2005-2007 by Albert Astals Cid <aacid@kde.org>          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "client.h"
#include "protocol.h"
#include "tile.h"

#include <QVariant>
#include <QVector>
#include <QTcpSocket>

#include <kdebug.h>
#include <klocale.h>

class gameOnServer
{
	public:
		gameOnServer(const QMap<QString,QString> &data, const QString &tag)
		{
			id = data[tag + "id"].toInt();
			name = data[tag + "name"];
			players = data[tag + "numberOfPlayers"].toInt();
			spectators = data[tag + "numberOfSpectators"].toInt();
			maxplayers = data[tag + "maxPlayers"].toInt();
			started = QVariant( data[tag + "started"] ).toBool();
			allowspectators = QVariant( data[tag + "allowSpectators"] ).toBool();
			comment = data[tag + "comment"];
			language = data[tag + "language"];
		}
	
		int id;
		QString name;
		int players;
		int spectators;
		int maxplayers;
		bool started;
		bool allowspectators;
		QString comment;
		QString language;
};


client::client(const QString &nick, bool joinAsSpectator) : QAbstractTableModel(), m_nick(nick), m_spectator(joinAsSpectator), m_phase(NOT_CONNECTED)
{
	m_clientSocket = new QTcpSocket();
	connect(m_clientSocket, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(slotSocketError(QAbstractSocket::SocketError)));

	m_messageHandler.setSocket(m_clientSocket);

	connect(&m_messageHandler, SIGNAL(newMessage(const message &)), this, SLOT(processNewMessage(const message &)));
}

client::~client()
{
	if (m_clientSocket)
	{
		m_clientSocket->close();
		delete m_clientSocket;
	}
	qDeleteAll(m_gamesOnServer);
}

QString client::nick() const
{
	return m_nick;
}

void client::tryConnect(const QString &hostname, int port)
{
	m_phase = WAITING_FOR_VERSION;
	m_clientSocket->connectToHost(hostname, port);
}

void client::joinGame(int gameIndex, bool joinAsSpectator)
{
	m_spectator = joinAsSpectator;
	
	message m;
	m.type = "join";
	m.data.insert("gameId", QString::number(m_gamesOnServer[gameIndex]->id));
	m.data.insert("asSpectator", QString::number(joinAsSpectator));
	m_messageHandler.sendMessage(m);
}

client::Phase client::currentPhase() const
{
	return m_phase;
}

void client::sendTiles(const tilePositionList &tiles)
{
	message m;
	m.type = "sendingTiles";
	m.data = tiles.toMap();
	m_messageHandler.sendMessage(m);
}

void client::passTurn()
{
	message m;
	m.type = "passTurn";
	m_messageHandler.sendMessage(m);
}

void client::exchangeTiles(const QVector<tile*> &tiles)
{
	m_phase = EXCHANGING_TILES;
	
	message m;
	m.type = "exchangeTiles";
	m.data = toMap(tiles);
	m_messageHandler.sendMessage(m);
}

void client::createServerGame(const QString &name, int maxPlayers, const QString &comment, const QString &languageCode, bool allowSpectators)
{
	message m;
	m.type = "createGame";
	m.data.insert("name", name);
	m.data.insert("maxPlayers", QString::number(maxPlayers));
	m.data.insert("comment", comment);
	m.data.insert("languageCode", languageCode);
	m.data.insert("allowSpectators", QString::number(allowSpectators));
	m_messageHandler.sendMessage(m);
}

bool client::isPlayer() const
{
	return !m_spectator;
}

int client::columnCount(const QModelIndex &) const
{
	return 6;
}

QVariant client::data(const QModelIndex &index, int role) const
{
	if (!index.isValid())
		return QVariant();

	if (role != Qt::DisplayRole)
		return QVariant();

	if (index.row() < m_gamesOnServer.size())
	{
		switch (index.column())
		{
			case 0:
				return m_gamesOnServer.at(index.row())->name;
				break;
			case 1:
				return i18nc("current players / total players", "%1 / %2", m_gamesOnServer.at(index.row())->players, m_gamesOnServer.at(index.row())->maxplayers);
				break;
			case 2:
				if (m_gamesOnServer.at(index.row())->allowspectators)
					return i18nc("Spectators not allowed", "Not allowed");
				else
					return m_gamesOnServer.at(index.row())->spectators;
				break;
			case 3:
				return m_gamesOnServer.at(index.row())->started;
				break;
			case 4:
				return m_gamesOnServer.at(index.row())->comment;
				break;
			case 5:
				return m_gamesOnServer.at(index.row())->language;
				break;
			default:
				return QVariant();
		}
	}
	
	return QVariant();
}

QVariant client::headerData(int section, Qt::Orientation orientation, int role) const
{
	if (orientation != Qt::Horizontal)
		return QVariant();

	if (role != Qt::DisplayRole)
		return QVariant();

	switch (section)
	{
		case 0:
			return i18n("Name of the game");
			break;
		case 1:
			return i18n("Players");
			break;
		case 2:
			return i18n("Spectators");
		case 3:
			return i18n("Game started");
			break;
		case 4:
			return i18n("Comment");
			break;
		case 5:
			return i18n("Language");
			break;
		default:
			return QVariant();
	}
}

int client::rowCount(const QModelIndex &parent) const
{
	(void)parent;
	return m_gamesOnServer.size();
}

QString client::gameLanguage(int gameIndex) const
{
	return m_gamesOnServer.at(gameIndex)->language;
}

bool client::gameAllowsSpectators(int gameIndex) const
{
	return m_gamesOnServer.at(gameIndex)->allowspectators;
}

void client::processNewMessage(const message &m)
{
	kDebug() << "client::processNewMessage" << " " << m.type << " " << m.data << endl;
	switch (m_phase)
	{
		case NOT_CONNECTED:
			// should not be necessary, but I don't want gcc warnings
		break;

		case WAITING_FOR_VERSION:
			if (m.data.value("version").toInt() != VERSION) emit error(i18n("The server uses a protocol version this client does not understand"));
			else
			{
				message m;
				m.type = "wantToJoin";
				m.data.insert("playerType", QString::number(m_spectator));
				m_messageHandler.sendMessage(m);

				m_phase = WAITING_TO_SEND_NICK;
			}
		break;
		
		case WAITING_TO_SEND_NICK:
			if (m.type == "tooManyPlayers")
			{
				emit error(i18n("The are too many users playing in that server"));
			}
			else if (m.type == "serverDoesNotAllowSpectators")
			{
				emit error(i18n("The server does not allow any spectator"));
			}
			else if (m.type == "sendNick")
			{
				m_phase = WAITING_TO_BE_ACCEPTED;
				
				message m;
				m.type = "sendNick";
				m.data.insert("nick", m_nick);
				m_messageHandler.sendMessage(m);
			}
			else wrongResponse();
		break;

		case WAITING_TO_BE_ACCEPTED:
			if (m.type == "nickInUse")
			{
				emit error(i18n("There is someone already using that nick in that server"));
			}
			else if (m.type == "acceptedIntoServer")
			{
				m_phase = WAITING_TO_ENTER_GAME;
				int numofplayers = m.data.value("players").toInt();
				QStringList playersnotingames;
				for (int j = 0; j < numofplayers; ++j)
				{
					playersnotingames << m.data.value(QString("player%1").arg(j));
				}
				int numofgames = m.data.value("games").toInt();
				for (int j = 0; j < numofgames; ++j)
				{
					gameOnServer *gos = new gameOnServer(m.data, QString("game%1").arg(j));
					m_gamesOnServer.append(gos);
				}
				emit enteredAMultiServer(playersnotingames);
			}
			else if (m.type == "acceptedIntoGame") handleAcceptedIntoGame(m);
			else if (m.type == "tooManyPlayers")
			{
				emit error(i18n("The are too many users in the server"));
			}
			else wrongResponse();
		break;
		
		case WAITING_TO_ENTER_GAME:
			if (m.type == "acceptedIntoGame") handleAcceptedIntoGame(m);
			else if (m.type == "newGame")
			{
				beginInsertRows(QModelIndex(), m_gamesOnServer.count(), m_gamesOnServer.count());
				m_gamesOnServer.push_back(new gameOnServer(m.data, "game"));
				endInsertRows();
			}
			else if (m.type == "otherPlayerEnteredAGame")
			{
				QString nick = m.data.value("nick");
				int gameId = m.data.value("gameId").toInt();
				int row = 0;
				foreach(gameOnServer *gos, m_gamesOnServer)
				{
					if (gos->id == gameId)
					{
						gos->players += 1;
						emit dataChanged(index(row, 1), index(row, 1));
						emit removePlayer(nick, gos->name);
					}
					else row++;
				}
			}
			else if (m.type == "tooManyPlayers")
			{
				emit showErrorBox(i18n("Could not join the game as it is full."), i18n("Too many players"));
			}
			else if (m.type == "canNotCreateGame")
			{
				if (m.data.value("reason") == "tooFewMaxPlayers")
				{
					emit showErrorBox(i18n("It is not possible to create a game with fewer than 2 players."), i18n("Could not create a game"));
				}
				else if (m.data.value("reason") == "notEmptyName")
				{
					emit showErrorBox(i18n("It is not possible to create a game with an empty name."), i18n("Could not create a game"));
				}
				else if (m.data.value("reason") == "repeatedName")
				{
					emit showErrorBox(i18n("It is not possible to create a game with this name as there is already one with it."), i18n("Could not create a game"));
				}
				else
				{
					// Can not happen
					emit showErrorBox(i18n("It was not possible to create the game."), i18n("Could not create a game"));
				}
			}
			else if (m.type == "newPlayer") handleNewPlayer(m, false);
			else if (m.type == "message") handleMessage(m);
			else wrongResponse();
		break;

		case WAITING_TO_START_GAME:
			if (m.type == "startingGame")
			{
				emit startGame();
				if (!m_spectator)
				{
					QList<tile*> tiles;
					fromMap(m.data, tiles);
					foreach(tile *t, tiles)
						emit newTile(t);
				}
				m_phase = NEXT_TURN_NEEDED;

			}
			else if (m.type == "message") handleMessage(m);
			else if (m.type == "newPlayer") handleNewPlayer(m, true);
			else if (m.type == "removePlayer") handleRemovePlayer(m);
			else if (m.type == "status") handleStatus(m);
			else wrongResponse();
		break;
		
		case NEXT_TURN_NEEDED:
			if (m.type == "nextTurn") handleNextTurn(m);
			else wrongResponse();
		break;

		case PLAYING_GAME:
			if (m.type == "nextTurn") handleNextTurn(m);
			else if (m.type == "message") handleMessage(m);
			else if (m.type == "newPlayer") handleNewPlayer(m, true);
			else if (m.type == "otherPlayerTiles")
			{
				// the players that did not send the tiles enter the if here
				tilePositionList tiles;
				tiles.fromMap(m.data);
				kDebug() << "number of tiles: " << tiles.count() << endl;
				emit gotTiles(tiles);
				
				m_phase = POST_GOT_TILES;
			}
			else if (m.type == "correctWords")
			{
				// the player that sent the tiles enter the if here
				int score = m.data.value("score").toInt();
				emit scoreForCurrentPlayer(score);
				emit confirmAccept();
				
				m_phase = POST_PLAYER_CORRECT_WORDS;
			}
			else if (m.type == "incorrectWords") handleIncorrectWords(m);
			else if (m.type == "unknownWords")
			{
				// OR the player that sent the tiles enter the if here
				m_phase = TRADING_PHASE;
			}
			else if (m.type == "passedTurn")
			{
				const QString &nick = m.data.value("nick");
				emit someonePassedHisTurn(nick);
				
				m_phase = POST_PASSED_TURN;
			}
			else if (m.type == "exchangedTiles") handleExchangedTiles(m);
			else if (m.type == "removePlayer") handleRemovePlayer(m);
			else if (m.type == "status") handleStatus(m);
			else wrongResponse();
		break;
		
		case POST_GOT_TILES:
			if (m.type == "correctWords")
			{
				int score = m.data.value("score").toInt();
				emit scoreForCurrentPlayer(score);

				m_phase = POST_NOTPLAYER_CORRECT_WORDS;
			}
			else if (m.type == "unknownWords")
			{
				m_phase = TRADING_PHASE;
				emit unknownWords();
			}
			else if (m.type == "incorrectWords") handleIncorrectWords(m);
			else wrongResponse();
		break;
		
		case POST_PLAYER_CORRECT_WORDS:
			if (m.type == "gameEnded") handleGameEnded(m);
			else if (m.type == "newTiles") handleNewTiles(m);
			else wrongResponse();
		break;
		
		case POST_NOTPLAYER_CORRECT_WORDS:
			if (m.type == "nextTurn")
			{
				emit confirmAccept();
				handleNextTurn(m);
			}
			else if (m.type == "gameEnded") handleGameEnded(m);
			else wrongResponse();
		break;
		
		case POST_PASSED_TURN:
			if (m.type == "nextTurn") handleNextTurn(m);
			else wrongResponse();
		break;

		case EXCHANGING_TILES:
			if (m.type == "newExchangedTiles")
			{
				QList<tile *> tiles;
				fromMap(m.data, tiles);
				foreach(tile *t, tiles)
					emit newTile(t);
				m_phase = POST_EXCHANGING_TILES;
			}
			else wrongResponse();
		break;
		
		case POST_EXCHANGING_TILES:
			if (m.type == "exchangedTiles") handleExchangedTiles(m);
			else wrongResponse();
		break;

		case TRADING_PHASE:
			if (m.type == "correctWords")
			{
				m_phase = POST_TRADING_PHASE;

				int score = m.data.value("score").toInt();
				emit scoreForCurrentPlayer(score);
				emit confirmAccept();
			}
			else if (m.type == "incorrectWords") handleIncorrectWords(m);
			else if (m.type == "status") handleStatus(m);
			else wrongResponse();
		break;
		
		case POST_TRADING_PHASE:
			if (m.type == "nextTurn")
			{
				emit confirmAccept();
				handleNextTurn(m);
			}
			else if (m.type == "gameEnded") handleGameEnded(m);
			else if (m.type == "newTiles") handleNewTiles(m);
			else wrongResponse();
		break;
	}
}

void client::handleAcceptedIntoGame(const message &m)
{
	m_phase = WAITING_TO_START_GAME;
	emit clearPlayers();
	
	int count = m.data.value("count").toInt();
	for (int i = 0; i < count; ++i)
	{
		QString nick = m.data.value( QString("nick%1").arg(i) );
		int status = m.data.value( QString("status%1").arg(i) ).toInt();
		emit newPlayer(nick, status, true);
	}
	QString languageCode = m.data.value("language");
	emit acceptedIntoGame(languageCode);
}

void client::handleNextTurn(const message &m)
{
	QString nick = m.data.value("nick");
	emit nextPlayer(nick, nick == m_nick);
	m_phase = PLAYING_GAME;
}

void client::handleMessage(const message &m)
{
	QString nick = m.data.value("nick");
	QString text = m.data.value("message");
	emit gotMessage(nick, text);
}

void client::handleGameEnded(const message &m)
{
	m_phase = WAITING_TO_START_GAME;
	QString finisher, winner;
	finisher = m.data.value("finisher");
	winner = m.data.value("winner");
	emit gameEnded(finisher, winner);
}

void client::handleRemovePlayer(const message &m)
{
	QString nick = m.data.value("nick");
	emit removePlayer(nick, QString());
}

void client::handleNewTiles(const message &m)
{
	QList<tile *> tiles;
	fromMap(m.data, tiles);
	foreach(tile *t, tiles)
		emit newTile(t);
	
	m_phase = NEXT_TURN_NEEDED;
}

void client::handleStatus(const message &m)
{
	int which = m.data.value("status").toInt();
	QString player = m.data.value("nick");
	emit changeStatus(player, which);
}

void client::handleNewPlayer(const message &m, bool isJoiningGame)
{
	QString nick = m.data.value("nick");
	int which;
	if (isJoiningGame) which = m.data.value("status").toInt();
	else which = STATUS_SPECTATOR;
	emit newPlayer(nick, which, isJoiningGame);
}

void client::handleIncorrectWords(const message &m)
{
	const QString &reason = m.data.value("reason");
	
	if (reason != "usesExclusiveTiles" && reason != "lostVoting") wrongResponse();
	else
	{
		int intReason = reason == "usesExclusiveTiles" ? USES_EXCLUSIVE_TILES : LOST_VOTING;
		const QString &nick = m.data.value("nick");
		emit confirmReject(intReason, nick);
		
		m_phase = NEXT_TURN_NEEDED;
	}
}

void client::handleExchangedTiles(const message &m)
{
	const QString &nick = m.data.value("nick");
	int nTiles = m.data.value("number").toInt();
	emit someoneExchangedHisTiles(nick, nTiles);
	
	m_phase = NEXT_TURN_NEEDED;
}

void client::sendMessage(const QString &text)
{
	message m;
	m.type = "message";
	m.data.insert("message", text);
	m_messageHandler.sendMessage(m);
}

void client::toggleReady()
{
	message m;
	m.type = "toggleReady";
	m_messageHandler.sendMessage(m);
}

void client::confirmAccepted()
{
	message m;
	m.type = "acceptWords";
	m_messageHandler.sendMessage(m);
}

void client::confirmRejected()
{
	message m;
	m.type = "rejectWords";
	m_messageHandler.sendMessage(m);
}

void client::slotSocketError(QAbstractSocket::SocketError error)
{
	switch (error)
	{
		case QAbstractSocket::ConnectionRefusedError:
			m_phase = NOT_CONNECTED;
			m_clientSocket->close();
			m_clientSocket->deleteLater();
			m_clientSocket = 0;
			emit serverDown();
		break;
			
		default:
			// TODO tell the user
			kDebug() << "QAbstractSocket error: " << error << endl;
		break;
	}
}

void client::wrongResponse()
{
	// TODO Tell the user to file a bug
	emit error(i18n("The server gave a wrong response."));
}

#include "client.moc"
