/***************************************************************************
 *   Copyright (C) 2005-2007 by Albert Astals Cid <aacid@kde.org>          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef CLIENT_H 
#define CLIENT_H 

#include <QAbstractTableModel>
#include <QList>
#include <QStringList>
#include <QAbstractSocket>

#include "tileposition.h"
#include "messagehandler.h"

class gameOnServer;
class tile;
class QTcpSocket;

/**
 * This is the representation of a player on the client side 
 */ 

class client : public QAbstractTableModel
{
Q_OBJECT
	public:
		enum Phase
		{
			NOT_CONNECTED,
			WAITING_FOR_VERSION,
			WAITING_TO_SEND_NICK,
			WAITING_TO_BE_ACCEPTED,
			WAITING_TO_ENTER_GAME,
			WAITING_TO_START_GAME,
			NEXT_TURN_NEEDED,
			PLAYING_GAME,
			POST_GOT_TILES,
			POST_PLAYER_CORRECT_WORDS,
			POST_NOTPLAYER_CORRECT_WORDS,
			POST_PASSED_TURN,
			EXCHANGING_TILES,
			POST_EXCHANGING_TILES,
			TRADING_PHASE,
			POST_TRADING_PHASE
		};

		// TODO bool joinAsSpectator -> enum
		client(const QString &nick, bool joinAsSpectator);
		~client();

		QString nick() const;
		void tryConnect(const QString &hostname, int port);
		
		void joinGame(int gameIndex, bool joinAsSpectator);

		Phase currentPhase() const;

		void sendTiles(const tilePositionList &tiles);
		void passTurn();
		void exchangeTiles(const QVector<tile*> &tiles);
		
		void createServerGame(const QString &name, int maxPlayers, const QString &comment, const QString &languageCode, bool allowSpectators);

		/**
		 * @returns true if we are a player, false if we are a spectator
		 */
		bool isPlayer() const;

		// stuff from QAbstractTableModel
		virtual int columnCount(const QModelIndex &parent = QModelIndex()) const;
		virtual QVariant headerData(int section, Qt::Orientation orientation, int role = Qt::DisplayRole) const;
		virtual QVariant data(const QModelIndex &index, int role = Qt::DisplayRole) const;
		virtual int rowCount( const QModelIndex &parent = QModelIndex()) const;
		
		// stuff to query the games in a more independent way
		QString gameLanguage(int gameIndex) const;
		bool gameAllowsSpectators(int gameIndex) const;
	
	signals:
		void showErrorBox(const QString &text, const QString &caption);
		void error(const QString &text);
		void enteredAMultiServer(const QStringList &playersNames);
		void acceptedIntoGame(const QString &languageCode);
		void gotMessage(const QString &nick, const QString &text);
		void newPlayer(const QString &nick, int which, bool isJoiningGame);
		void startGame();
		void newTile(tile *t);
		void nextPlayer(const QString &nick, bool isItMe);
		void gotTiles(const tilePositionList &tiles);
		void gameEnded(const QString &nickFinisher, const QString &nickWinner);
		void removePlayer(const QString &nick, const QString &game);
		void serverDown();
		void unknownWords();
		void changeStatus(const QString &nick, int which);
		void scoreForCurrentPlayer(int score);
		void someoneExchangedHisTiles(const QString &nick, int nTiles);
		void someonePassedHisTurn(const QString &nick);

		void confirmAccept();
		void confirmReject(int reason, const QString &);
		
		void clearPlayers();

	public slots:
		void sendMessage(const QString &message);
		void toggleReady();
		void confirmAccepted();
		void confirmRejected();
	
	private slots:
		void slotSocketError(QAbstractSocket::SocketError error);
		
		void processNewMessage(const message &m);

	private:
		void handleAcceptedIntoGame(const message &m);
		void handleNextTurn(const message &m);
		void handleMessage(const message &m);
		void handleGameEnded(const message &m);
		void handleRemovePlayer(const message &m);
		void handleNewTiles(const message &m);
		void handleStatus(const message &m);
		void handleNewPlayer(const message &m, bool isJoiningGame);
		void handleIncorrectWords(const message &m);
		void handleExchangedTiles(const message &m);
		void wrongResponse();
		
		QTcpSocket *m_clientSocket;
		messageHandler m_messageHandler;
		QString m_nick;
		bool m_spectator;

		Phase m_phase;

		// the data we get from the server about what we have on the server
		QList<gameOnServer*> m_gamesOnServer;
};

#endif
