/****************************************************************************
 *   Copyright (C) 2005 by Albert Astals Cid <tsdgeos@terra.es>             *
 * Copyright (C) 2007-2008 by Michal Sciubidlo <michal.sciubidlo@gmail.com> *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License, or      *
 *   (at your option) any later version.                                    *
 ****************************************************************************/
#include "kombination.h"
#include "kombinationscene.h"

#include "boardwidget.h"
#include "chatwidget.h"
#include "client.h"
#include "dictionary.h"
#include "error.h"
#include "gamelistwidget.h"
#include "ignoredialog.h"
#include "logdialog.h"
#include "multiwidget.h"
#include "ui_newgame.h"
#include "passturndialog.h"
#include "protocol.h"
#include "server.h"
#include "settings.h"
#include "tile.h"
#include "tilefactory.h"
#include "tilesscoresdialog.h"
#include "tilewidget.h"
#include "ui_creategame.h"

#include <kaction.h>
#include <kactioncollection.h>
#include <ktoggleaction.h>
#include <kstandardaction.h>
#include <kstandardgameaction.h>
#include <kdebug.h>

#include <kglobal.h>
#include <kinputdialog.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kmenubar.h>

#include <QGraphicsView>

Kombination::Kombination()
 : KXmlGuiWindow(), m_client(0), m_server(0), m_gameListWidget(0), m_ready(false)
{
	KStandardGameAction::gameNew(this, SLOT(newGame()), actionCollection());

	scene = new KombinationScene(this);
	scene->setSceneRect(0.0, 0.0, 32*15, 32*16);
	m_board = scene->getBoard();
	m_tileWidget = scene->getTileWidget();
	m_exchangeWidget = scene->getExchangeWidget();

	view = new KombinationView(scene, this);

	QWidget *w = new QWidget(this);

	QHBoxLayout *layout = new QHBoxLayout(w);
	layout->addWidget(view);

	QVBoxLayout * vlayout = new QVBoxLayout;
	layout->addLayout(vlayout);

	m_readyButton = new QPushButton(this);
	m_readyButton->setEnabled(false);
	m_readyButton->setText(i18n("Set me as ready"));

	m_passButton = new QPushButton(this);
	m_passButton->setEnabled(false);
	m_passButton->setText("Pass turn");

	m_multiWidget = new multiWidget(this);
	m_multiWidget->setEnabled(false);

	m_chatWidget = new chatWidget(this);
	m_chatWidget->setEnabled(false);

	QHBoxLayout *buttonLayout = new QHBoxLayout();
	buttonLayout->addWidget(m_readyButton);
	buttonLayout->addWidget(m_passButton);

	vlayout->addWidget(m_multiWidget);
	vlayout->addWidget(m_chatWidget);
	vlayout->addLayout(buttonLayout);

	setCentralWidget(w);

	connect(m_board, SIGNAL(moveTileAway(tile *)), m_tileWidget, SLOT(addTile(tile *)));
	connect(m_readyButton, SIGNAL(clicked()), this, SLOT(buttonPressed()));
	connect(m_passButton, SIGNAL(clicked()), this, SLOT(passTurn()));

	QAction *action = KStandardGameAction::gameNew(this, SLOT(newGame()), actionCollection());
	actionCollection()->addAction("newGame", action);

	action = KStandardGameAction::quit(this, SLOT(close()), actionCollection());
	actionCollection()->addAction("quit", action);

	m_ignorePlayersAction = actionCollection()->addAction("ignorePlayers");
	m_ignorePlayersAction->setText(i18n("&Ignore Players"));
	m_ignorePlayersAction->setEnabled(false);
	connect(m_ignorePlayersAction, SIGNAL(triggered()), this, SLOT(ignorePlayers()));

	m_tilesScoresAction = actionCollection()->addAction("showTilesScores");
	m_tilesScoresAction->setText(i18n("&Show Tiles' Scores"));
	m_tilesScoresAction->setEnabled(false);
	connect(m_tilesScoresAction, SIGNAL(triggered()), this, SLOT(showTilesScores()));
	
	m_viewLogAction = actionCollection()->addAction("log");
	m_viewLogAction->setText(i18n("Play &Log"));
	m_viewLogAction->setEnabled(false);
	connect(m_viewLogAction, SIGNAL(triggered()), this, SLOT(viewLog()));
	
	//setCentralWidget(w);
	setupGUI(Keys | Save | Create);
	show();
}


Kombination::~Kombination()
{
}

void Kombination::newGame()
{
	QStringList codesList = dictionary::languages(); //= m_broker->languages();	

	//Sonnet::Speller speller("");
	//QStringList codesList = speller.availableLanguages();

	QStringList namesList;
	KLocale *kloc = KGlobal::locale();
	QMap<QString, QString> languageCodeMap;

	QDialog *d = new QDialog(this);
	Ui_newGame ng;
	ng.setupUi(d);
	ng.clientNick->setText(Settings::clientNick());
	ng.clientPort->setValue(Settings::clientPort());
	ng.server->setText(Settings::clientServer());
	ng.serverNick->setText(Settings::serverNick());
	ng.numberOfPlayers->setValue(Settings::numberOfPlayers());
	ng.serverPort->setValue(Settings::serverPort());
	if (Settings::addAi())
		ng.addAi->setCheckState(Qt::Checked);
	else
		ng.addAi->setCheckState(Qt::Unchecked);

	if (Settings::allowSpectators())
		ng.allowSpectators->setCheckState(Qt::Checked);
	else
		ng.allowSpectators->setCheckState(Qt::Unchecked);

	foreach (const QString &code, codesList)
	{
		QString lang = kloc->languageCodeToName(code);
		kDebug() << code << endl;
		kDebug() << lang << endl;
		kDebug() << tileFactory::haveTilesFor(code) << endl;
		if (!lang.isEmpty() && tileFactory::haveTilesFor(code))
		{
			namesList.append(lang);
			languageCodeMap.insert(lang, code);
		}
	}
	namesList.sort();
	ng.languagesCombo->addItems(namesList);
	ng.languagesCombo->setCurrentIndex(ng.languagesCombo->findText(Settings::dictionaryLanguage()));
	if (d->exec() == QDialog::Accepted)
	{
		if (m_client || m_server)
		{
			if (KMessageBox::questionYesNo(this, i18n("There is a game already in progress, are you sure you want to start another one?"), i18n("Game already in progress")) == KMessageBox::No) return;
		}
		closeConnection();
		disableServer();
		m_multiWidget->clearPlayerList();
		if (ng.tabWidget->currentIndex() == 1)
		{
			// we are the server
			QString nick = ng.serverNick->text().trimmed();
			if (nick.isEmpty())
			{
				KMessageBox::information(this, i18n("The nick can not be empty."), i18n("Empty Nick"));
			}
			else if (ng.languagesCombo->currentText().isEmpty())
			{
				KMessageBox::information(this, i18n("You have to select a language for the game"), i18n("No language chosen"));
			}
			else
			{
				int aiCount = 0;
				if (ng.addAi->isChecked()) aiCount = 1;

				m_server = new server(0, server::SINGLESERVER);
				m_server->createGame(NULL, "game", ng.numberOfPlayers->value(), QString(), languageCodeMap[ng.languagesCombo->currentText()], ng.allowSpectators->isChecked(), aiCount);
				m_nick = nick;
				m_port = ng.serverPort->value();
				Settings::setServerNick(ng.serverNick->text());
				Settings::setNumberOfPlayers(ng.numberOfPlayers->value());
				Settings::setServerPort(ng.serverPort->value());
				Settings::setDictionaryLanguage(ng.languagesCombo->currentText());
				Settings::setAddAi(ng.addAi->checkState() == Qt::Checked);
				Settings::setAllowSpectators(ng.allowSpectators->checkState() == Qt::Checked);
				connect(m_server, SIGNAL(connected()), this, SLOT(serverStarted()));
				connect(m_server, SIGNAL(error(const QString &)), this, SLOT(serverError(const QString &)));
				m_server->start(ng.serverPort->value());
			}
		}
		else
		{
			QString nick = ng.clientNick->text().trimmed();
                        if (nick.isEmpty())
			{
				KMessageBox::information(this, i18n("The nick can not be empty."), i18n("Empty Nick"));
			}
			else
			{
				createClient(ng.server->text(), ng.clientPort->value(), nick, ng.joinAsSpectator->isChecked());
				Settings::setClientNick(ng.clientNick->text());
				Settings::setClientPort(ng.clientPort->value());
				Settings::setClientServer(ng.server->text());
			}
		}

		m_board->reset();
		m_tileWidget->reset();
		m_exchangeWidget->reset();

		m_readyButton->setText(i18n("Set me as ready"));
		view->invalidateScene();
	}
	Settings::self()->writeConfig();
	delete d;
}

void Kombination::createClient(const QString &host, int port, const QString &nick, bool joinAsSpectator)
{
	m_client = new client(nick, joinAsSpectator);
	connect(m_client, SIGNAL(enteredAMultiServer(const QStringList &)), this, SLOT(enteredAMultiServer(const QStringList &)));
	connect(m_client, SIGNAL(acceptedIntoGame(const QString &)), this, SLOT(acceptedIntoGame(const QString &)));
	connect(m_client, SIGNAL(acceptedIntoGame(const QString &)), m_chatWidget, SLOT(connected(const QString &)));
	connect(m_client, SIGNAL(error(const QString &)), this, SLOT(clientError(const QString &)));
	connect(m_client, SIGNAL(showErrorBox(const QString &, const QString &)), this, SLOT(showErrorBox(const QString &, const QString &)));
	connect(m_client, SIGNAL(newPlayer(const QString &, int, bool)), m_chatWidget, SLOT(addPlayer(const QString &, int, bool)));
	connect(m_client, SIGNAL(newPlayer(const QString &, int, bool)), m_multiWidget, SLOT(addPlayer(const QString &, int)));
	connect(m_client, SIGNAL(gotMessage(const QString &, const QString &)), m_chatWidget, SLOT(addMessage(const QString &, const QString &)));
	connect(m_client, SIGNAL(changeStatus(const QString &, int)), m_multiWidget, SLOT(changeStatus(const QString &, int)));
	connect(m_client, SIGNAL(startGame()), m_chatWidget, SLOT(startGame()));
	connect(m_client, SIGNAL(startGame()), m_multiWidget, SLOT(startGame()));
	connect(m_client, SIGNAL(startGame()), this, SLOT(startGame()));
	connect(m_client, SIGNAL(newTile(tile *)), m_tileWidget, SLOT(addTile(tile *)));
	connect(m_client, SIGNAL(nextPlayer(const QString &, bool)), m_chatWidget, SLOT(nextPlayer(const QString &, bool)));
	connect(m_client, SIGNAL(nextPlayer(const QString &, bool)), m_multiWidget, SLOT(nextPlayer(const QString &, bool)));
	connect(m_client, SIGNAL(nextPlayer(const QString &, bool)), m_board, SLOT(nextPlayer(const QString &, bool)));
	connect(m_client, SIGNAL(nextPlayer(const QString &, bool)), this, SLOT(nextPlayer(const QString &, bool)));
	connect(m_client, SIGNAL(gotTiles(const tilePositionList &)), m_board, SLOT(slotSetPlayerTiles(const tilePositionList &)));
	connect(m_client, SIGNAL(confirmAccept()), m_board, SLOT(slotLockTilesAdded()));
	connect(m_client, SIGNAL(confirmReject(int, const QString &)), this, SLOT(confirmReject(int, const QString &)));
	connect(m_client, SIGNAL(confirmReject(int, const QString &)), m_chatWidget, SLOT(tilesRejected(int, const QString &)));
	connect(m_client, SIGNAL(gameEnded(const QString &, const QString &)), m_chatWidget, SLOT(gameEnded(const QString &, const QString &)));
	connect(m_client, SIGNAL(gameEnded(const QString &, const QString &)), this, SLOT(gameEnded()));
	connect(m_client, SIGNAL(removePlayer(const QString &, const QString &)), m_chatWidget, SLOT(removePlayer(const QString &, const QString &)));
	connect(m_client, SIGNAL(removePlayer(const QString &, const QString &)), m_multiWidget, SLOT(removePlayer(const QString &)));
	connect(m_client, SIGNAL(serverDown()), m_chatWidget, SLOT(serverDown()));
	connect(m_client, SIGNAL(serverDown()), m_multiWidget, SLOT(serverDown()));
	connect(m_client, SIGNAL(serverDown()), this, SLOT(serverDown()));
	connect(m_client, SIGNAL(unknownWords()), m_multiWidget, SLOT(showAndStartTimeout()));
	connect(m_client, SIGNAL(scoreForCurrentPlayer(int)), this, SLOT(addScoreCurrentPlayer(int)));
	connect(m_client, SIGNAL(someonePassedHisTurn(const QString &)), m_chatWidget, SLOT(someonePassedHisTurn(const QString &)));
	connect(m_client, SIGNAL(someoneExchangedHisTiles(const QString &, int)), m_chatWidget, SLOT(someoneExchangedHisTiles(const QString &, int)));
	connect(m_client, SIGNAL(clearPlayers()), m_multiWidget, SLOT(clearPlayerList()));

	connect(m_multiWidget, SIGNAL(confirmAccepted()), m_client, SLOT(confirmAccepted()));
	connect(m_multiWidget, SIGNAL(confirmRejected()), m_client, SLOT(confirmRejected()));

	m_client->tryConnect(host, port);
}

void Kombination::closeConnection()
{
	if (!m_client) return;

	disconnect(m_multiWidget, SIGNAL(confirmAccepted()), m_client, SLOT(confirmAccepted()));
	disconnect(m_multiWidget, SIGNAL(confirmRejected()), m_client, SLOT(confirmRejected()));

	m_ignorePlayersAction->setEnabled(false);
	m_tilesScoresAction->setEnabled(false);
	m_viewLogAction->setEnabled(false);

	m_client->deleteLater();
	m_client = 0;
}

void Kombination::disableServer()
{
	if (!m_server) return;

	m_server->deleteLater();
	m_server = 0;
}

void Kombination::serverStarted()
{
	// besides beign the server we are also going to play :-D
	createClient("localhost", m_port, m_nick, false);
	//m_client->joinGame(1, false);
}

void Kombination::acceptedIntoGame(const QString &languageCode)
{
	delete m_gameListWidget;
	m_gameListWidget = 0;
	
	if (!tileFactory::haveTilesFor(languageCode))
	{
		// first, close the connection(s), so the messagebox won't block it/them
		closeConnection();
		disableServer();
		// TODO improve this message :-D
		KMessageBox::error(this, i18n("You cannot join this game as you do not have the tile set for that language."), i18n("No Valid Tile Set"));
		return;
	}
	
	tileFactory tf;
	tf.readTileSet(languageCode);
	m_board->setTileStringList(tf.tileStringList());
	dictionary *dict = new dictionary(languageCode);
	m_board->setDictionary(dict);
	m_chatWidget->setEnabled(true);
	m_multiWidget->setEnabled(true);
	m_readyButton->setEnabled(m_client->isPlayer());
	m_chatWidget->setIAm(m_client->nick());
	m_multiWidget->setIAm(m_client->nick());
	m_chatWidget->addPlayer(m_client->nick(), m_client->isPlayer() ? STATUS_PLAYER_NOT_READY : STATUS_SPECTATOR, true);
	m_multiWidget->addPlayer(m_client->nick(), m_client->isPlayer() ? STATUS_PLAYER_NOT_READY : STATUS_SPECTATOR);
	connect(m_chatWidget, SIGNAL(typedMessage(const QString &)), m_client, SLOT(sendMessage(const QString &)));
	//KMessageBox::information(this, i18n("You are in."), i18n("Connected"));
	m_ignorePlayersAction->setEnabled(true);
	m_tilesScoresAction->setEnabled(true);
	m_viewLogAction->setEnabled(true);
}

void Kombination::gameEnded()
{
	m_readyButton->setText(i18n("Set me as ready"));
	m_readyButton->setEnabled(true);
	m_passButton->setEnabled(false);
}

void Kombination::serverDown()
{
	m_readyButton->setEnabled(false);
	m_readyButton->setText(i18n("Set me as ready"));
	closeConnection();
	KMessageBox::error(this, i18n("The connection to the server has been lost. Cannot continue the game."), i18n("Server Error"));
}

void Kombination::ignorePlayers()
{
	ignoreDialog id(this);
	id.setNicks(m_multiWidget->getPlayersNicks(multiWidget::DoNotGiveMyOwnNick), m_chatWidget->ignoredNicks());
	if (id.exec() == QDialog::Accepted)
	{
		m_chatWidget->setIgnoredNicks(id.ignoredNicks());
	}
}

void Kombination::showTilesScores()
{
	tilesScoresDialog tsd(this, m_board->getTileStringList());
	tsd.exec();
}

void Kombination::passTurn()
{
	//if there are player tiles on board put them to rack
	QList<tile*> t = m_board->tilesAdded().tiles();
	QList<tile*>::iterator i;
	for (i = t.begin(); i != t.end(); ++i)
		m_tileWidget->addTile(*i);
	m_board->slotDeleteLastTiles();

	if (m_exchangeWidget->freeSpaces() == 7)
		m_client->passTurn();
	else
	{
		m_board->setAcceptDrops(false);
		m_client->exchangeTiles(m_exchangeWidget->tiles());
		m_exchangeWidget->removeTiles(m_exchangeWidget->tiles());
		m_board->setAcceptDrops(true);
	}
}

void Kombination::viewLog()
{
	logDialog ld(this, m_scoresLog);
	ld.exec();
}

void Kombination::addScoreCurrentPlayer(int score)
{
	int score2, error;
	wordsAndTilePositions wtp = m_board->checkValidTilesPositions(score2, error);
	// we could check score2 == score but atm that gives different scores
	// because extra points are not correctly checked on the client side
	if (!error)
	{
		QPair<QString, tilePositionList> p;
		QStringList wordList;
		for(int i=0; i<wtp.size(); i++)
			wordList << wtp[i].first;
		m_scoresLog.addScore(m_multiWidget->currentPlayerNick(), wordList, score);
		m_chatWidget->addScoreCurrentPlayer(wordList, score);
		m_multiWidget->addScoreCurrentPlayer(wordList, score);
	}
	else
	{
		kDebug() << "GOT ERROR ON addScoreCurrentPlayer" << endl;
		// TODO that means the server is cheating
	}
}

void Kombination::enteredAMultiServer(const QStringList &playerList)
{
	delete m_gameListWidget;
	//delete m_board->layout();
	//QHBoxLayout *boardLayout = new QHBoxLayout(m_board);
	m_gameListWidget = new gameListWidget((QWidget*)m_board, m_client);
	m_gameListWidget->setAutoFillBackground(true);
	//TODO: boardLayout->addWidget(m_gameListWidget);
	connect(m_gameListWidget, SIGNAL(createServerGame()), this, SLOT(createServerGame()));
	connect(m_gameListWidget, SIGNAL(joinServerGame()), this, SLOT(joinServerGame()));
	
	m_chatWidget->setEnabled(true);
	m_chatWidget->setIAm(m_client->nick());
	m_chatWidget->addPlayer(m_client->nick(), STATUS_SPECTATOR, false);
	m_multiWidget->addPlayer(m_client->nick(), STATUS_SPECTATOR);
	foreach(const QString &nick, playerList)
	{
		m_chatWidget->addPlayer(nick, STATUS_SPECTATOR, false);
		m_multiWidget->addPlayer(nick, STATUS_SPECTATOR);
	}
	connect(m_chatWidget, SIGNAL(typedMessage(const QString &)), m_client, SLOT(sendMessage(const QString &)));
}

void Kombination::createServerGame()
{
	QStringList codesList = dictionary::languages(); //= m_broker->languages();

	//Sonnet::Speller speller("");
	//QStringList codesList = speller.availableLanguages();

	QStringList namesList;
	KLocale *kloc = KGlobal::locale();
	QMap<QString, QString> languageCodeMap;
	
	QDialog *d = new QDialog(this);
	Ui_createGame cg;
	cg.setupUi(d);
	foreach (const QString &code, codesList)
	{
		QString lang = kloc->languageCodeToName(code);
		if (!lang.isEmpty() && tileFactory::haveTilesFor(code))
		{
			namesList.append(lang);
			languageCodeMap.insert(lang, code);
		}
	}
	namesList.sort();
	cg.languagesCombo->addItems(namesList);
	if (d->exec() == QDialog::Accepted)
	{
		if (cg.gameName->text().isEmpty())
		{
			KMessageBox::error(this, i18n("You cannot create games without name."), i18n("Name cannot be empty"));
		}
		else
		{
			m_client->createServerGame(cg.gameName->text(), cg.numberOfPlayers->value(), cg.gameComment->text(), languageCodeMap[cg.languagesCombo->currentText()], cg.allowSpectators->isChecked());
		}
	}
}

void Kombination::joinServerGame()
{
	QStringList codesList = dictionary::languages(); //m_broker->languages();

	//Sonnet::Speller speller("");
	//QStringList codesList = speller.availableLanguages();

	int selectedRow = m_gameListWidget->selectedGameRow();
	QString gameLanguage = m_client->gameLanguage(selectedRow);
	if (codesList.contains(gameLanguage))
	{
		if (m_client->gameAllowsSpectators(selectedRow))
		{
			QStringList options;
			bool ok;
			options << i18n("Join as Player");
			options << i18n("Join as Spectator");
			QString result = KInputDialog::getItem(i18n("Join as..."), i18n("Play as..."), options, 0, false, &ok);
			if (ok)
			{
				m_client->joinGame(selectedRow, result == i18n("Join as Spectator"));
			}
		}
		else m_client->joinGame(selectedRow, false);
	}
	else KMessageBox::error(this, i18n("You can not join this game as you do not have the dictionary of the game language"), i18n("Language missing"));
}

void Kombination::clientError(const QString &text)
{
	closeConnection();
	KMessageBox::error(this, text, i18n("Client Error"));
}

void Kombination::serverError(const QString &text)
{
	disableServer();
	KMessageBox::error(this, text, i18n("Server Error"));
}

void Kombination::showErrorBox(const QString &text, const QString &caption)
{
	KMessageBox::error(this, text, caption);
}

void Kombination::startGame()
{
	m_board->reset();
	m_tileWidget->reset();
	m_exchangeWidget->reset();

	m_readyButton->setText(i18n("Send word"));
	m_scoresLog.start(m_multiWidget->getPlayersNicks(multiWidget::GiveMyOwnNick));
}

void Kombination::buttonPressed()
{
	if (m_client->currentPhase() == client::WAITING_TO_START_GAME)
	{
		m_ready = !m_ready;
		if (m_ready)
		{
			m_readyButton->setText(i18n("Set me as not ready"));
		}
		else
		{
			m_readyButton->setText(i18n("Set me as ready"));
		}
		m_multiWidget->setMyselfReady(m_ready);
		m_client->toggleReady();
	}
	else if (m_client->currentPhase() == client::PLAYING_GAME)
	{
		bool newtiles = m_board->wereTilesAdded();
		if (!newtiles)
		{
			if (KMessageBox::warningContinueCancel(this, i18n("You have placed no tiles during this turn. Do you want to pass the turn or come back placing tiles?"), i18n("No Tiles Added"), KGuiItem(i18n("Pass")), KStandardGuiItem::cancel(), "warn-about-no-tiles-added") == KMessageBox::Continue)
				m_client->passTurn();
		}
		else
		{
			int error = 0;
			int score = 0;
			
			m_board->setPlayedTiles(m_board->tilesAdded());
			wordsAndTilePositions words = m_board->checkValidTilesPositions(score, error);
			QStringList w;
			QPair<QString, tilePositionList> i;
			for(int j=0; j<words.size(); j++)
				w << words[j].first;
			kDebug() << endl
			          << "words = " << w << endl;
			if (error > 0)
			{
				// check the possible error, and warn the user
				QString errmsg;
				switch (error)
				{
					case ERROR_FIRST_WORD_NOT_CENTERED:
						errmsg = i18n("The first word of a game should cover the central position.");
						break;
					case ERROR_BLANK_TILE_NOT_REPLACED:
						errmsg = i18n("You have added a blank tile but you did not give it a proper value.");
						break;
					case ERROR_TILES_NOT_ALIGNED:
						errmsg = i18n("The tiles you added are not aligned (horizontally or vertically), so they cannot create any valid word.");
						break;
					default:
						// generic message
						errmsg = i18n("Generic error #%1.",error);
				
				}
				KMessageBox::information(this, errmsg, i18n("Error"));
			}
			else
			{
				// The inserted tiles are ok;
				m_client->sendTiles(m_board->tilesAdded());
			}
		}
	}
}

void Kombination::nextPlayer(const QString &, bool itsMe)
{
	m_readyButton->setEnabled(itsMe);
	m_passButton->setEnabled(itsMe);
}

void Kombination::confirmReject(int z, const QString & a)
{
	QList<tile*> t = m_board->tilesAdded().tiles();
	QList<tile*>::iterator i;
	for (i = t.begin(); i != t.end(); ++i)
		m_tileWidget->addTile(*i);
	m_board->slotDeleteLastTiles();
}

#include "kombination.moc"
