/***************************************************************************
 *   Copyright (C) 2005-2007 by Albert Astals Cid <aacid@kde.org>          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "player.h"

#include <QList>
#include <QVector>

#include <kdebug.h>

#include "messagehandler.h"
#include "protocol.h"
#include "tile.h"

player::player(spectator *s) : spectator(s), m_myturn(false), m_score(0), m_ready(false)
{
	prepareForNewVoting();
}

player::player() : spectator(), m_myturn(false), m_score(0), m_ready(true)
{
	prepareForNewVoting();
}

player::~player()
{
	deleteSocket();
}

bool player::ready() const
{
	return m_ready;
}

void player::startGame(const QList<tile*> &tiles)
{
	m_phase = PLAYING_GAME;
	m_tiles = tiles;
	
	message m;
	m.type = "startingGame";
	m.data = toMap(tiles);
	m_messageHandler->sendMessage(m);
}

void player::sendNewTiles(const QList<tile*> &tiles)
{
	m_tiles << tiles;
	
	message m;
	m.type = "newTiles";
	m.data = toMap(tiles);
	m_messageHandler->sendMessage(m);
}

void player::nextTurn(const QString &nick)
{
	spectator::nextTurn(nick);
	m_myturn = nick == m_nick;
	qDeleteAll(m_tilesIHaveUsedInThisTurn);
	m_tilesIHaveUsedInThisTurn.clear();
}

void player::sendExchangedTiles(const QList<tile*> &tiles)
{
	m_tiles << tiles;
	
	message m;
	m.type = "newExchangedTiles";
	m.data = toMap(tiles);
	m_messageHandler->sendMessage(m);
}

void player::sendTiles(const tilePositionList &tiles)
{
	kDebug() << endl << endl << endl << "tiles count: " << tiles.count() << endl << endl << endl;
	message m;
	m.type = "otherPlayerTiles";
	m.data = tiles.toMap();
	m_messageHandler->sendMessage(m);
}

void player::correctWord(int score)
{
	m_phase = PLAYING_GAME;
	
	message m;
	m.type = "correctWords";
	m.data.insert("score", QString::number(score));
	m_messageHandler->sendMessage(m);
}

void player::unknownWords()
{
	spectator::unknownWords();
	
	message m;
	m.type = "unknownWords";
	m_messageHandler->sendMessage(m);
}

void player::rejectVotedTiles(const QString &nick)
{
	m_phase = PLAYING_GAME;
	m_tiles << m_tilesIHaveUsedInThisTurn;
	m_tilesIHaveUsedInThisTurn.clear();
	
	message m;
	m.type = "incorrectWords";
	m.data.insert("nick", nick);
	m.data.insert("reason", "lostVoting");
	m_messageHandler->sendMessage(m);
}

void player::usesExclusiveTiles(const QString &nick)
{
	m_phase = PLAYING_GAME;
	m_tiles << m_tilesIHaveUsedInThisTurn;
	m_tilesIHaveUsedInThisTurn.clear();
	
	message m;
	m.type = "incorrectWords";
	m.data.insert("nick", nick);
	m.data.insert("reason", "usesExclusiveTiles");
	m_messageHandler->sendMessage(m);
}

int player::freeTiles() const
{
	return 7 - m_tiles.count();
}

player::VotingState player::votingState() const
{
	return m_votingState;
}

void player::prepareForNewVoting()
{
	m_votingState = VOTING_UNKNOWN;
}

int player::score() const
{
	return m_score;
}

void player::addToScore(int s)
{
	m_score += s;
}

bool player::hasTile(const QString &tileText) const
{
	foreach(tile *t, m_tiles)
		if (t->string() == tileText) return true;

	return false;
}		

void player::processNewMessage(const message &m)
{
	kDebug() << "player::processNewMessage" << " " << m.type << " " << endl;
	switch (m_phase)
	{
		case WAITING_FOR_SPECTATOR:
			// Never happens
		break;
		
		case WAITING_FOR_NICK:
			if (m.type == "sendNick")
			{
				m_nick = m.data.value("nick");
				m_phase = WAITING_TO_START_GAME;
				emit gotNick(this);
			}
			// else wrongResponse()
		break;

		case WAITING_TO_START_GAME:
			if (m.type == "message") handleMessage(m);
			else if (m.type == "toggleReady")
			{
				m_ready = !m_ready;
				emit readyChanged(this);
			}
			// else wrongResponse()
		break;

		case PLAYING_GAME:
			if (m.type == "message") handleMessage(m);
			else if (m.type == "passTurn") emit passTurn(this);
			else if (m.type == "exchangeTiles")
			{
				QList<tile*> tiles;
				fromMap(m.data, tiles);
				bool validTiles = checkValidTiles(tiles);
				if (validTiles)
					emit exchangeTiles(this, tiles);
				// else cheater kill him
			}
			else if (m.type == "sendingTiles")
			{
				tilePositionList tiles;
				tiles.fromMap(m.data);
				if (m_myturn)
				{
					bool validTiles = checkValidTiles(tiles.tiles());
					if (validTiles)
						emit gotTiles(this, tiles);
					// else cheater kill him
				}
				// else cheater kill him
			}
			// else wrongResponse()
		break;

		case TRADING_PHASE:
		{
			bool votechanged = false;
			if (m.type == "acceptWords")
			{
				votechanged = true;
				m_votingState = VOTING_ACCEPTED;
			}
			else if (m.type == "rejectWords")
			{
				votechanged = true;
				m_votingState = VOTING_REJECTED;
			}
			// else wrongResponse()
			if (votechanged)
				emit votingChanged();
		}
		break;
		
		case NICK_ALREADY_IN_USE:
			// wrongResponse()
		break;
	}
}

bool player::checkValidTiles(const QList<tile*> &tiles)
{
	bool validtiles = true;
	int i = 0;
	do
	{
		bool tilefound = false;
		int j = 0;
		do
		{
			if (*(tiles.at(i)) == *(m_tiles.at(j)))
			{
				tilefound = true;
				m_tilesIHaveUsedInThisTurn.append(m_tiles.at(j));
				m_tiles.removeAt(j);
			}
			++j;
		}
		while ((j < m_tiles.count()) && !tilefound);
		if (!tilefound && tiles.at(i)->isBlank())
		{
			j = 0;
			// it may be a blank file check if we have one
			do
			{
				tilefound = m_tiles.at(j)->isBlank();
				++j;
			}
			while ((j < m_tiles.count()) && !tilefound);
			if (tilefound)
			{
				--j;
				m_tilesIHaveUsedInThisTurn.append(m_tiles.at(j));
				m_tiles.removeAt(j);
			}
		}
		validtiles &= tilefound;
		++i;
	}
	while ((i < tiles.count()) && validtiles);

	return validtiles;
}

void player::gameEnded(const QString &finisher, const QString &winner)
{
	m_phase = WAITING_TO_START_GAME;
	m_score = 0;
	m_myturn = false;
	m_ready = false;
	qDeleteAll(m_tilesIHaveUsedInThisTurn);
	m_tilesIHaveUsedInThisTurn.clear();
	qDeleteAll(m_tiles);
	m_tiles.clear();

	message m;
	m.type = "gameEnded";
	m.data.insert("winner", winner);
	m.data.insert("finisher", finisher);
	m_messageHandler->sendMessage(m);
}

#include "player.moc"
