/**************************************************************************
*   Copyright (C) 2005 by Lukas Tinkl                                     *
*   lukas@kde.org                                                         *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.             *
***************************************************************************/

#include "server.h"
#include "gametypes.h"

#include <stdlib.h>
#include <strings.h>

#include <qstring.h>
#include <qsimplerichtext.h>
#include <qregexp.h>
#include <qfontmetrics.h>
#include <qpainter.h>

#include <klocale.h>
#include <kdebug.h>

void Server::init()
{
    m_ping = 999;
    m_players = m_maxPlayers = 0;
    m_lastRefresh = m_lastAnwser = QDateTime::currentDateTime();
    m_manager = 0;
}

void Server::setGameManager( GameManager * manager )
{
    m_manager = manager;
}

QString Server::gameName() const
{
    QStringList rulenames = m_manager->gameNameRule( m_gameID );

    for ( QStringList::Iterator it = rulenames.begin(); it != rulenames.end(); ++it )
    {
        QString value = m_rules[( *it )];
        if ( !value.isNull() )
            return value;
    }

    return QString::null;
}

QString Server::gameType() const
{
    QStringList rulenames = m_manager->gameTypeRule( m_gameID );

    for ( QStringList::Iterator it = rulenames.begin(); it != rulenames.end(); ++it )
    {
        QString value = m_rules[( *it )];
        if ( !value.isNull() )
            return decodeGametype( value );
    }

    return QString::null;
}

int Server::privateSlots() const
{
    if ( m_gameID == "WOETS" )  // Enemy Territory family
        return rule( "sv_privateClients" ).toInt();
    else if ( m_gameID == "HLS" || m_gameID == "A2S" )  // Half Life (2)
        return rule( "reserve_slots" ).toInt();
    else                        // other games don't support this
        return 0;
}

int Server::freeSlots() const
{
    return m_maxPlayers - m_players - privateSlots();
}

// virtual
QString Server::text( int col ) const
{
    if ( col == 0 )
        return strippedName();
    else if ( col == 1 )
        return m_address;
    else if ( col == 2 )
        return QString::number( m_ping );
    else if ( col == 3 )
        // there's no text in priv status
        ;
    else if ( col == 4 )
        return i18n( "players / max players", "%1/%2 (%3)" ).arg( m_players ).arg( m_maxPlayers ).arg( freeSlots() );
    else if ( col == 5 )
        return m_mapName;
    else if ( col == 6 )
        return gameName();
    else if ( col == 7 )
        return gameType();

    return QString::null;
}

void Server::setName( const QString & name )
{
    QString tmp = name;

    // with markup
    m_name = name;

    // without markup
    QRegExp rx( "</?font.*>" );
    rx.setMinimal( true );
    m_strippedName = tmp.remove( rx );
}

// virtual
int Server::compare( QListViewItem * i, int col, bool ascending ) const
{
    if ( col == 2 || col == 4 )             // integers
    {
        int us = 0;
        int them = 0;

        if ( col == 2 )         // ping
        {
            us = m_ping;
            them = static_cast<Server *>( i )->ping();
        }
        else if ( col == 4 )    // number of players
        {
            us = m_players;
            them = static_cast<Server *>( i )->numPlayers();
        }

        if ( us < them )
            return -1;
        else if ( us > them )
            return 1;
        else
            return 0;
    }
    else
        return QListViewItem::compare( i, col, ascending );
}

// virtual
void Server::paintCell( QPainter * p, const QColorGroup & cg, int column, int width, int align )
{
    if ( column == 0 )          // player name
    {
        QRect r( 0, 0, width, height() );
        p->fillRect( r, Qt::gray );
        p->drawPixmap( 2, 0, *pixmap( 0 ) );
        //kdDebug() << "Drawing server: " << m_name << endl;
        QSimpleRichText rich( m_name, listView()->font() );
        rich.draw( p, 20, 0, r, cg ); // 18 + 2 == width of the game icon + some space
    }
    else
        QListViewItem::paintCell ( p, cg, column, width, align );
}

// virtual
int Server::width( const QFontMetrics & fm, const QListView * lv, int c ) const
{
    if ( c == 0 )               // server name
    {
        //kdDebug() << "Width: " << fm.width( strippedName() ) << endl;
        return 20 + fm.width( strippedName() );
    }
    else
        return QListViewItem::width( fm, lv, c );
}

#ifdef HAVE_LIBGEOIP
QString Server::countryCode( GeoIP * gi ) const
{
    // strip the port from the IP, GeoIP doesn't like that
    return QString::fromLatin1( GeoIP_country_code_by_addr( gi, m_address.section( ":", 0, 0 ) ) ).lower();
}
#endif

void Server::setRules( const QDomElement & elem )
{
    QDomNodeList childList = elem.childNodes();
    for ( uint i = 0; i < childList.count(); i++ )
    {
        QDomElement celem = childList.item( i ).toElement();
        if ( !celem.hasAttribute( "name" ) )
            continue;
        m_rules.insert( celem.attribute( "name" ), celem.text() );
    }
}

QString Server::rule( const QString & name ) const
{
    return m_rules[name];
}

bool Server::isLocked() const
{
    QString rulename = m_manager->passwordRule( m_gameID );
    QString rulevalue = m_rules[rulename];
    if ( rulevalue == QString::null ) // rule doesn't exist
        return false;
    else if ( rulevalue == "1" || rulevalue == "true" || rulevalue == "yes" )
        return true;

    return false;
}

bool Server::hasPunkbuster() const
{
    QString rulename = m_manager->pbRule( m_gameID );
    QString rulevalue = m_rules[rulename];
    if ( rulevalue == QString::null ) // rule doesn't exist
        return false;
    else if ( rulevalue == "1" || rulevalue == "true" || rulevalue == "yes" )
        return true;

    return false;
}

QString Server::decodeGametype( const QString & gameType ) const
{
    if ( gameName().isNull() )
        return QString::null;

    //kdDebug() << QString( "decoding gametype, game ID: %1, orig. gametype: %2, game mod: %3" ).
    //    arg( m_gameID ).arg( gameType ).arg( gameName() ) << endl;

    char *endptr;
    int n = strtol( gameType, &endptr, 10 );

    // strtol returns a pointer to the first invalid digit, if both pointers
    // are equal there was no number at all
    if ( gameType == endptr )
        return QString::null;

    struct q3a_gametype_s* ptr = 0;

    if ( m_gameID == "Q3S" )      /* Quake 3 */
        ptr = q3a_gametype_map;
    else if ( m_gameID == "WOS" ) /* Wolfenstein */
        ptr = wolf_gametype_map;
    else if ( m_gameID == "EFS" ) /* Elite Force */
        ptr = ef_gametype_map;
    else if ( m_gameID == "WOETS" ) /* Enemy Territory */
        ptr = wolfet_gametype_map;
    else if ( m_gameID == "JK3S" ) /* Jedi Academy */
        ptr = jk3_gametype_map;

    for ( ; ptr && ptr->mod != NULL; ptr++ )
    {
        if ( !strcasecmp ( gameName(), ptr->mod ) && n >=0 && n < ptr->number && ptr->gametypes[n] )
        {
            return QString::fromLatin1( ptr->gametypes[n] );
        }
    }

    return gameType;            // return unchanged if not found
}

QDataStream & operator<<( QDataStream & s, const Server & server )
{
    s << server.gameID() << server.name() << server.address() << server.comment();
    return s;
}

QDataStream & operator>>( QDataStream & s, Server & server)
{
    QString str;
    s >> str;
    server.setGameID( str );

    s >> str;
    server.setName( str );

    s >> str;
    server.setAddress( str );

    s >> str;
    server.setComment( str );
    return s;
}

QDataStream & operator<<( QDataStream & s, const Favorite & fav )
{
    s << fav.gameID << fav.name << fav.address << fav.comment;
    return s;
}

QDataStream & operator>>( QDataStream & s, Favorite & fav )
{
    s >> fav.gameID >> fav.name >> fav.address >> fav.comment;
    return s;
}

bool Favorite::operator==( const Favorite & fav )
{
    return ( ( gameID == fav.gameID ) && ( name == fav.name )
             && ( address == fav.address ) && ( comment == fav.comment ) );
}



