/****************************************************************
Copyright 2008  Matt Williams matt@milliams.com
Copyright 2008  Parker Coates parker.coates@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) version 3, or any
later version accepted by the membership of KDE e.V. (or its
successor approved by the membership of KDE e.V.), which shall
act as a proxy defined in Section 6 of version 3 of the license.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/

#ifndef KSCOREBOARDBASE_H
#define KSCOREBOARDBASE_H

#include <libkdegames_export.h>
#include <kscoremanager.h>

#include <QtCore/QMap>

/**
 * \class KScoreBoardBase kscoreboardbase.h <KScoreBoardBase>
 *
 * \brief The base class for score UIs
 *
 * All score board implememtations must derive from this class. Call
 * loadScores() to retrieve the scores from KScoreManager and submitName() to
 * submit the name for the most recently added score.
 *
 * \since KDE 4.2
 * \author Matt Williams <matt@milliams.com>
 */
class KDEGAMES_EXPORT KScoreBoardBase
{
    public:
        /**
         * \enum OutcomeFlag
         *
         * \brief The possible outcomes
         *
         * This provides the values for the \a Outcome flag set
         *
         * \see setMessage()
         */
        enum OutcomeFlag
        {
            TopScore, ///<The score is the top in its group
            Highscore, ///<The score made it to the highscore table
            NoHighscore ///<The score didn't make it to the table
        };
        Q_DECLARE_FLAGS(Outcome, OutcomeFlag)

        /**
         * \enum ScoreDisplayFlag
         *
         * \brief Flag for setting what should be highlighted in the dialog
         *
         * This provides the values for the \a ScoreDisplay flag set
         */
        enum DisplayOptionsFlag
        {
            HighlightLastScore = 0x1, ///<Highlight the last score to be entered into the table. This will also make sure that the group which the score belongs to is showing. If no highscore was achieved, the group it would have been in will be shown.
            DisplayMessage = 0x2, ///<Should the comment message relevant to the last added score be displayed
            LeaderBoard = 0x4///<Display the scores as a leader board. This will only display the top score in each group
        };
        Q_DECLARE_FLAGS(DisplayOptions, DisplayOptionsFlag)

        KScoreBoardBase();
        virtual ~KScoreBoardBase();
        void setDisplayOptions(const DisplayOptions & options);
        DisplayOptions displayOptions() const;

    protected:
        const QMap<QString, QString> groupNames() const;
        const QList<KScoreManager::Field> visibleFields() const;
        const QMap<KScoreManager::Field, QString> fieldHeaders() const;
        const QList<QString> groups() const;
        const KScoreManager::ScoreList scores(const QString& groupKey = QString()) const;
        QPair<QString, int> newestScore() const;
        bool nameHasBeenRequested() const;
        /**
         * \param name the name to be entered for the most recent score
         */
        void submitName(const QString& name);

    private:
        class KScoreBoardBasePrivate;
        KScoreBoardBasePrivate * const d;
};

Q_DECLARE_OPERATORS_FOR_FLAGS(KScoreBoardBase::DisplayOptions)

#endif
