/****************************************************************
Copyright 2008  Parker Coates parker.coates@gmail.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) version 3, or any
later version accepted by the membership of KDE e.V. (or its
successor approved by the membership of KDE e.V.), which shall
act as a proxy defined in Section 6 of version 3 of the license.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************/

#include "kscoretable.h"

#include "kscoremanager.h"

#include <KDebug>
#include <KLocale>
#include <KSeparator>

#include <QtGui/QGridLayout>
#include <QtGui/QKeyEvent>
#include <QtGui/QLabel>
#include <QtGui/QLineEdit>
#include <QtGui/QStackedWidget>

class KScoreTable::KScoreTablePrivate
{
    public:
        KScoreTablePrivate(KScoreTable* const parent);
        void updateDisplay();
        void updateTable();
        void highlightEntry(int rank);
        void showNameEdit(int rank);
        void newScore(const QString& groupKey, int rank);

        KScoreTable * const q;
        QString groupKey;
        QList< QMap<int, QLabel*> > labels;
        QList<QStackedWidget*> stacks;
        QLineEdit * nameEdit;
        QLabel * nameLabel;
};

KScoreTable::KScoreTablePrivate::KScoreTablePrivate(KScoreTable * const parent)
  : q(parent)
{
}

void KScoreTable::KScoreTablePrivate::updateTable()
{
    q->cancelNameRequest();

    int i = 0;
    foreach (const KScoreManager::ScoreEntry & score, q->scores(groupKey))
    {
        foreach (KScoreManager::Field field, q->visibleFields())
        {
            labels[i][field]->setText(score[field].toString());
        }
        ++i;
    }
}

void KScoreTable::KScoreTablePrivate::updateDisplay()
{
    int index = 0;
    if (q->newestScore().first == groupKey && ( q->nameHasBeenRequested() || q->displayOptions().testFlag(HighlightLastScore)) )
    {
        index = q->newestScore().second;
    }

    highlightEntry(index);

    if (q->newestScore().first == groupKey && q->nameHasBeenRequested())
    {
        showNameEdit(index);
    }
}

void KScoreTable::KScoreTablePrivate::highlightEntry(int rank)
{
    QFont normal = q->font();
    QFont bold = normal;
    bold.setBold(true);

    for (int entry = 1; entry <= KScoreManager::self()->scoreBoardLength(); ++entry)
    {
        labels[entry-1][-1]->setFont(entry == rank ? bold : normal);
        foreach (KScoreManager::Field field, q->visibleFields())
        {
            labels[entry-1][field]->setFont(entry == rank ? bold : normal);
        }
    }
}

void KScoreTable::KScoreTablePrivate::showNameEdit(int rank)
{
    if (stacks.size())
    {
        nameLabel = labels[rank-1][KScoreManager::Name];

        delete nameEdit;
        nameEdit = new QLineEdit(nameLabel->text());
        int preferredWidth = qMin(nameEdit->sizeHint().width(), 300);

        QStackedWidget * stack = stacks[rank - 1];
        stack->addWidget(nameEdit);
        stack->setCurrentWidget(nameEdit);
        nameEdit->setFocus();
        
        connect(nameEdit, SIGNAL(returnPressed()), q, SLOT(acceptName()), Qt::QueuedConnection);
    }
}

void KScoreTable::KScoreTablePrivate::newScore(const QString& groupKey, int rank)
{
    Q_UNUSED(rank)
    if (groupKey == groupKey)
    {
        updateTable();
    }
    else
    {
        q->cancelNameRequest();
    }
    updateDisplay();
}




KScoreTable::KScoreTable(const QString& groupKey, QWidget* parent)
  : QWidget(parent),
    d(new KScoreTablePrivate(this))
{
    d->groupKey = groupKey;
    d->nameEdit = 0;
    d->nameLabel = 0;

    QFont bold = font();
    bold.setBold(true);

    QLabel * label; // Reusable label pointer

    QGridLayout * layout = new QGridLayout(this);
    layout->setMargin(20);
    layout->setHorizontalSpacing(8);
    layout->setVerticalSpacing(0);

    const int headerRow = 0;
    const int firstDataRow = headerRow + 2;
    const int minimumRowHeight = QLineEdit(this).sizeHint().height();

    // Create headers
    label = new QLabel(i18n("Rank"));
    label->setFont(bold);
    layout->addWidget(label, headerRow, 0, Qt::AlignHCenter);
    layout->setColumnMinimumWidth(0, 50);

    int column = 1;
    foreach (KScoreManager::Field field, visibleFields())
    {
        label = new QLabel(fieldHeaders()[field]);
        label->setFont(bold);
        layout->addWidget(label, headerRow, column, field == KScoreManager::Name ? Qt::AlignLeft : Qt::AlignRight);
        layout->setColumnMinimumWidth(column, field == KScoreManager::Name ? 80 : 50);
        ++column;
    }

    layout->setRowMinimumHeight(0, minimumRowHeight);
    layout->addWidget(new KSeparator(Qt::Horizontal), 1, 0, 1, layout->columnCount());

    // Create the score entry labels
    for (int row = firstDataRow; row < KScoreManager::self()->scoreBoardLength() + firstDataRow; ++row)
    {
        d->labels << QMap<int, QLabel*>();

        layout->setRowMinimumHeight(row, minimumRowHeight);

        label = new QLabel(QString::number(row - firstDataRow + 1));
        layout->addWidget(label, row, 0, Qt::AlignHCenter);
        d->labels.last().insert(-1, label);

        int column = 1;
        foreach (KScoreManager::Field field, visibleFields())
        {
            label = new QLabel();
            if (field == KScoreManager::Name)
            {
                QStackedWidget * stack = new QStackedWidget();
                stack->addWidget(label);
                stack->setCurrentWidget(label);
                layout->addWidget(stack, row, column, Qt::AlignLeft);
                d->stacks << stack;
            }
            else
            {
                layout->addWidget(label, row, column, Qt::AlignRight);
            }
            d->labels.last().insert(field, label);
            ++column;
        }
    }

    d->updateTable();
    d->updateDisplay();

    connect(KScoreManager::self(), SIGNAL(newScore(QString,int)), this, SLOT(newScore(QString,int)));
}

KScoreTable::~KScoreTable()
{
    delete d;
}

void KScoreTable::setGroup(const QString& groupKey)
{
    d->groupKey = groupKey;
    d->updateTable();
}

QString KScoreTable::group() const
{
    return d->groupKey;
}

void KScoreTable::setDisplayOptions(const KScoreBoardBase::DisplayOptions & options)
{
    KScoreBoardBase::setDisplayOptions(options);
    d->updateDisplay();
}

void KScoreTable::cancelNameRequest()
{
    delete d->nameEdit;
    d->nameEdit = 0;
    d->nameLabel = 0;
}

void KScoreTable::acceptName()
{
    if (d->nameEdit)
    {
        QString name = d->nameEdit->text();

        // Double check that KScoreManager still wants the name.
        if (nameHasBeenRequested() && newestScore().first == d->groupKey )
        {
            d->nameLabel->setText(name);
            submitName(name);
        }

        cancelNameRequest();
    }
}

void KScoreTable::keyPressEvent(QKeyEvent * e)
{
    // Eat all Enter keypresses coming from nameEdit to prevent automatically accepting the containing dialog.
    if (d->nameEdit && (e->key() == Qt::Key_Return || e->key() == Qt::Key_Enter))
    {
        e->accept();
    }
    else if (d->nameEdit && e->key() == Qt::Key_Escape)
    {
//         cancelNameRequest();
         e->accept();
    }
    else
    {
        e->ignore();
        QWidget::keyPressEvent(e);
    }
}

#include "kscoretable.moc"
