/*
    This file is part of KSpiral.
    Copyright 2009 Felix Lemke <hobbyblobby@arcor.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "ball.h"

#include <cmath>
#include <QPointF>
#include <KGameRenderer>
#include <QTimer>
#include <QTimeLine>
#include "svgreader.h"

KSpiral::Ball::Ball(const QPointF& pos, KGameRenderer* renderer, const QString& type, QGraphicsItem* parent)
	: KGameRenderedObjectItem(renderer, type, parent)
	, m_nextBall(0)
	, m_prevBall(0)
	, m_radius(renderer->boundsOnSprite(type).width() / 2)
	, m_type(type)
	, m_states(KSpiral::ballNotInAction)
	, m_timer(0)
	, m_lastFrame(0)
	, m_arrangeStepCounter(0)
{
  const QSizeF size = renderer->boundsOnSprite(type).size();
  setOffset(-size.width() / 2, -size.height() / 2);
  setFixedSize(size.toSize());
	setPos(pos); //move ball to right position
}

KSpiral::Ball::~Ball() {
}

void KSpiral::Ball::addState(KSpiral::BallState state) {
	m_states |= state;
}

#include <kdebug.h>
void KSpiral::Ball::removeState(KSpiral::BallState state) {
	m_states &= ~state;
}

void KSpiral::Ball::shoot(const QPointF& startPos, qreal angle, qreal velocity) {
	//calculate fly parameters
	const int stepTime = 50; //milliseconds
	const qreal d = velocity / 1000 * stepTime; //distance after one stepTime
	m_dx = d * cos(angle);
	m_dy = d * sin(angle);
	setPos(startPos);

	//initialize timer and start
	if (m_timer) delete m_timer;
	m_timer = new QTimer;
	connect(m_timer, SIGNAL(timeout()), this, SLOT(step()));
	m_timer->start(stepTime);
}

void KSpiral::Ball::step() {
	setPos(pos() + QPointF(m_dx, m_dy));
	emit ballStepped(this);
}

void KSpiral::Ball::stopShoot() {
	m_timer->stop();
}

void KSpiral::Ball::setPosAnimated(const QPointF& targetPos, int duration, int framesPerSecond) {
	QTimeLine* timeline = new QTimeLine();
	timeline->setDuration(duration);
	timeline->setFrameRange(0, duration * framesPerSecond / 1000);
	connect(timeline, SIGNAL(frameChanged(int )), this, SLOT(frameChanged(int )));
	connect(timeline, SIGNAL(finished()), this, SLOT(setPosCompleted()));
	m_stepPerFrame = (targetPos - pos()) / (duration * framesPerSecond / 1000);
	timeline->start();
}

void KSpiral::Ball::frameChanged(int frame) {
	const QPointF nextPoint = (frame - m_lastFrame) * m_stepPerFrame + pos();
	setPos(nextPoint);
	emit ballStepped(this);
	m_lastFrame = frame;
}

QList<KSpiral::Ball* > KSpiral::Ball::chainStep(qreal distance, SvgReader* path, int direction) {
	addState(KSpiral::ballInMove);
	QList<KSpiral::Ball* > ballsToDelete;
	qreal newPosition = position() + distance;
	// O-O->O
	if(direction >= 0 && m_nextBall) {
		if(euklid(pos() - m_nextBall->pos()) <= 2 * radius()) {
			removeState(KSpiral::ballIsStartball);
			if(m_states & KSpiral::ballIsEndball) kDebug() << "endball";
			ballsToDelete << m_nextBall->chainStep(distance, path, 1);
		}
		else {
			addState(KSpiral::ballIsStartball);
			m_nextBall->chainStep(0, path, 1);
		}
	}
	// O<-O-O
	if(direction <= 0 && m_prevBall) {
		if(euklid(pos() - m_prevBall->pos()) <= 2 * radius()) {
			removeState(KSpiral::ballIsEndball);
			ballsToDelete << m_prevBall->chainStep(distance, path, -1);
		}
		else {
			addState(KSpiral::ballIsEndball);
			m_prevBall->chainStep(0, path, -1);
		}
	}
	if (m_states & KSpiral::ballInChain) {
		if (newPosition < 0 || newPosition > path->length()) return ballsToDelete << this;

		setPosition(newPosition);
		setPos(path->C(newPosition));
		removeState(KSpiral::ballInMove);
	} else {
		arrangeBall();
	}
	return ballsToDelete;
}

void KSpiral::Ball::setPosition(qreal newPos) {
	m_position = newPos;
}

#include <kdebug.h>
void KSpiral::Ball::arrangeBall() {
	if(m_states & KSpiral::ballInChain) kDebug() << "bla";
	if(m_arrangeStepCounter >= 5) {
		addState(KSpiral::ballInChain);
		m_arrangeStepCounter = 0;
// 		emit rearrangeNeeded();
	}

	if(m_prevBall && m_nextBall)
		kDebug() << ballType() << m_prevBall->ballType() << m_nextBall->ballType();

	++m_arrangeStepCounter;
}

void KSpiral::Ball::rewriteStates() {
	if((m_nextBall && (abs(m_nextBall->position() - m_position) >= 2 * m_radius)) || !m_nextBall) {
		addState(KSpiral::ballIsStartball);
	} else {
		removeState(KSpiral::ballIsStartball);
	}
	if((m_prevBall && (abs(m_prevBall->position() - m_position) >= 2 * m_radius)) || !m_prevBall) {
		addState(KSpiral::ballIsEndball);
	} else {
		removeState(KSpiral::ballIsEndball);
	}
}

#include "ball.moc"
