/*
    This file is part of KSpiral.
    Copyright 2009 Felix Lemke <hobbyblobby@arcor.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#ifndef BALL_H
#define BALL_H

#include <cmath>
class KGameRenderer;
#include <KGameRenderedObjectItem>
#include <QBrush>
#include <QObject>
class SvgReader;

namespace KSpiral {
	enum BallState {
		ballNotInAction = 0x0,
		ballAsCannonball = 0x1,
		ballInCollidingCheck = 0x2,
		ballInChain = 0x4,
		ballInMove = 0x8,
		ballIsEndball = 0x10,		// -O-O-O   X-O-O    X = endBall
		ballIsStartball = 0x20      // -O-O-X   O-O-X    X = startBall
	};
	Q_DECLARE_FLAGS(BallStates, BallState)

	class Ball : public KGameRenderedObjectItem {
		Q_OBJECT
		public:
			Ball(const QPointF& pos, KGameRenderer* renderer, const QString& type, QGraphicsItem* parent = 0);
			virtual ~Ball();

			enum ItemTypes {BallItem = UserType + 1};
			int type() const {return KSpiral::Ball::BallItem;}

			KSpiral::BallStates states() const {return m_states;}
			void addState(KSpiral::BallState state);
			void removeState(KSpiral::BallState state);

			void setNextBall(KSpiral::Ball* nextBall) {m_nextBall = nextBall;}
			void setPrevBall(KSpiral::Ball* prevBall) {m_prevBall = prevBall;}
			KSpiral::Ball* nextBall() const {return m_nextBall;}
			KSpiral::Ball* prevBall() const {return m_prevBall;}

			void setPosition(qreal newPos);
			qreal position() const {return m_position;}
			void rewriteStates();
			qreal radius() const {return m_radius;}

			int increaseArrangeStepCounter() {return ++m_arrangeStepCounter;}

			QString ballType() const {return m_type;}

			//moving types
			//shoot = move until stopShoot();
			void shoot(const QPointF& startPos, qreal angle, qreal velocity); //velocity in points per second
			//setPosAnimated = animated move
			void setPosAnimated(const QPointF& targetPos, int duration, int framesPerSecond = 20);
			//chainStep = move that moves also m_nextBall and m_prevBall
			QList<KSpiral::Ball* > chainStep(qreal distance, SvgReader* path, int direction = 0);
			//returns a list of balls aren't on path
			//direction: 1 = forward, 0 = both directions, -1 = backward
			void arrangeBall(); //called from chainStep if ball is not in Chain, but chain moves him

			void stopShoot();
			QPair<qreal, qreal> moveDirection() const {return qMakePair(m_dx, m_dy);}

			qreal euklid(const QPointF& point) {return sqrt(pow(point.x(),2) + pow(point.y(),2));}
			qreal euklid(const qreal& x, const qreal& y) {return euklid(QPointF(x, y));}

		public Q_SLOTS:
			void step(); //for shoot
			void frameChanged(int frame); //for setPosAnimated
			void setPosCompleted() {emit setPosCompleted(this);}
		Q_SIGNALS:
			void ballStepped(KSpiral::Ball* );
			void setPosCompleted(KSpiral::Ball* ); //emitted, when setPosAnimated is complete
			void rearrangeNeeded();

		private:
			KSpiral::Ball* m_nextBall;
			KSpiral::Ball* m_prevBall;
			//ball property
			qreal m_radius;
			QString m_type;
			KSpiral::BallStates m_states;
			//position absolut on path as distance from start
			qreal m_position;
			//for ball flying
			QTimer* m_timer;
			qreal m_dx;
			qreal m_dy;
			//for setPosAnimated
			int m_lastFrame;
			QPointF m_stepPerFrame;
			//arrangeBall
			int m_arrangeStepCounter;
  };
}

#endif // BALL_H
