/***************************************************************************
 *   Copyright (C) 2003-2004 by Albert Astals Cid                          *
 *   tsdgeos@terra.es                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "board.h"

#include <kapplication.h>
#include <kglobal.h> // for Q_ASSERT

#include "evaluation.h"
#include "settings.h"

board::board(QObject *o, bool createKcases) : QObject(o)
{
	cases = new kcase** [3];
	for (int i = 0; i < 3; i++)
	{
		cases[i] = new kcase* [3];
		for (int j = 0; j < 3; j++)
		{
			if (createKcases) cases[i][j] = new kcase(i, j);
		}
	}
}

board::board(const board *original)
{
	currentPlayer = original -> currentPlayer;
	currentMode = original -> currentMode;
	player1 = original -> player1;
	player2 = original -> player2;
	player1PiecesLeft = original -> player1PiecesLeft;
	player2PiecesLeft = original -> player2PiecesLeft;
	selected = original -> selected;
	selected_x = original -> selected_x;
	selected_y = original -> selected_y;

	cases = new kcase** [3];
	for (int i = 0; i < 3; i++)
	{
		cases[i] = new kcase* [3];
		for (int j = 0; j < 3; j++)
		{
			cases[i][j] = new kcase(original -> cases[i][j]);
		}
	}
}

board::~board()
{
	for (int i = 0; i < 3; i++)
	{
		for (int j = 0; j < 3; j++)
		{
			delete cases[i][j];
		}
		delete[] cases[i];
	}
	delete[] cases;
}

void board::newGame()
{
	// we are not playing when we begin the game
	// we are playing once the first move is done
	setPlaying(false);
	currentMode = ktttSettings::game_mode();
	player1 = ktttSettings::player1();
	player2 = ktttSettings::player2();

	currentPlayer = kcase::PLAYER1;
	player1PiecesLeft = 3;
	player2PiecesLeft = 3;
	selected = false;
	selected_x = -1;
	selected_y = -1;

	for (int i = 0; i < 3; i++)
	{
		for (int j = 0; j < 3; j++)
		{
			cases[i][j] -> setEmpty();
		}
	}
	
	if (currentPlayer == kcase::PLAYER1 && player1 == ktttSettings::EnumPlayer1::Computer ||
		currentPlayer == kcase::PLAYER2 && player2 == ktttSettings::EnumPlayer2::Computer)
	{
		setPlaying(true);
		moveComputer();
	}
}

void board::endGame()
{
	Q_ASSERT(currentPlayer != kcase::NOBODY);
	currentPlayer = kcase::NOBODY;
	setPlaying(false);
}

void board::place(int x, int y)
{
	Q_ASSERT(x >= 0);
	Q_ASSERT(x < 3);
	Q_ASSERT(y >= 0);
	Q_ASSERT(y < 3);
	emit movement();
	cases[x][y] -> setOcupation(currentPlayer);
	nextTurn();
}

void board::select(int x, int y)
{
	Q_ASSERT(x >= 0);
	Q_ASSERT(x < 3);
	Q_ASSERT(y >= 0);
	Q_ASSERT(y < 3);
	selected = true;
	selected_x = x;
	selected_y = y;
	emit selection(true);
	cases[x][y] -> toggleSelected();
}

void board::unselect()
{
	cases[selected_x][selected_y] -> toggleSelected();
	selected = false;
	selected_x = -1;
	selected_y = -1;
	emit selection(false);
}

void board::placeSelected(int x, int y)
{
	selected = false;
	cases[selected_x][selected_y] -> setEmpty();
	place(x, y);
}

// TODO if i change to arbitrary size that should change
// steps > 3 is not necessary, event steps = ! is good enough
// but hey, this is a programming exercise not a commercial game
boardEvaluation* board::suggestMove(int steps)
{
	Q_ASSERT(steps > 0);
	bool allOccupied;
	boardEvaluation *area = new boardEvaluation();
	boardEvaluation *auxarea;
	evaluation *aux;
	board *copy;

	allOccupied = allCasesOccupied();
	if (ttt())
	{
		int value;
		// the current player has lost so mark all the positions
		// with it's worst score
		if (currentPlayer == kcase::PLAYER1) value = -10000;
		else value = 10000;
		for (int i = 0; i < 3; i++)
		{
			for (int j = 0; j < 3; j++)
			{
				area -> getEvaluation(i, j) -> setValue(value);
			}
		}
	}
	else
	{
		if (currentMode == ktttSettings::EnumGame_mode::Classic
		    || (currentPlayer == kcase::PLAYER1 && player1PiecesLeft > 0)
		    || (currentPlayer == kcase::PLAYER2 && player2PiecesLeft > 0) )
		{
			if (allOccupied)
			{
				for (int i = 0; i < 3; i++)
				{
					for (int j = 0; j < 3; j++)
					{
						area -> getEvaluation(i, j) -> setValue(evaluate());
					}
				}
			}
			else
			{
				for (int i = 0; i < 3; i++)
				{
					for (int j = 0; j < 3; j++)
					{
						if (cases[i][j] -> isFree())
						{
							copy = new board(this);
							copy -> cases[i][j] -> setOcupation(currentPlayer);
							if (currentPlayer == kcase::PLAYER1) copy -> player1PiecesLeft--;
							else /* currentPlayer = kcase::PLAYER2 */ copy -> player2PiecesLeft--;
							copy -> currentPlayer = kcase::otherPlayer(currentPlayer);
							if (steps > 1)
							{
								auxarea = copy -> suggestMove(steps-1);
								area -> setEvaluation(i, j, auxarea -> bestMove(copy -> currentPlayer));
								delete auxarea;
							}
							else
							{
								area -> getEvaluation(i, j) -> setValue(copy -> evaluate());
							}
							delete copy;
						}
					}
				}
			}
		}
		else
		{
			// PARTIAL OR TOTAL MODE WHEN MUST SELECT AND MOVE
			for (int i = 0; i < 3; i++)
			{
				for (int j = 0; j < 3; j++)
				{
					if (cases[i][j] -> getOcupation() == currentPlayer)
					{
						for (int k = 0; k < 3; k++)
						{
							for (int l = 0; l < 3; l++)
							{
								if (cases[k][l] -> isFree() && canMoveFrom(i, j, k, l))
								{
									copy = new board(this);
									copy -> cases[i][j] -> setOcupation(kcase::NOBODY);
									copy -> cases[k][l] -> setOcupation(currentPlayer);
									copy -> currentPlayer = kcase::otherPlayer(currentPlayer);
									if (steps > 1)
									{
										auxarea = copy -> suggestMove(steps-1);
										aux = auxarea -> bestMove(copy -> currentPlayer);
										aux -> setRemove(i, j);
 										area -> setEvaluation(k, l, area -> getEvaluation(k, l) -> bestScore(aux, currentPlayer));
										delete auxarea;
									}
									else
									{
										evaluation *t = new evaluation();
										t -> setValue(copy -> evaluate());
										area -> setEvaluation(k, l, area -> getEvaluation(k, l) -> bestScore(t, copy -> currentPlayer));
									}
									delete copy;
								}
							}
						}
					}
				}
			}
		}
	}

	return area;
}

bool board::playing()
{
	return p_playing;
}

void board::setPlaying(bool b)
{
	p_playing = b;
}

void board::nextTurn()
{
	setPlaying(true);
	if (ttt())
	{
		if (player1 == player2)
		{
			if (currentPlayer == kcase::PLAYER1) emit playerwin(1, true);
			else emit playerwin(2, true);
		}
		else if (player1 == ktttSettings::EnumPlayer1::Human)
		{
			if (currentPlayer == kcase::PLAYER1) emit win();
			else emit loose();
		}
		else
		{
			if (currentPlayer == kcase::PLAYER2) emit win();
			else emit loose();
		}
		endGame();
	}
	else
	{
		if (allCasesOccupied())
		{
			emit draw();
			endGame();
		}
		else
		{
			currentPlayer = kcase::otherPlayer(currentPlayer);
			if (currentPlayer == kcase::PLAYER1 && player1 == ktttSettings::EnumPlayer1::Computer ||
				currentPlayer == kcase::PLAYER2 && player2 == ktttSettings::EnumPlayer2::Computer) moveComputer();
		}
	}
}

void board::moveComputer()
{
	Q_ASSERT(currentPlayer == kcase::PLAYER1 && player1 == ktttSettings::EnumPlayer1::Computer ||
			currentPlayer == kcase::PLAYER2 && player2 == ktttSettings::EnumPlayer2::Computer);
	board *copy = new board(this);

	boardEvaluation *area = copy -> suggestMove(3);

	evaluation *move = area -> bestMove(currentPlayer);
	Q_ASSERT(move->possible());

	if (currentMode != ktttSettings::EnumGame_mode::Classic)
	{
		if (
			(currentPlayer == kcase::PLAYER1 && player1PiecesLeft == 0)
			||
			(currentPlayer == kcase::PLAYER2 && player2PiecesLeft == 0)
			)
		{
			Q_ASSERT(move->r_x() >= 0);
			Q_ASSERT(move->r_x() < 3);
			Q_ASSERT(move->r_y() >= 0);
			Q_ASSERT(move->r_y() < 3);
			cases[move->r_x()][move->r_y()] -> setEmpty();
		}
		else if (currentPlayer == kcase::PLAYER1) player1PiecesLeft--;
		else /* currentPlayer = kcase::PLAYER2 */ player2PiecesLeft--;
	}

	place(move->p_x(), move->p_y());
	
	delete copy;
	delete area;
	delete move;
}

bool board::allCasesOccupied() const
{
	bool Occupied = true;
	if (currentMode == ktttSettings::EnumGame_mode::Partial || currentMode == ktttSettings::EnumGame_mode::Total) Occupied = false;
	else
	{
		for (int i = 0; i < 3; i++)
		{
			for (int j = 0; j < 3; j++)
			{
				Occupied = Occupied && !cases[i][j]->isFree();
			}
		}
	}
	return Occupied;
}

bool board::ttt() const
{
	return (checkH(0) || checkH(1) || checkH(2) || checkV(0) || checkV(1)  || checkV(2) || checkD(0) || checkD(1));
}

bool board::canMoveFrom(int selected_x, int selected_y, int x, int y) const
{
	Q_ASSERT(selected_x >= 0 && selected_x < 3 && selected_y >= 0 && selected_y < 3
			&& x >= 0 && x < 3 && y >= 0 && y < 3);
	Q_ASSERT(currentMode != ktttSettings::EnumGame_mode::Classic);
	if (currentMode == ktttSettings::EnumGame_mode::Total) return true;
	int i = selected_x - x;
	int j = selected_y - y;
	if (i < 0) i = -i;
	if (j < 0) j = -j;
	if (i == 2 || j == 2) return false;
	else return true;
}

int board::evaluate()
{
	int score = 0;
	for (int i = 0; i < 3; i++) score += scoreH(i);
	for (int i = 0; i < 3; i++) score += scoreV(i);
	for (int i = 0; i < 2; i++) score += scoreD(i);
	return score;
}

bool board::checkH(int i) const
{
	Q_ASSERT(i >= 0 && i < 3);
	return (cases[0][i] -> getOcupation() == cases[1][i] -> getOcupation() &&
			cases[0][i] -> getOcupation() == cases[2][i] -> getOcupation() &&
			cases[0][i] -> getOcupation() != kcase::NOBODY);
}

bool board::checkV(int i) const
{
	Q_ASSERT(i >= 0 && i < 3);
	return (cases[i][0] -> getOcupation() == cases[i][1] -> getOcupation() &&
			cases[i][0] -> getOcupation() == cases[i][2] -> getOcupation() &&
			cases[i][0] -> getOcupation() != kcase::NOBODY);
}

bool board::checkD(int i) const
{
	Q_ASSERT(i >= 0 && i <= 1);
	if (i == 0)
	{
		return (cases[0][0] -> getOcupation() == cases[1][1] -> getOcupation() &&
				cases[0][0] -> getOcupation() == cases[2][2] -> getOcupation() &&
				cases[0][0] -> getOcupation() != kcase::NOBODY);
	}
	else
	{
		return (cases[2][0] -> getOcupation() == cases[1][1] -> getOcupation() &&
				cases[2][0] -> getOcupation() == cases[0][2] -> getOcupation() &&
				cases[2][0] -> getOcupation() != kcase::NOBODY);
	}
}

int board::scoreH(int i) const
{
	Q_ASSERT(i >= 0 && i < 3);
	int score = 0;
	int ocurrences = numberPlayerOcurrencesH(kcase::PLAYER1, i);
	if (ocurrences == 1) score += 1;
	else if (ocurrences == 2 && currentPlayer != kcase::PLAYER1) score += 10;
	else if (ocurrences == 2 && currentPlayer == kcase::PLAYER1) score += 100;
	else if (ocurrences == 3) score += 1000;

	ocurrences = numberPlayerOcurrencesH(kcase::PLAYER2, i);
	if (ocurrences == 1) score -= 1;
	else if (ocurrences == 2 && currentPlayer != kcase::PLAYER1) score -= 10;
	else if (ocurrences == 2 && currentPlayer == kcase::PLAYER1) score -= 100;
	else if (ocurrences == 3) score -= 1000;

	return score;
}

int board::scoreV(int i) const
{
	Q_ASSERT(i >= 0 && i < 3);
	int score = 0;
	int ocurrences = numberPlayerOcurrencesV(kcase::PLAYER1, i);
	if (ocurrences == 1) score += 1;
	else if (ocurrences == 2 && currentPlayer != kcase::PLAYER1) score += 10;
	else if (ocurrences == 2 && currentPlayer == kcase::PLAYER1) score += 100;
	else if (ocurrences == 3) score += 1000;

	ocurrences = numberPlayerOcurrencesV(kcase::PLAYER2, i);
	if (ocurrences == 1) score -= 1;
	else if (ocurrences == 2 && currentPlayer != kcase::PLAYER1) score -= 10;
	else if (ocurrences == 2 && currentPlayer == kcase::PLAYER1) score -= 100;
	else if (ocurrences == 3) score -= 1000;

	return score;
}

int board::scoreD(int i) const
{
	Q_ASSERT(i >= 0 && i < 2);
	int score = 0;
	int ocurrences = numberPlayerOcurrencesD(kcase::PLAYER1, i);
	if (ocurrences == 1) score += 1;
	else if (ocurrences == 2 && currentPlayer != kcase::PLAYER1) score += 10;
	else if (ocurrences == 2 && currentPlayer == kcase::PLAYER1) score += 100;
	else if (ocurrences == 3) score += 1000;

	ocurrences = numberPlayerOcurrencesD(kcase::PLAYER2, i);
	if (ocurrences == 1) score -= 1;
	else if (ocurrences == 2 && currentPlayer != kcase::PLAYER2) score -= 10;
	else if (ocurrences == 2 && currentPlayer == kcase::PLAYER2) score -= 100;
	else if (ocurrences == 3) score -= 1000;

	return score;
}

int board::numberPlayerOcurrencesH(kcase::PLAYER o, int i) const
{
	bool alone = true;
	int ocurrences = 0;
	int j = 0;
	while (alone && j < 3)
	{
		if (cases[i][j]->getOcupation() == o) ocurrences++;
		else if (cases[i][j]->getOcupation() == kcase::otherPlayer(o)) alone = false;
		j++;
	}
	if (alone) return ocurrences;
	else return 0;
}

int board::numberPlayerOcurrencesV(kcase::PLAYER o, int i) const
{
	bool alone = true;
	int ocurrences = 0;
	int j = 0;
	while (alone && j < 3)
	{
		if (cases[j][i]->getOcupation() == o) ocurrences++;
		else if (cases[j][i]->getOcupation() == kcase::otherPlayer(o)) alone = false;
		j++;
	}
	if (alone) return ocurrences;
	else return 0;
}

int board::numberPlayerOcurrencesD(kcase::PLAYER o, int i) const
{
	bool alone = true;
	int ocurrences = 0;
	int j, x, stepx;
	if (i == 1){ x = 0; stepx = 1; }
	else { x = 2; stepx = -1; }

	j = 0;
	while (alone && j < 3)
	{
		if (cases[x][j]->getOcupation() == o) ocurrences++;
		else if (cases[x][j]->getOcupation() == kcase::otherPlayer(o)) alone = false;
		j++;
		x += stepx;
	}
	if (alone) return ocurrences;
	else return 0;
}

#include "board.moc"

