//Magazynier, a warehouse assistant game.
//Copyright (C) 2009 Niels Slot nielsslot@gmail.com
//
//This program is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License along
//with this program; if not, write to the Free Software Foundation, Inc.,
//51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

#include "astar.h"

#include "levelModel.h"
#include "item.h"

#include <KDebug>

struct PathNode {
	// The position of this PathNode on the grid
	QPoint pos;
	// The parent is used to determine how we got to this
	//   PathNode starting at the starting point.
	PathNode *parent;
	// G is the movement cost to move from the starting point to this PathNode, following the path it took to get there
	int G;
	// H is the estimated cost to move from this PathNode to the goal point
	int H;
	// F is G + H
	int F;

	PathNode(QPoint p) : pos(p), parent(0), G(0), H(0), F(0)
	{
	}

	PathNode(QPoint p, PathNode* pa, int g, int h) : pos(p), parent(pa), G(g), H(h)
	{
		calculateF();
	}

	void calculateF()
	{
		F = G + H;
	}
};


static inline int indexOfNodeWithPos( const QPoint& pos, const QList<PathNode*>& list )
{
	for(int i=0;i<list.count(); ++i)
		if( list.at(i)->pos == pos )
			return i;

	return -1;
}

AStar::AStar()
{
}

QList<QPoint> *AStar::findPath(const LevelModel* levelModel, const QPoint& from, const QPoint& to)
{
	// Implementation of the A* pathfinding algorithm
	// This code is based on the code found in KLines (Kolour Lines)
	// See animator.cpp in the KLines source code
	// Thanks goes to the KLines developers and again to
	// Patrick Lester for the tutorial on gamedev.net
	// See http://www.gamedev.net/reference/articles/article2003.asp

	QList<PathNode*> openList;
	QList<PathNode*> closedList;

	openList.append(new PathNode(from));

	PathNode *curNode = 0;
	bool pathFound = false;

	// There are two exit conditions for this while
	// 1. We have found the 'to' position and are about
	//	to add it to the closed list.
	// 2. We have checked all the squares, and as
	//	as a result, the openList is empty
	while(true) {
		PathNode* minF = openList.at(0);
		for(int i=1; i<openList.count(); ++i)
			if( openList.at(i)->F < minF->F )
				minF = openList.at(i);

		curNode = minF;

		// Exit condition 1, 'to' position found
		if(curNode->pos==to) {
			pathFound = true;
			break; // while
		}

		closedList.append(curNode);
		openList.removeAll(curNode);

		QList<QPoint> adjacentSquares = findAdjacentSquares(curNode, levelModel);
		foreach(const QPoint& pos, adjacentSquares) {
			// Check whether the position is solid or not
			// Empty cell or Goal => Not solid
			const MagazynierItem* item = levelModel->getItem(pos);
			if(item->itemType()!=Empty && item->itemType()!=Goal)
				continue;

			// skip if closed list contains this square
			if(indexOfNodeWithPos(pos, closedList) != -1)
				continue;

			// search for node with position 'pos' in openList
			int inOpenList = indexOfNodeWithPos(pos, openList);
			if(inOpenList==-1) { //Not found
				PathNode* node = new PathNode(pos, curNode, curNode->G,
					(to-pos).manhattanLength());
				openList.append(node);
			}else{ // Found in openList
				PathNode* node = openList.at(inOpenList);
				// check if this path to square is better
				if( curNode->G < node->G )	{
					// yup, it's better, reparent and recalculate G,F
					node->parent = curNode;
					node->G = curNode->G;
					node->calculateF();
				}
			} // inOpenList==-1
		} //foreach

		// Is the openList empty?
		if(openList.isEmpty()) {
			// Yes? Then we've done all we could.
			// Exit condition 2, there is no path.
			pathFound = false;
			break;
		}
	} // while

	QList<QPoint> *path = NULL;

	if(pathFound)
	{
		path = new QList<QPoint>;
		// restoring path starting from last node:
		PathNode* node = curNode;
		while(node)
		{
			if(node->parent)
			    path->prepend(node->pos);
			node = node->parent;
		}
	}

	// cleanup
	qDeleteAll(openList);
	qDeleteAll(closedList);

	return path;
}

QList<QPoint> AStar::findAdjacentSquares(const PathNode* node, const LevelModel* levelModel)
{
	QList<QPoint> adjacentSquares;

	int x = node->pos.x();
	int y = node->pos.y();

	if(x != 0)
		adjacentSquares.append(QPoint(x-1, y));

	if(y != 0)
		adjacentSquares.append(QPoint(x, y-1));

	if(x != levelModel->getColumnCount()-1)
		adjacentSquares.append(QPoint(x+1,y));

	if(y != levelModel->getRowCount()-1)
		adjacentSquares.append(QPoint(x, y+1));

	return adjacentSquares;
}
