/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#ifndef DATA_H
#define DATA_H

#include <QList>
#include <QObject>
#include <QColor>

#include "description.h"
#include <KDebug>

class Data : public QObject
{
	Q_OBJECT

public:
	Data(QObject *parent=0);
	~Data() { }

	/**States each box in the grid can have.*/
	enum State {BOX_ON, BOX_OFF, BOX_UNDEFINED};

	/**Each box in the grid has a state and a color.*/
	struct MatrixEntry {
	    State box_status;
	    QColor box_color;
	};

	/**Each hint consists of an amount (consecutive boxes of same color),
	 * a color*/
	struct HintEntry {
	    int amount;
	    QColor hintColor;
	    bool active;
	};

	void setRowHints(int row, QList<HintEntry*> * values);
	void setRowHints(QList< QList<Data::HintEntry *> *> * values);

	void setColumnHints(int column, QList<HintEntry *> * values);
	void setColumnHints(QList< QList<Data::HintEntry *> *> * values);

	QList<Data::HintEntry *> * rowHints(int row);
	QList< QList<Data::HintEntry *> *> * rowHints();


	QList<Data::HintEntry *> * columnHints(int column);
	QList< QList<Data::HintEntry *> *> * columnHints();


	QList< QList<Data::MatrixEntry*> *> * matrix();

	int getRowCount();
	int getColumnCount();

	static QString toString(QList<HintEntry*> * entries);
	static QString toString(QList< QList<HintEntry*> *> * entries);

	static QList< QList<HintEntry*> *> * fromString(const QString & entries_string);

	static QString toString(QList< QList<MatrixEntry*> *> * entries);

	void setDescription(Description * description);
	Description * description();

	MatrixEntry* entry(int, int) const;

	QList<QColor> * availableColors();
	void setAvailableColors(QList<QColor> *);

	void update();

    virtual bool event(QEvent *);


public slots:
    /**Switches between the three states UNDEFINED, ON, OFF,
     * where ON is using the active color
     * */
	void togglePosition(int, int);
    void toggle_UndefinedOn(int x, int y);
    void toggle_UndefinedOff(int x, int y);
    void continueLastChange(int, int);

    void setUndefined(int x, int y);
    void setOff(int x, int y);
    void setOn(int x, int y);

	void resetGame();

	void setActiveColor(const QColor &);

signals:
	void positionChanged();
    void positionChanged(int, int);
	void colorsChanged(Data *);

protected:
    virtual void connectNotify(const char *signal);
    virtual void disconnectNotify(const char *signal);

private:



    /**Hold number of blocks in rows*/
    QList< QList<HintEntry * > * > * rowList;

    /**Hold number of blocks in columns*/
    QList< QList<HintEntry * > * > * colList;

    /**Hold currently set colors of boxes in the grid*/
    QList< QList<MatrixEntry * > * > * m_matrix;

    /**Which color will be used for the next block*/
    QColor m_currentColor;

    /**The color that's used for background painting*/
    QColor m_backGround;

    /**The description of this game*/
    Description * m_description;

    /**These colors can be used*/
    QList<QColor> * m_availableColors;

    State lastChange;

};

inline QDebug operator<<(QDebug dbg, const Data::HintEntry &e)
{

    QStringList l = QColor::colorNames();
    QString name = e.hintColor.name();
    for(int i = 0; i < l.size(); i++) {
        if (QColor(l.at(i)) == e.hintColor) {
            name = l.at(i);
        }
    }
 dbg.nospace() << "(" << e.amount << "@" << name << ")";

 return dbg.space();
}

inline QDebug operator<<(QDebug dbg, const Data::HintEntry *e)
{
    if (e == NULL){
        dbg.nospace() << "HintEntry is NULL - this won't work";
    } else {
     dbg.nospace() << *e;
    }

 return dbg;
}

#endif
