/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/



#include "ds/solver.h"

#include <KDebug>
#include <QPoint>
#include <QHash>

#include <iostream>

inline void wait() {
    kDebug() << "Waiting for you: ";
//     std::cin.get();
}

Solver::Solver(){
}

void Solver::setData(Data* data) {
    m_data = data;
}

void Solver::run() {
    findSolution();
}



void Solver::findSolution() {
    dirtyIntervals = new QList<Interval*>();
    entryToInterval = new QHash<DPoint, QList<Interval*>*>();

    //initial calculation of intervals: complete row/column
    kDebug() << "Initializing intervals";
    initializeIntervals(m_data->columnHints(), m_data->getRowCount(), true);
    initializeIntervals(m_data->rowHints(), m_data->getColumnCount(), false);


    kDebug() << "Getting rid of dirty intervals";
    while (! dirtyIntervals->isEmpty()) {
        Interval * interval = dirtyIntervals->first();
        dirtyIntervals->removeFirst();
        interval->setNotDirty();

        makeIntervalSmaller(interval);

        paintBoxesInInterval(interval);

        if (interval->dirty()) {
            dirtyIntervals->append(interval);
        }

        wait();
    }

    delete dirtyIntervals;
    delete entryToInterval;

    emit finished();
}

void Solver::initializeIntervals(
        QList<QList<Data::HintEntry*>*>* hintListList,
        int initialIntervalSize,
        bool markInColumns)
{
    for(int hintOrdinate = 0; hintOrdinate < hintListList->size(); hintOrdinate++) {
        QList<Data::HintEntry*>* currentHintlist =  hintListList->at(hintOrdinate);

        int hintSum = 0;
        int startIndex = 0;
        Data::HintEntry* previousHint = NULL;
        QList<Interval*>*intervalsInThisColumn = new QList<Interval*>();
        for(int h = 0; h < currentHintlist->size(); h++) {

            Data::HintEntry * currentHint = currentHintlist->at(h);

            if (previousHint != NULL &&
                    previousHint->hintColor == currentHint->hintColor ) {
                //two blocks of same color have at least one off-box between them
                startIndex ++;
            }

            hintSum += currentHint->amount;
            Interval * interval = new Interval(startIndex, initialIntervalSize, hintOrdinate, currentHint, markInColumns);
            interval->setDirty();
            dirtyIntervals->append(interval);
            intervalsInThisColumn->append(interval);
            for(int c = interval->beginIndex(); c < interval->endIndex(); c++ ) {
                int x = markInColumns?hintOrdinate:c;
                int y = markInColumns?c:hintOrdinate;
                DPoint p = DPoint(x, y, interval->isColumn());
                QList<Interval*>* intervalList = entryToInterval->take(p);
                if (intervalList == NULL) {
                    intervalList = new QList<Interval*>();
                }
                intervalList->append(interval);

                entryToInterval->insert(p, intervalList);
            }

            previousHint = currentHint;
            startIndex += currentHint->amount;
        }

        previousHint = NULL;
        int endIndex = initialIntervalSize;
        for(int h = intervalsInThisColumn->size() -1; h >= 0 ; h--) {

            Interval * currentInterval = intervalsInThisColumn->at(h);
            if (previousHint != NULL &&
                    previousHint->hintColor == currentInterval->block()->hintColor ) {
                //two blocks of same color have at least one off-box between them
                endIndex --;
            }
            currentInterval->setBlockSum(hintSum);
            adjustInterval(currentInterval, endIndex, true);
//             currentInterval->setEndIndex(endIndex);

            previousHint = currentInterval->block();
            endIndex -= currentInterval->block()->amount;
        }

        delete intervalsInThisColumn;
    }
}

void Solver::makeIntervalSmaller(Interval* interval) {
    if (interval->block()->amount==0) {
        //complete block is off
        for(int i = interval->beginIndex(); i < interval->endIndex(); i++) {
            setToState(interval, i, false);
        }
    } else {
        kDebug() << "Starting off with interval " << *interval;
        int firstMarked = interval->beginIndex()-1;
        int beginGap = -1;
        int lastMarked = interval->endIndex()+1;
        int endGap = interval->endIndex()+1;

        int gapAdjustedBegin = interval->beginIndex();
        int gapAdjustedEnd = interval->endIndex();

        bool correctlyAdjustedBeginning = false;
        bool everFoundPossibleGap = false;

        for(int i = interval->beginIndex(); i < interval->endIndex(); i++) {
            DPoint p = toPoint(interval, i, interval->isColumn());
            Data::MatrixEntry * entry = m_data->entry(p.x(), p.y());

            //if the orthogonal hints _only_ contain different colors, the row is impossible!
            QList<Data::HintEntry*> * orthogonalHints = interval->isColumn()?m_data->rowHints(i):m_data->columnHints(i);
            bool possibleBox = false;

            for(int hint = 0; hint < orthogonalHints->size() && !possibleBox; hint++) {

                possibleBox =
                        (orthogonalHints->at(hint)->hintColor == interval->block()->hintColor)
                        && orthogonalHints->at(hint)->amount > 0;

            }


            if (possibleBox) {
            //this box could be the beginning of a gap.

                if (beginGap < 0) {
                    kDebug() << "Found a place to be: " << p;
                    beginGap = i;
                }

//              remember this box, it could be the last one in the gap.
                endGap = i;
            } else {
                kDebug() << "This is no place to be: " << p;
//                 it counts only as a gap if it is large enough to hold
//                  the complete block.


                if (endGap - beginGap < (interval->block()->amount-1) || beginGap < 0) {
//                  too small => the complete gap can be skipped.
                    kDebug() << "And the previous gap was too small: " << beginGap << ".." << endGap << "=="<< endGap - beginGap <<"<<"<<(interval->block()->amount-1);
//                  adjust the beginning if it has not been adjusted yet
                    if (!correctlyAdjustedBeginning) {
                        correctlyAdjustedBeginning = everFoundPossibleGap;
                        gapAdjustedBegin = beginGap>0?beginGap:i+1;
                        beginGap = -1;
                    }
                } else {
                    everFoundPossibleGap = true;
                }
            }

            QList<Interval*>* correspondingIntervalList = entryToInterval->value(p);

            /*find first and last marked box - only if we know that this interval is exclusively
            * using this box.*/
            if (correspondingIntervalList == NULL
                    || correspondingIntervalList->size() == 1) {

                if (entry->box_status == Data::BOX_ON
                        && entry->box_color == interval->block()->hintColor) {

                    if (correspondingIntervalList != NULL)
                        kDebug() << "Found ON: "<<correspondingIntervalList << " at " << p;
                    else
                        kDebug() << "Found ON: NULL at " << p;

                    if (firstMarked == interval->beginIndex()) {
                        firstMarked = i;
                    }

                    lastMarked = i;
                }
            }
        }

        gapAdjustedEnd = endGap;
        kDebug() << "Found adjustable gap for " << *interval <<": " << gapAdjustedBegin << " .. " << gapAdjustedEnd;

        if (gapAdjustedBegin > interval->beginIndex()) {
            adjustInterval(interval, gapAdjustedBegin, false);
        }
        if (gapAdjustedEnd < interval->endIndex()) {
            adjustInterval(interval, gapAdjustedEnd, true);
        }

        //index of first marked box + length of block == max end index of interval.
        if (firstMarked > interval->beginIndex()){
            int maxEnd = firstMarked + interval->block()->amount;
            kDebug() << "firstMarked" << firstMarked << "maximum end: " << maxEnd;
            if (maxEnd < interval->endIndex()) {
                adjustInterval(interval, maxEnd, true);
            }

            kDebug() << "maxEnd:" << maxEnd <<" actual End:" << interval->endIndex() ;
        }

        if (lastMarked < interval->endIndex()) {
            //last marked box - length of block == min begin index of interval
            int minBegin = lastMarked - interval->block()->amount;
            kDebug() << "lastMarked"<<lastMarked << "minimum begin: " << minBegin;
            if (minBegin > interval->beginIndex()) {
                adjustInterval(interval, minBegin, false);
            }

            kDebug() <<"min Begin:"<< minBegin << " actual Begin:" << interval->beginIndex() ;
        }

    }
}

void Solver::adjustInterval(Interval * interval, int border, bool adjustTheEnd) {

    int excludeFromHere, excludeTillHere;
    if (adjustTheEnd) {
        excludeFromHere = border+1;
        excludeTillHere = interval->endIndex()+1;
        kDebug() << "Adjusting end: "<<excludeTillHere<<"->"<<excludeFromHere;
    } else {

        excludeFromHere = interval->beginIndex();
        excludeTillHere = border;
        kDebug() << "Adjusting begin: "<<excludeFromHere<<"->"<<excludeTillHere;
    }

    for(int difference = excludeFromHere; difference < excludeTillHere; difference++) {
        kDebug() << "Calculating point ";
        //in all intervallists of this box, remove this interval
        DPoint p = toPoint(interval, difference, interval->isColumn());
        kDebug() << "Get other interval at " << p;
        QList<Interval*>* otherInterval = entryToInterval->value(p);

        if (otherInterval != NULL) {
            kDebug() << "Found " << interval << "in between "<<(adjustTheEnd?"END":"BEGIN")<<": "<<otherInterval  << " at " << p;
            kDebug() << "End: beforeRemoval "<<*otherInterval;
            otherInterval->removeOne(interval);
            kDebug() << "End: afterRemoval "<<*otherInterval;
            if (otherInterval->isEmpty()) {
                setToState(interval, difference, false);
            }
        }
        kDebug()<<"Done";
    }
    kDebug() << "Adjusting "<<(adjustTheEnd?"END":"BEGIN")<<" from "
            << (adjustTheEnd?interval->endIndex():interval->beginIndex()) << " to " << border;
    if (adjustTheEnd) {
        interval->setEndIndex(border);
    } else {
        interval->setBeginIndex(border);
    }
    kDebug() << "Adjusted interval: " << *interval;
}

void Solver::paintBoxesInInterval(Interval* interval) {
    kDebug() << "Painting in interval " << *interval;
    int freeFields = interval->length() - interval->block()->amount +1;
    kDebug() << "freeFields" << interval->length() <<"-"<< interval->block()->amount <<"="<<freeFields;
    int mark = interval->block()->amount - freeFields;
    kDebug() << "mark: " << interval->block()->amount <<"-"<< freeFields <<"="<<mark;
    int startIndex = interval->beginIndex();
    startIndex += (double) (interval->length() - mark) / 2 +0.5;
    int endIndex = startIndex+mark;

    for(int i = startIndex; mark > 0 && i<endIndex; i++) {

        if (i==startIndex) {
            QString q = QString(interval->isColumn()?"column":"row");
            kDebug() << "MARKING ["<<mark<<"] beginning at ["<<startIndex<<"] in "<<q<<" ["<<interval->column()<<"]";
        }
        m_data->setActiveColor(interval->block()->hintColor);

        setToState(interval, i, true);

        DPoint p = toPoint(interval, i, interval->isColumn());

        QList<Interval*>* intervalList = entryToInterval->value(p);
        if (intervalList != NULL) {
            kDebug() << "Removing interval "<< *interval <<" from position "<<p << "which has: " << intervalList;
            intervalList->removeAll(interval);
            if (intervalList->isEmpty()) {
                entryToInterval->take(p);

                delete intervalList;
            }
            kDebug() << "Finished removing";
        }

    }

    if (startIndex == interval->beginIndex() && endIndex == interval->endIndex()) {
        kDebug() << "Interval finished - deleting it";
        dirtyIntervals->removeAll(interval);
        delete interval;
    }
}

DPoint Solver::toPoint(Interval* interval, int index, bool direction) {
    int x, y;
    if (interval->isColumn()) {
        x = interval->column();
        y = index;
    } else {
        x = index;
        y = interval->column();
    }
    return DPoint(x, y, direction);
}

void Solver::setToState(Interval* interval, int index, bool on) {

    DPoint p = toPoint(interval, index, interval->isColumn());

    Data::State state = m_data->entry(p.x(), p.y())->box_status;
    if (state == (on ? Data::BOX_ON : Data::BOX_OFF) ) {
        return;
    }

    if (state != Data::BOX_UNDEFINED) {
        kDebug() << "I want to turn " << p << (on ? " ON":" OFF") << ". But it is already " << state;
        wait();
    }

    if (on) {
        kDebug() << "Setting " << p << " ON";
        m_data->setOn(p.x(), p.y());
    } else {
        kDebug() << "Setting " << p << " OFF";
        m_data->setOff(p.x(), p.y());
    }
    kDebug() << "Done setting";

    markDirty(p, interval, true);
    markDirty(p, interval, false);
}

void Solver::markDirty(QPoint point, Interval* interval, bool direction) {

    DPoint p = DPoint(point, direction);
    //each row that has been touched here must be set dirty.
    QList<Interval*>* intervalList = entryToInterval->value(p);

    if (intervalList != NULL)
        kDebug() << "These intervals are at"<<p<<":"<<intervalList;

    for(int intervalCount = 0; intervalList != NULL && intervalCount<intervalList->size(); intervalCount++) {

        Interval* dirtyInterval = intervalList->at(intervalCount);
        if (dirtyInterval != interval &&
            ! dirtyInterval->dirty()) {
            dirtyInterval->setDirty();
            kDebug() << "no more intervals";
            kDebug() << "Appending newly dirtyfied interval: "<< *dirtyInterval;
            kDebug() << "no more intervals";
            dirtyIntervals->append(dirtyInterval);
        }
    }
}



Interval::Interval(const int beginAt, const int endAt, const int column,
                    Data::HintEntry* const block, const bool isColumn)
:    _beginIndex(beginAt), _endIndex(endAt), _column(column),
     _block(block), _isColumn(isColumn)
{
    setNotDirty();
}

void Interval::setBeginIndex(int newBegin) {
    _beginIndex = newBegin;
}


int Interval::beginIndex() const {
    return _beginIndex;
}

void Interval::setEndIndex(int newEnd) {
    _endIndex = newEnd;
}


int Interval::endIndex() const {
    return _endIndex;
}

int Interval::length() const {
    return _endIndex - _beginIndex;
}

int Interval::column() const {
    return _column;
}

bool Interval::isColumn() const {
    return _isColumn;
}

bool Interval::dirty() const {
    return _dirty;
}

void Interval::setDirty() {
    _dirty = true;
}

void Interval::setNotDirty() {
    _dirty = false;
}

int Interval::blockSum() {
    return _blockSum;
}

void Interval::setBlockSum(int newBlockSum) {
    _blockSum = newBlockSum;
}

Data::HintEntry* Interval::block() const {
    return _block;
}

DPoint::DPoint(int x, int y, bool direction) : QPoint(x, y), direction(direction) {

}

DPoint::DPoint(QPoint p, bool direction) : QPoint(p), direction(direction) {

}


bool DPoint::getDirection() const {
    return direction;
}







