/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "game_info.h"

#include "ds/data.h"
#include "ds/package.h"
#include "ds/config.h"

#include "io/nop_format.h"

#include <QPainter>
#include <QDir>
#include <KAction>
#include <klocalizedstring.h>
#include <KActionCollection>
#include <KComboBox>
#include <KDebug>
#ifndef KXMLGUIWINDOW_H
    #include <kxmlguiwindow.h>
#endif
#include <KComponentData>
#include <KConfig>
#include <KConfigGroup>
#include <KGlobal>

#include <QHBoxLayout>
#include <QPushButton>
#include <QLabel>

#include <ThreadWeaver/Job>
#include <ThreadWeaver/JobSequence>
#include <ThreadWeaver/Weaver>
#include <ThreadWeaver/Thread>

#include <KStandardDirs>


LoaderJob::LoaderJob(QString filename, Package * package, QObject *parent)
    : ThreadWeaver::Job(parent), m_filename(filename), m_package(package)
{
}

void LoaderJob::doRun() {
    run();
    emit done(this);
}

Package* LoaderJob::package() const {
    return m_package;
}


void LoaderJob::run() {

    NopFormat * in = new NopFormat(m_package, new QObject());
    in->readPackageFile(m_filename);

//     Exaggeration for educational purposes:
//     thread()->msleep(800);

    kDebug() << thread() << " Found [" << m_package->name() <<"] in package file: " << m_package;
}

GameInfo::GameInfo(KXmlGuiWindow *parent) :
    QWidget(parent)
{
    setAutoFillBackground(true);

    m_currentGame = -1;

    setupConnectionsAndLayout();
    setupActions(parent);

    readConfig();
}


void GameInfo::readPackages() {

    bool runThreaded = false;

    KStandardDirs standardDirs;
    QStringList gamedirs = standardDirs.findDirs("data", "nonogram/games");

    for (int i = 0; i < gamedirs.size(); ++i) {
        QString gameDirName = gamedirs.at(i);

        kDebug() << "Searching for game packages in " << gameDirName;

        QDir * gameDir = new QDir(gameDirName);
        QStringList games = gameDir->entryList();

        for (int j = 0; j < games.size(); ++j) {
            QString gamefileName = games.at(j);

            if (!gamefileName.endsWith("nop") ) {
                continue;
            }
            QString gamefile = gameDirName + '/' + gamefileName;

            LoaderJob * job = new LoaderJob(gamefile, new Package());

            NopFormat * in = new NopFormat();
            Package * package= in->readPackageFile(gamefile);

            kDebug() << thread() << " Found [" << package->name() <<"] in package file: " << gamefile;
            connect(job, SIGNAL(done(ThreadWeaver::Job*)), this, SLOT(jobFinished(ThreadWeaver::Job*)));


            if (runThreaded) {
                ThreadWeaver::Weaver::instance()->enqueue (job);
            } else {
                job->doRun();
            }
        }
    }
}


void GameInfo::addPackage(Package* package) {
    kDebug() << "Adding package " << package->name() << "in thread " << thread();

    QVariant variant = qVariantFromValue( package );
    m_packageSelector->addItem(package->name(), variant );
    m_packageSelector->update();

    if (package->filename() == m_lastPackage) {
        setPackage(package);
    }
}

void GameInfo::jobFinished(ThreadWeaver::Job* job) {
    kDebug() << "Job is finished";

    LoaderJob * lJob = dynamic_cast<LoaderJob*>(job);
    addPackage(lJob->package());

    ThreadWeaver::Weaver::instance()->dequeue( job );
    delete job;
}

void GameInfo::setPackage(Package * package)
{
    kDebug() << "Changing current package to " << package->name();

    m_package = package;
    m_packageSelector->setCurrentItem(m_package->name());

    m_gameSelector->clear();

    for (int i = 0; i < m_package->size(); ++i) {
        QString hint = m_package->at(i)->description()->name();
        m_gameSelector->addItem(hint);
    }

    KConfigGroup savegame = KGlobal::config()->group(Config::CONFIG_GROUP_SAVEGAME);
    savegame.writeEntry( Config::CONFIG_LASTUSED_PACKAGE, m_package->filename() );

    game(m_lastGameInPackage);
}

void GameInfo::setPackage(int i)
{

    kDebug() << "Changing current package to " << QString::number(i);

    if (i>=m_packageSelector->count() || i<0) {
        return;
    }

    m_packageSelector->setCurrentIndex(i);

    QVariant variant = m_packageSelector->itemData(i);
    Package * package = variant.value< Package * >();
    m_lastGameInPackage = 0;
    setPackage(package);
}


void GameInfo::nextPackage() {
    setPackage( m_packageSelector->currentIndex() +1 );
}


void GameInfo::previousPackage() {
    setPackage( m_packageSelector->currentIndex() -1 );
}


void GameInfo::nextGame()
{
    if (m_currentGame + 1 < m_package->size()) {
        game(m_currentGame+1);
    }
}

void GameInfo::previousGame()
{
    if (m_currentGame > 0) {
        game(m_currentGame-1);
    }
}

void GameInfo::game(int i)
{
    m_currentGame = i;

    KConfigGroup savegame = KGlobal::config()->group(Config::CONFIG_GROUP_SAVEGAME);
    savegame.writeEntry( Config::CONFIG_LASTUSED_GAME, i );

    m_gameSelector->setCurrentIndex(i);

    Data * data = m_package->at(m_currentGame);
    emit changedGameTo(data);
}

void GameInfo::setupActions(KXmlGuiWindow* parent) {
    KAction* prevAction = new KAction(this);
    prevAction->setText(i18n("Previous Game"));
    prevAction->setIcon(KIcon("arrow-left"));
    prevAction->setShortcut(Qt::ALT+Qt::Key_Left);
    parent->actionCollection()->addAction("prevGame", prevAction);
    connect(prevAction, SIGNAL(triggered(bool)), SLOT(previousGame()));

    KAction* nextAction = new KAction(this);
    nextAction->setText(i18n("Next Game"));
    nextAction->setIcon(KIcon("arrow-right"));
    nextAction->setShortcut(Qt::ALT+Qt::Key_Right);
    parent->actionCollection()->addAction("nextGame", nextAction);
    connect(nextAction, SIGNAL(triggered(bool)), SLOT(nextGame()));

    KAction* prevPackageAction = new KAction(this);
    prevPackageAction->setText(i18n("Previous Package"));
    prevPackageAction->setIcon(KIcon("arrow-left-double"));
    prevPackageAction->setShortcut(Qt::ALT+Qt::Key_Up);
    parent->actionCollection()->addAction("prevPackage", prevPackageAction);
    connect(prevPackageAction, SIGNAL(triggered(bool)), SLOT(previousPackage()));

    KAction* nextPackageAction = new KAction(this);
    nextPackageAction->setText(i18n("Next Package"));
    nextPackageAction->setIcon(KIcon("arrow-right-double"));
    nextPackageAction->setShortcut(Qt::ALT+Qt::Key_Down);
    parent->actionCollection()->addAction("nextPackage", nextPackageAction);
    connect(nextPackageAction, SIGNAL(triggered(bool)), SLOT(nextPackage()));
}

void GameInfo::setupConnectionsAndLayout() {
    QPushButton *prev = new QPushButton("<<", this);
    prev->setStyleSheet("color: black");
    connect(prev, SIGNAL(clicked()), SLOT(previousGame()));

    QPushButton *next = new QPushButton(">>", this);
    next->setStyleSheet("color: black");
    connect(next, SIGNAL(clicked()), SLOT(nextGame()));

    m_gameSelector = new KComboBox();
    m_gameSelector->setEditable(false);
    connect(m_gameSelector, SIGNAL(activated(int)), SLOT(game(int)));

    m_packageSelector = new KComboBox();
    m_packageSelector->setEditable(false);
    connect(m_packageSelector, SIGNAL(activated(int)), SLOT(setPackage(int)));

    QHBoxLayout *layout = new QHBoxLayout;
    layout->addWidget(m_packageSelector);
    layout->addWidget(prev);
    layout->addWidget(m_gameSelector);
    layout->addWidget(next);

    setLayout(layout);
}



void GameInfo::readConfig() {
    KConfigGroup config = KGlobal::config()->group(Config::CONFIG_GROUP_COLORS);

    QColor background_infopart = config.readEntry( Config::CONFIG_COLOR_BACKGROUND_GAMESELECT, Config::DEFAULT_COLOR_BACKGROUND_GAMESELECT );

    setPalette(QPalette(background_infopart));

    KConfigGroup savegame = KGlobal::config()->group(Config::CONFIG_GROUP_SAVEGAME);
    m_lastPackage = savegame.readEntry( Config::CONFIG_LASTUSED_PACKAGE, QString("") );
    m_lastGameInPackage = savegame.readEntry( Config::CONFIG_LASTUSED_GAME, 0 );

    update();

}


