/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "grid_center.h"

#include <QPainter>
#include <QMouseEvent>
#include <QPaintEvent>

#include <KDebug>

#include "ds/data.h"

#include <KComponentData>
#include <KConfig>
#include <KConfigGroup>
#include <KGlobal>
#include <ds/config.h>

GridCenter::GridCenter(QWidget *parent) :
    Abstract_Grid(parent)
{
    lastMoveGridPosition = new QPoint();
    lastPressedButton = Qt::NoButton;
}

bool GridCenter::doPaintHorizontal()
{
    return true;
}

bool GridCenter::doPaintVertical()
{
    return true;
}

int GridCenter::widthOfLastLine()
{
    return m_maxWidth;
}

int GridCenter::heightOfLastLine()
{
    return m_maxHeight;
}

void GridCenter::setData(Data* newData) {
    Abstract_Grid::setData(newData);

    kDebug() << "connecting additional signals to data " << newData->description()->name();
    //signals from the grid: tell the data to react to a click
    connect(this, SIGNAL(togglePosition(int, int)), m_data, SLOT(togglePosition(int, int)));
    connect(this, SIGNAL(toggle_UndefinedOn(int, int)), m_data, SLOT(toggle_UndefinedOn(int, int)));
    connect(this, SIGNAL(toggle_UndefinedOff(int, int)), m_data, SLOT(toggle_UndefinedOff(int, int)));

    connect(this, SIGNAL(continueLastChange(int,int)), m_data, SLOT(continueLastChange(int,int)));
}

void GridCenter::paintAdditionalVertical(QPainter * /*painter*/, int /*pos_y*/, int /*row*/)
{
    /*do nothing*/
}

void GridCenter::paintAdditionalHorizontal(QPainter * painter, int pos_y, int row)
{
    /*paint boxes first so that lines overdraw them*/
    for (int x=0; x<m_cols; x++) {
        int pos_x = x*m_h_spacing;
        Data::MatrixEntry * m = m_data->entry(x, row);
        paintBox(pos_x+1+HORIZONTAL_LEFT_OFFSET, pos_y+1+VERTICAL_UPPER_OFFSET, m_h_spacing-2, m_v_spacing-2, m, painter);
    }
}

void GridCenter::paintBox(int x, int y, int width, int height,
        Data::MatrixEntry * entry, QPainter * painter)
{
    if (entry->box_status == Data::BOX_OFF) {

        //first fill with background
        painter->setPen(entry->box_color);
        painter->setBrush(entry->box_color);
        painter->drawRect(x, y, width, height);

        //and a little circle on that
        painter->setPen(offStateDotColor);
        painter->setBrush(offStateDotColor);
        painter->drawEllipse(x+(width/2), y+(height/2), width/8, height/8);
    } else if (entry->box_status == Data::BOX_ON) {

        painter->setPen(entry->box_color);
        painter->setBrush(entry->box_color);
        painter->drawRect(x, y, width, height);
    } else {
        //TODO: Should BOX_UNDEFINED status be painted?
    }
}

void GridCenter::mousePressEvent(QMouseEvent * event)
{

    QPoint *p = getGridPositions(event);

    /*propagate the click only if it is reasonable */
    if ( p->x() >= m_cols
        || p->y() >= m_rows
        || p->x() < 0
        || p->y() < 0) {

        return;
    }

    lastPressedButton = event->button();

    kDebug() << "["<<p->x()<<"/"<< p->y()<<"]";

    /*sane click? then tell everyone who's interested.*/

    if (event->button() == Qt::LeftButton) {

        toggleCurrent();
    } else if (event->button() == Qt::MidButton) {

        emit toggle_UndefinedOn(p->x(), p->y());
    } else if (event->button() == Qt::RightButton) {

        emit toggle_UndefinedOff(p->x(), p->y());
    }

}

void GridCenter::mouseReleaseEvent(QMouseEvent* event) {
    QWidget::mouseReleaseEvent(event);
    lastPressedButton = Qt::NoButton;
}


void GridCenter::mouseMoveEvent(QMouseEvent* event) {
    Abstract_Grid::mouseMoveEvent(event);

    if (lastPressedButton != Qt::NoButton) {
        kDebug() << "Move with pressed Button? ";

        QPoint *p = getGridPositions(event);

        /*propagate the click only if it is reasonable */
        if ( p->x() >= m_cols
            || p->y() >= m_rows
            || p->x() < 0
            || p->y() < 0) {

            return;
        }

         /*if nothing has changed, there's no need to get all fussy.*/
        if ( lastMoveGridPosition->x() == p->x() && lastMoveGridPosition->y() == p->y() )
        {
            return;
        }

        lastMoveGridPosition->setX(p->x());
        lastMoveGridPosition->setY(p->y());

        kDebug() << "Propagate change to: ["<<p->x()<<"/"<< p->y()<<"]";

        emit continueLastChange(p->x(), p->y());
        emit cellSelected(p);
    }
}

void GridCenter::up() {
    QPoint * newSelected = new QPoint(selectedCell->x(), selectedCell->y()-1);
    adjustManualSelection(newSelected);
    emit cellSelected(newSelected);
}

void GridCenter::down() {
    QPoint * newSelected = new QPoint(selectedCell->x(), selectedCell->y()+1);
    adjustManualSelection(newSelected);
    emit cellSelected(newSelected);
}

void GridCenter::left() {
    QPoint * newSelected = new QPoint(selectedCell->x()-1, selectedCell->y());
    adjustManualSelection(newSelected);
    emit cellSelected(newSelected);
}

void GridCenter::right() {
    QPoint * newSelected = new QPoint(selectedCell->x()+1, selectedCell->y());
    adjustManualSelection(newSelected);
    emit cellSelected(newSelected);
}

QPoint* GridCenter::adjustManualSelection(QPoint* p) {

    if (selectedCell->x()<0) p->setX(0);
    else if (p->x() < 0) p->setX(m_cols);
    else if (p->x() > m_cols) p->setX(0);

    if (selectedCell->y()<0) p->setY(0);
    else if (p->y() < 0) p->setY(m_rows);
    else if (p->y() > m_rows) p->setY(0);

    return p;
}


void GridCenter::toggleCurrent() {
    kDebug() << "Toggling current [" << selectedCell->x() << "/" << selectedCell->y() << "]";
//     m_data->togglePosition(selectedCell->x(), selectedCell->y());
    emit togglePosition(selectedCell->x(), selectedCell->y());
}

void GridCenter::readConfig() {
    Abstract_Grid::readConfig();

    KConfigGroup config = KGlobal::config()->group(Config::CONFIG_GROUP_COLORS);

    offStateDotColor = config.readEntry( Config::CONFIG_COLOR_BOX_OFFSTATE_DOT,
        Config::DEFAULT_COLOR_BOX_OFFSTATE_DOT );

}

