/***************************************************************************
 *   Copyright (C) 2008 Copyright 2008  Tom Vollerthun <vollerthun@gmx.de>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public
 *   License as published by the Free Software Foundation; either
 *   version 2 of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 ***************************************************************************/

#include "grid_header.h"

#include <QPainter>
#include <KDebug>
#include <QPoint>
#include <QString>

#include <KComponentData>
#include <KConfig>
#include <KConfigGroup>
#include <KGlobal>
#include <ds/config.h>



GridHeader::GridHeader(QWidget *parent, bool isHorizontal) :
    Abstract_Grid(parent)
{
    hasHorizontalOrientation = isHorizontal;
    m_hintRect = NULL;
}

bool GridHeader::doPaintHorizontal()
{
    return hasHorizontalOrientation;
}

bool GridHeader::doPaintVertical()
{
    return !hasHorizontalOrientation;
}

int GridHeader::widthOfLastLine()
{
    if (hasHorizontalOrientation) {
        return width()-2;
    }
    else {
        return m_maxWidth;
    }
}

int GridHeader::heightOfLastLine()
{
    if (hasHorizontalOrientation) {
        return m_maxHeight;
    }
    else {
        return height()-2;
    }
}

void GridHeader::calculateHintRect(QPainter * painter)
{
    if (m_hintRect == NULL) {
        QFont hintFont = painter->font();
        hintFont.setBold(true);
        painter->setFont(hintFont);

        QFontMetrics fm = QFontMetrics(hintFont);
        m_hintRect = new QRect(0, 0, fm.width(QString::number(99))+2, fm.height());
    }
}

/*paints row-hints*/
void GridHeader::paintAdditionalHorizontal(QPainter * painter, int pos_y,
        int row)
{
    calculateHintRect(painter);

    QList<Data::HintEntry *> * rowHints = m_data->rowHints(row);
    for (int i = rowHints->size() -1; i >=0; i--) {

        Data::HintEntry * hint = rowHints->at(rowHints->size()-1-i);

        QRect * r = new QRect(width()-((1+i)*m_hintRect->width()), pos_y+VERTICAL_UPPER_OFFSET+(m_v_spacing - m_hintRect->height())/2,
                              m_hintRect->width(), m_hintRect->height());

        paintHint(painter, r, hint);
    }
}

/*paints column-hints*/
void GridHeader::paintAdditionalVertical(QPainter * painter, int pos_x,
        int column)
{
    calculateHintRect(painter);

    QList<Data::HintEntry *> * columnHints = m_data->columnHints(column);
    for (int i = columnHints->size() -1; i >=0; i--) {

        Data::HintEntry * hint = columnHints->at(columnHints->size() -1 -i);

        QRect * r = new QRect(pos_x+HORIZONTAL_LEFT_OFFSET+(m_h_spacing - m_hintRect->width())/2, height()-((1+i)*m_hintRect->height()),
                              m_hintRect->width(), m_hintRect->height());

        paintHint(painter, r, hint);
    }
}

void GridHeader::paintHint(QPainter * painter, QRect * inset, Data::HintEntry * hintEntry)
{
    /*paint number with correct color*/
    QFont hintFont = painter->font();
    hintFont.setBold(true);
    painter->setFont(hintFont);

    painter->setPen(hintEntry->hintColor);
    painter->drawText(*inset, Qt::AlignCenter, QString::number(hintEntry->amount));

    if (! hintEntry->active) {
        painter->setPen(hintCrossthroughColor);
        QPen p = painter->pen();
        p.setWidth(2);
        painter->setPen(p);

        painter->drawLine(inset->topLeft(), inset->bottomRight());
        painter->drawLine(inset->bottomLeft(), inset->topRight());
    }


}

void GridHeader::mousePressEvent(QMouseEvent * event)
{
    if (m_hintRect == NULL) {
        /*mousepress before first paint event? --> Discard*/
        return;
    }

    calculatePositions();

    int x_world = event->x() - HORIZONTAL_LEFT_OFFSET;
    int y_world = event->y() - VERTICAL_UPPER_OFFSET;

    if (hasHorizontalOrientation) {

        int y_grid = y_world / m_v_spacing;

        QList<Data::HintEntry *> * row = m_data->rowHints(y_grid);

        int x_grid = (width() - x_world) / m_hintRect->width();
        x_grid = (row->size()-1) - x_grid;

        if (x_grid >=0 && x_grid < row->size()) {

            kDebug() << "["<<x_grid<<"/"<< y_grid<<"]";

            Data::HintEntry * hintEntry = row->at(x_grid);
            hintEntry->active = !hintEntry->active;
        }
    }
    else {

        int x_grid = x_world / m_h_spacing;

        QList<Data::HintEntry *> * col = m_data->columnHints(x_grid);

        int y_grid = (height() - y_world) / m_hintRect->height();
        y_grid = (col->size()-1) - y_grid;

        if (y_grid >= 0 && y_grid < col->size()) {

            kDebug() << "["<<x_grid<<"/"<< y_grid<<"]";

            Data::HintEntry * hintEntry = col->at(y_grid);
            hintEntry->active = !hintEntry->active;
        }
    }

    update();

}

void GridHeader::readConfig() {
    Abstract_Grid::readConfig();

    KConfigGroup config = KGlobal::config()->group(Config::CONFIG_GROUP_COLORS);

    hintCrossthroughColor = config.readEntry( Config::CONFIG_COLOR_HINT_CROSSTHROUGH,
        Config::DEFAULT_COLOR_HINT_CROSSTHROUGH );
}


